/*
** File  : AudioAccept.c
** Author: Koen Martens <koen.martens@tomtom.com>
** Desc. : Abstraction layer for sending sound to the
**	   audio_accept interface
**
*/

#include <assert.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <basetypes.h>
#include <TTMP.h>
#include <sysvsem.h>
#include <stdlib.h>
#include <signal.h>

#ifdef NMP3DEBUG
#undef DEBUG
#endif

#include "AudioAccept.h"
#include "Debug.h"
#include "A2DPErrors.h"

#define MIN( a, b ) ( ( ( a ) < ( b ) ) ? ( a ) : ( b ) )

#define MAX_NUMBUFS     4
#define RECONNECT_COUNT 50

typedef enum { EBlocking, ENonBlocking } ESendMessageType;

static TTMPConnection ttmp_conn       = NULL;
static TTMPMessage    ttmp_sync_msg   = NULL;
static TTMPMessage    ttmp_start_msg  = NULL;
static TTMPMessage    ttmp_stop_msg   = NULL;
static TTMPMessage    ttmp_ui_msg     = NULL;
static TTMPMessage    ttmp_pause_msg  = NULL;
static TTMPMessage    ttmp_resume_msg = NULL;

static void *ptr = NULL;

static int fd  = -1;
static int sem = -1;
 
static volatile int             buflen              = 0;
static volatile TUint32*        len[ MAX_NUMBUFS ]  = { NULL };
static volatile void*           bufs[ MAX_NUMBUFS ] = { NULL };
static volatile int             offsets[ MAX_NUMBUFS ];
static volatile int             current_buffer      = 0; // Number of current buffer
static volatile int             sync_mode           = 0;
static int                      buf_count           = 0;
static char                     file_name[ 128 ];
static char                     sock_name[ 128 ];

// A2DP error storage, done for compiling
unsigned int iLastA2DPErrorCode = A2DP_ERROR_UNKNOWN;

static ui_output_contoller_t* ui_controller     = NULL;
static int                    tmpReconnectCount = RECONNECT_COUNT;

static void ReportError( const char* aText, int ec )
{
 char errStr[ 2048 ];

 assert( NULL != aText );

 TTMPErrorString( ec, errStr, 2048 );
 DBG( "%s: %s, errno %d, text %s\n", aText, errStr, errno, strerror( errno ) );
}

static int AudioAcceptReconnect( const char* aSockName )
{ 

 assert( NULL != aSockName );

 if( NULL != ttmp_conn )
 {
  TTMPConnectionDestroy( ttmp_conn );
  ttmp_conn = NULL;
 }

 // open TTMP connection
 int ec = TTMPConnectionCreate( &ttmp_conn, TTMP_TO_INFINITE, 1, 0 );

 if( ec != TTMP_ERR_NONE) {
    ReportError( "Error creating connection", ec );
    return -1;
  }

  ec = TTMPConnectionConnect( ttmp_conn, aSockName );

  if( ec != TTMP_ERR_NONE ) {
    ReportError( "Error connecting connection", ec );
    TTMPConnectionDestroy(ttmp_conn);
    ttmp_conn = NULL;
    return -1;
  }

 return 0;
}

int AudioAcceptInit(
      int         BUFLEN,
      int         aBufferCount,
      const char* memname,
      const char* sockname,
      int         output,
      int         aSamplingRate,
      int         aChannelsNumber,
      int         aBitsPerSample,
      int         extraOptions )
{ 
  int i;
  TTMPMessage ttmp_msg;
  char errStr[2048];
  int ec;
  TUint8 code;
  TUint32 options;

  assert( BUFLEN );
  assert( aBufferCount );
  assert( aBufferCount <= MAX_NUMBUFS );
  assert( NULL != memname );
  assert( NULL != sockname );
  assert( aSamplingRate );
  assert( aChannelsNumber );
  assert( 8 == aBitsPerSample || 16 == aBitsPerSample );

  DBG( "SR %d, Ch %d, BPS %d\n", aSamplingRate, aChannelsNumber, aBitsPerSample );

  strcpy( sock_name, sockname );

 	// for output plugins, set output flag
        // for input plugins, set wakeup flag
  options = output ? ( SAH_OUTPUT | SAH_NO_BUF_RESET /*| SAH_WAKEUP */): SAH_WAKEUP;

  options |= extraOptions;

  current_buffer = 0;
  buf_count      = aBufferCount;

  strcpy( file_name, memname );

	// create shared memory handle & mmap the region
  fd=open(memname,O_RDWR | O_CREAT | O_TRUNC,S_IRWXU);
  if(fd<0) {
    fprintf(stderr,"open failed: %s\n",strerror(errno));
    return -1;
  }

  buflen = buf_count * ( sizeof( TUint32 ) + BUFLEN );

  if(ftruncate(fd, buflen )<0) {
    fprintf(stderr,"error in truncate setting shared memory size\n");
    close(fd);
    return -1;
  }

  ptr = mmap(NULL, buflen, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);

  if(ptr==NULL) {
    fprintf(stderr,"error in mmap\n");
    close(fd);
    return -1;
  }  

  memset( ptr, 0, buflen );

  DBG( "Allocated %d buffers %d each, total len %d\n", buf_count, BUFLEN, buflen );

	// create semaphore
  if(!output) {
    if((sem=sysv_create_sem(memname,0))<0) {
      fprintf(stderr,"semaphore creation on %s failed\n",memname);
      return -1;
    }
  }

  // Initialising buffers
  for( i = 0; i < buf_count; i++ ) 
  {
      len[ i ]     = ptr + ( sizeof( TUint32 ) * i );
      bufs[ i ]    = ptr + ( sizeof( TUint32 ) * buf_count ) + ( i * BUFLEN );
      offsets[ i ] = 0;
  }

  // open TTMP connection
  i = 5;
  while( TTMP_ERR_NONE != ( ec = AudioAcceptReconnect( sockname ) ) && i-- )
  {   
   ReportError( errStr, ec );
   DBG( "Connect error, retrying, try %d...\n", i );
   usleep( 1000000 );
  }
  
  if( TTMP_ERR_NONE != ec )
  {
   return -1; // FATAL
  }
  

  assert( NULL != ttmp_conn );

	// send stream initialisation
	//  create message
  ec=TTMPMessageCreate(&ttmp_msg);
  if(ec!=TTMP_ERR_NONE) {
    ReportError( "Error creating TTMP message", ec );
    return -1;
  }

  
	//  set fields
  ec=TTMPSetFieldUint8(ttmp_msg,AA_SOUND_CODE_FIELD_NAME,1);
  if(ec!=TTMP_ERR_NONE) {
    sprintf( errStr, "Error setting field name %s", AA_SOUND_CODE_FIELD_NAME );
    ReportError( errStr, ec );    
    return -1;
  }

  ec=TTMPSetFieldStr(ttmp_msg,"shname",memname);
  if(ec!=TTMP_ERR_NONE) {
    ReportError( "Error setting field name shname", ec );    
    return -1;
  }

  ec=TTMPSetFieldUint32(ttmp_msg, AA_SOUND_RATE_FIELD_NAME, aSamplingRate );
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n", AA_SOUND_RATE_FIELD_NAME, errStr );
    return -1;
  }

  ec=TTMPSetFieldUint8(ttmp_msg, AA_SOUND_BPS_FIELD_NAME, aBitsPerSample );
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n", AA_SOUND_BPS_FIELD_NAME, errStr );
    return -1;
  }

  ec=TTMPSetFieldUint8(ttmp_msg, AA_SOUND_CHANNELS_FIELD_NAME, aChannelsNumber );
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n", AA_SOUND_CHANNELS_FIELD_NAME, errStr);
    return -1;
  }

  ec=TTMPSetFieldUint8(ttmp_msg,"num_bufs", buf_count );
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name num_bufs: %s\n",errStr);
    return -1;
  }

		// (ask for sysv semaphore wakeup calls)
  ec=TTMPSetFieldUint32(ttmp_msg,"options",options);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name options: %s\n",errStr);
    return -1;
  }

	//  send message
  ec=TTMPConnectionSend(ttmp_conn,ttmp_msg);
  if(ec!=TTMP_ERR_NONE) {
    char *strErr=strerror(errno);
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error sending init message: %s (%s)\n",errStr,strErr);
    return -1;
  }

	//  discard 
  TTMPMessageDestroy(ttmp_msg);

	// wait for stream ready msg
  ec=TTMPConnectionReceive(ttmp_conn,&ttmp_msg);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error receiving message: %s\n",errStr);
    return -1;
  }
  
  ec=TTMPGetFieldUint8(ttmp_msg,AA_SOUND_CODE_FIELD_NAME,&code);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error getting field name %s: %s\n", AA_SOUND_CODE_FIELD_NAME, errStr);
    return -1;
  }
  
  if(code!=0x81) {
    fprintf(stderr,"Error, return packet not 0x81..\n");
    return -1;
  }

	// pre-create messages
	//  sync
  ec=TTMPMessageCreate(&ttmp_sync_msg);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error creating sync_msg: %s\n",errStr);
    return -1;
  }
  ec=TTMPSetFieldUint8(ttmp_sync_msg, AA_SOUND_CODE_FIELD_NAME, 0x6);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n", AA_SOUND_CODE_FIELD_NAME, errStr);
    return -1;
  }

	//  stop
  ec=TTMPMessageCreate(&ttmp_stop_msg);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error creating stop_msg: %s\n",errStr);
    return -1;
  }
  ec=TTMPSetFieldUint8(ttmp_stop_msg, AA_SOUND_CODE_FIELD_NAME, 0x3);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n",AA_SOUND_CODE_FIELD_NAME,errStr);
    return -1;
  }

	//  start
  ec=TTMPMessageCreate(&ttmp_start_msg);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error creating start_msg: %s\n",errStr);
    return -1;
  }
  ec=TTMPSetFieldUint8(ttmp_start_msg,AA_SOUND_CODE_FIELD_NAME,0x2);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n",AA_SOUND_CODE_FIELD_NAME,errStr);
    return -1;
  }

	//  pause
  ec=TTMPMessageCreate(&ttmp_pause_msg);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error creating pause_msg: %s\n",errStr);
    return -1;
  }
  ec=TTMPSetFieldUint8(ttmp_pause_msg,AA_SOUND_CODE_FIELD_NAME,0x4);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n",AA_SOUND_CODE_FIELD_NAME,errStr);
    return -1;
  }

	//  resume
  ec=TTMPMessageCreate(&ttmp_resume_msg);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error creating pause_msg: %s\n",errStr);
    return -1;
  }
  ec=TTMPSetFieldUint8(ttmp_resume_msg,AA_SOUND_CODE_FIELD_NAME,0x5);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n",AA_SOUND_CODE_FIELD_NAME,errStr);
    return -1;
  }

	// send play command
  ec=TTMPMessageCreate(&ttmp_msg);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error creating TTMP message: %s\n",errStr);
    return -1;
  }

  ec=TTMPSetFieldUint8(ttmp_msg,AA_SOUND_CODE_FIELD_NAME,2);
  if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n",AA_SOUND_CODE_FIELD_NAME,errStr);
    return -1;
  }

  ec=TTMPConnectionSend(ttmp_conn,ttmp_msg);
  if(ec!=TTMP_ERR_NONE) {
    char *strErr=strerror(errno);
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error sending play message: %s (%s)\n",errStr,strErr);
    return -1;
  }

  TTMPMessageDestroy(ttmp_msg);

  return 0;
}


static int AudioAcceptSendMessage( TTMPMessage aMsg, ESendMessageType aSendType )
{
 char errStr[ 2048 ];
 int  ec = TTMP_ERR_NONE;

 assert( NULL != aMsg );
 assert( NULL != ttmp_conn );

resend:
 if( EBlocking == aSendType )
 {
  ec = TTMPConnectionSendBlocking( ttmp_conn, aMsg );
 }
 else
 {
  ec = TTMPConnectionSend( ttmp_conn, aMsg );
 }

 switch( ec )
 {
  case TTMP_ERR_CONNEOF:
   DBG( "TTMPConnection lost, reconnecting...\n" );
   if( !AudioAcceptReconnect( sock_name ) )
   {
    DBG( "Reconnected, resending message...\n" );
    goto resend;
   }
   break;

  case TTMP_ERR_NONE:
   break;

  default:
   TTMPErrorString( ec, errStr, 2048 );
   fprintf( stderr, "Error sending message: %s\n", errStr );
  }

 return ( TTMP_ERR_NONE == ec ) ? 0: -1;
}

int AudioAcceptReconfigure( int aSamplingRate,
                            int aChannelsNumber,
                            int aBitsPerSample )
{
 DBG( "SR %d, Ch %d, BpS %d, Sending SYNC command...\n", aSamplingRate, aChannelsNumber, aBitsPerSample );

 int tmpResult = AudioAcceptSync();

 if( TTMP_ERR_NONE == tmpResult )
 {
  DBG( "Waiting for SYNC responce...\n" );
  
  if( !AudioAcceptWaitForSync() )
  {
   char errStr[ 2048 ];
   
   DBG( "Got SYNC responce\n" );

   // sending format command
   TTMPMessage ttmp_msg = NULL;

   int ec = TTMPMessageCreate( &ttmp_msg );
   if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error creating TTMP message: %s\n",errStr);
    return -1;
   }
  
	//  set fields
   ec=TTMPSetFieldUint8(ttmp_msg,AA_SOUND_CODE_FIELD_NAME, AA_FORMAT_COMMAND );
   if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n",AA_SOUND_CODE_FIELD_NAME, errStr);
    TTMPMessageDestroy(ttmp_msg);
    return -1;
   }

   ec=TTMPSetFieldUint32(ttmp_msg, AA_SOUND_RATE_FIELD_NAME, aSamplingRate );
   if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n", AA_SOUND_RATE_FIELD_NAME, errStr );
    TTMPMessageDestroy(ttmp_msg);
    return -1;
   }

   ec=TTMPSetFieldUint8(ttmp_msg, AA_SOUND_BPS_FIELD_NAME, aBitsPerSample );
   if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n", AA_SOUND_BPS_FIELD_NAME, errStr );
    TTMPMessageDestroy(ttmp_msg);
    return -1;
   }

   ec=TTMPSetFieldUint8(ttmp_msg, AA_SOUND_CHANNELS_FIELD_NAME, aChannelsNumber );
   if(ec!=TTMP_ERR_NONE) {
    TTMPErrorString(ec,errStr,2048);
    fprintf(stderr,"error setting field name %s: %s\n", AA_SOUND_CHANNELS_FIELD_NAME, errStr);
    TTMPMessageDestroy(ttmp_msg);
    return -1;
   }

   //  send message
   DBG( "Sending FORMAT command...\n" );

   if( AudioAcceptSendMessage( ttmp_msg, EBlocking ) )
   {
    TTMPMessageDestroy(ttmp_msg);
    return -1;
   }

   //  discard 
   TTMPMessageDestroy(ttmp_msg);

   DBG( "Sending START command...\n" );
   tmpResult = AudioAcceptSendStart();
  }
  else
  {
   DBG( "Can not receive SYNC confirmation!\n" );
   tmpResult = 1;
  }
 }
 else
 {
  DBG( "Error sending SYNC command\n" );
 }

 return tmpResult;
}

	// return buffer pointer
volatile void *AudioAcceptBuffer(void) {
  return bufs[0];
}

TUint32 AudioAcceptGetLen(int buf) {
  return *(len[buf]);
}

void AudioAcceptSetLen(int buf,TUint32 newlen) {
  *(len[buf])=newlen;
}

int AudioAcceptSync(void) {
 if( !AudioAcceptSendMessage( ttmp_sync_msg, EBlocking ) )
 {
  sync_mode = 1;
  return 0;
 }

 return 1; // Error
}

int AudioAcceptSendStart(void) {
  return AudioAcceptSendMessage( ttmp_start_msg, ENonBlocking );
}

int AudioAcceptSendStop(void) {
  int tmpRetries = 40;

  while( sync_mode && tmpRetries-- )
  {
   DBG( "Syncing, can not send STOP yet, retries %d...\n", tmpRetries );
   usleep( 100000 );
  }
  DBG( "STOP command sent, expecting nofitication %d, retires %d\n", AA_STOP_NOTIFY, tmpRetries );

  return AudioAcceptSendMessage( ttmp_stop_msg, ENonBlocking );
}

int AudioAcceptWaitForSync(void)
{
 const int tmpResult = AudioAcceptWaitForAck( AA_SYNC_NOTIFY );

 sync_mode = 0;

 return tmpResult;
}

int AudioAcceptWaitForAck( int aAckCode )
{
  char errStr[2048];
  TTMPMessage msg  = NULL;
  TUint8      code = 0;
  int ec;
  int retry = 5;

  DBG( "Code %02x\n", aAckCode );

  while(code != aAckCode && retry--) {
    ec = TTMPConnectionReceiveTimeout( ttmp_conn, &msg, 10000000 );

    switch( ec )
    {
     case TTMP_ERR_CONNEOF:
      DBG( "TTMPConnection lost, reconnecting...\n" );
      if( !AudioAcceptReconnect( sock_name ) )
      {
       DBG( "Reconnected, listening again...\n" );
       TTMPMessageDestroy( msg );
       continue; // retry
      }
      {
       TTMPMessageDestroy( msg );
       return -1;
      }
      break;

     case TTMP_ERR_NONE:
      break;

     default:
      TTMPErrorString( ec, errStr, 2048 );
      fprintf(stderr,"error reading reply message: %s\n",errStr );
      TTMPMessageDestroy( msg );
      return -1;
    }

    ec=TTMPGetFieldUint8(msg, AA_SOUND_CODE_FIELD_NAME,&code);
    if(ec!=TTMP_ERR_NONE) {
      TTMPErrorString(ec,errStr,2048);
      fprintf(stderr,"error getting field %s: %s\n",AA_SOUND_CODE_FIELD_NAME,errStr);
      TTMPMessageDestroy(msg);
      return -1;
    }
    TTMPMessageDestroy(msg);
  }

  DBG( "Code %02x, Reply %02x\n", aAckCode, code );
  
  return ( code == aAckCode ) ? 0:-1;
}


int AudioAcceptWaitForStop(void) 
{
 return AudioAcceptWaitForAck( AA_STOP_NOTIFY );
}
  

int AudioAcceptSendPause(void) {
 return AudioAcceptSendMessage( ttmp_pause_msg, ENonBlocking );
}

int AudioAcceptSendResume(void) {
  return AudioAcceptSendMessage( ttmp_resume_msg, ENonBlocking );
}

void AudioAcceptDestroy(void) {
  int i;

  TTMPConnectionDestroy(ttmp_conn);
  ttmp_conn = NULL;

  if( ptr!=NULL ) munmap( ptr, buflen );
  ptr    = NULL;
  buflen = buf_count = current_buffer = 0;

  if(fd >= 0) close( fd );
  unlink( file_name );
  fd         = -1;
  *file_name = 0;

  for( i = 0; i < buf_count; ++i )
  {
   bufs[ i ]    = NULL; 
   offsets[ i ] = 0;
   len[ i ]     = NULL;
  }

  TTMPMessageDestroy( ttmp_start_msg );
  TTMPMessageDestroy( ttmp_stop_msg );
  TTMPMessageDestroy( ttmp_pause_msg );
  TTMPMessageDestroy( ttmp_resume_msg );
  TTMPMessageDestroy( ttmp_sync_msg );
  ttmp_sync_msg = ttmp_stop_msg = ttmp_start_msg = ttmp_pause_msg = ttmp_resume_msg = NULL;
}

int AudioAcceptWait(void) {
  return sysv_lock_sem(sem);
}

int AudioAcceptNotify(void) {
  return sysv_unlock_sem(sem);
}

int AudioAcceptNotifyConditional(void) {
  return sysv_trylock_sem(sem);
}

void AudioAcceptDiscardBuffers( void )
{ 
 while( AudioAcceptGetLen( current_buffer ) )
 {
  DBG( "Discarding buffers, cur buffer %d, len %d...\n", current_buffer, AudioAcceptGetLen( current_buffer ) );
  
  AudioAcceptSetLen( current_buffer, 0 );
  offsets[ current_buffer ] = 0;
  current_buffer = ( current_buffer + 1 ) % buf_count;  
 }

 DBG( "Buffers discarded, cur buffer %d, len %d...\n", current_buffer, AudioAcceptGetLen( current_buffer ) );
}

int AudioAcceptReadData( void** aptr, int alen )
{
 assert( NULL != ptr );
 assert( len );
 assert( NULL != bufs[ 0 ] );
 assert( NULL != len[ 0 ] );
 assert( buf_count );
 
 *aptr = NULL;

/*#ifdef DEBUG
 if( !AudioAcceptGetLen( current_buffer ) )
 {
  DBG( "Start: Buffer %d len %d !!!!\n", current_buffer, AudioAcceptGetLen( current_buffer ) );
 }
#endif
*/
 /*
 DBG( "Read %d bytes, cur buf %d, cur offset %d, cu size %d\n", 
      alen, 
      current_buffer, 
      offsets[ current_buffer ], 
      AudioAcceptGetLen( current_buffer ) );
*/

 // Reading from current buffer
 int tmpResult = MIN( alen, ( int )AudioAcceptGetLen( current_buffer ) );
 
 if( !tmpResult ) // current buffer empty
 { 
  return -1;
 }

 *aptr = ( void * )( bufs[ current_buffer ] + offsets[ current_buffer ] );

 offsets[ current_buffer ] += tmpResult;

 int len = AudioAcceptGetLen( current_buffer ) - tmpResult;
 AudioAcceptSetLen( current_buffer, len );

 if( len == 0 )
 { // Switching to next buffer
  offsets[ current_buffer ] = 0;

  current_buffer = ( current_buffer + 1 ) % buf_count;  

  /*DBG( "Switch to buffer %d, len %d\n", 
       current_buffer, 
       AudioAcceptGetLen( current_buffer ) ); */
 }

/*
#ifdef DEBUG
 if( !AudioAcceptGetLen( current_buffer ) )
 {
  DBG( "End: Buffer %d len %d !!!!\n", current_buffer, AudioAcceptGetLen( current_buffer ) );
 }
#endif
*/

 return tmpResult;
}
/*int AudioAcceptAreBuffersEmpty( void )
{
 int i, tmpRes = 0;

 for( i = 0; i < NUMBUFS; ++i )
 {
  if( !AudioAcceptGetLen( i ) )
  {
   tmpRes++;
  }
 }

 return ( tmpRes == NUMBUFS ) ? 1: 0;
}
*/

static void SigtermHandler( int aSigNum )
{
 DBG( "Received SIGTERM\n" );

 if( NULL != ui_controller )
 {
  DBG( "Calling stop of UIController\n" );

  ui_controller->iStopPlayback(); 
 }
 else
 { 
  DBG( "No UIController set, just exiting\n" );

  exit( 0 );
 }
}

void InitUIOutputController( ui_output_contoller_t* aController )
{ 
  /*int ec=TTMPMessageCreate( &ttmp_ui_msg );
  if(ec!=TTMP_ERR_NONE) {
    fprintf(stderr,"FATAL: can not create TTMP message\n");
    exit( 0 );
  }*/

 ui_controller = aController;

 if( SIG_ERR != signal( SIGTERM, SigtermHandler ) )
 {
  DBG( "Installed SIGTERM handler\n" );
 }
 else
 {
  DBG( "Error %d installing SIGTERM handler, %s\n", errno, strerror( errno ) );
 }
}

void DoneUIOutputController( ui_output_contoller_t* aController )
{
 if( NULL != ttmp_ui_msg )   
 {
  TTMPMessageDestroy( ttmp_ui_msg );
  ttmp_ui_msg = NULL;
 }

 signal( SIGTERM, NULL );
 ui_controller = NULL;
}

void DispatchNextUIOutputMessage( ui_output_contoller_t* aController )
{
 int   ec;

 assert( NULL != aController );
 assert( NULL != ttmp_conn );

 // Deleting previous message
 if( NULL != ttmp_ui_msg )   
 {
  TTMPMessageDestroy( ttmp_ui_msg );
  ttmp_ui_msg = NULL;
 }

 ec = TTMPConnectionReceiveTimeout( ttmp_conn, &ttmp_ui_msg, TTMP_TO_NONE );

 if( ec == TTMP_ERR_NONE )
 {
  TUint8 code;
  TUint8 int8Arg, tmpMode;
  char*  strPtr;
  TChr   tmpBTAddr[ 2048 ], tmpPin[ 100 ];
  int    intArg;

#ifdef DEBUG
  TTMPGetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_CODE_FIELD_NAME, &code );
  DBG( "A2DP command:\n" );
  TTMPMessageDump( ttmp_ui_msg, stderr);
#endif

  ec=TTMPGetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_CODE_FIELD_NAME, &code);
  if(ec==TTMP_ERR_NONE) {
    switch( code )
    {
     case EStartPlayback:
      DBG( "Start playback\n" );
      aController->iStartPlayback();
      break;

     case EPausePlayback:
      DBG( "Pause playback\n" );

      int8Arg = 0;

      TTMPGetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_DISCARD_FIELD_NAME, &int8Arg );

      aController->iPausePlayback( int8Arg );      
      break;

     case EResumePlayback:
      DBG( "Resume playback\n" );
      aController->iResumePlayback();      
      break;

     case EStopPlayback:
      DBG( "Stop playback\n" );
      aController->iStopPlayback();
      break;
    
     case ESyncPlayback:
      DBG( "Sync playback\n" );
      aController->iSyncPlayback();
      break;

     case EBTParams:
      ec=TTMPGetFieldStr( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_BTADDR_FIELD_NAME, tmpBTAddr, 2048 );
      if(ec!=TTMP_ERR_NONE) {
        DBG( "error %i getting btaddr from command_cmd\n",ec );
        return;
      }

      ec=TTMPGetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_BTQUALITY_FIELD_NAME, &int8Arg );
      if(ec!=TTMP_ERR_NONE) {
        DBG( "error %i getting quality from command_cmd\n",ec);
        return;
      }

      ec=TTMPGetFieldStr( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_PIN_FIELD_NAME, tmpPin, 100 );
      if(ec!=TTMP_ERR_NONE) 
      {
        DBG( "error %i getting PIN from command_cmd\n",ec);
        return;
      }


      ec=TTMPGetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_MODE_FIELD_NAME, &tmpMode );
      if(ec!=TTMP_ERR_NONE) {
        DBG( "error %i getting %s command_cmd, assuming %d\n",ec, MEDIAPLAYER_MESSAGE_MODE_FIELD_NAME, ENormalA2DPMode );
        tmpMode = ENormalA2DPMode;
      }

      DBG( "Set output to bluetooth, addr %s, PIN %s, bitpool %d\n", tmpBTAddr, tmpPin, int8Arg );
      aController->iSetBlueToothOutput( tmpBTAddr, int8Arg, tmpPin, tmpMode );
      break;

     case ELastCommandResult:
      DBG( "Get result\n" );

      int8Arg = aController->iLastResult( &strPtr, &intArg );

      TTMPMessageClear( ttmp_ui_msg );
      
      TTMPSetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_CODE_FIELD_NAME, EResultNotify );
      TTMPSetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_RESULT_FIELD_NAME, int8Arg );
      TTMPSetFieldStr( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_ERRORTEXT_FIELD_NAME, ( NULL != strPtr ) ? strPtr: "" );
      TTMPSetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_ERROR_CODE_FIELD_NAME, iLastA2DPErrorCode );
      TTMPSetFieldUint8( ttmp_ui_msg, MEDIAPLAYER_MESSAGE_STATUS_FIELD_NAME, intArg );

      TTMPConnectionSend( ttmp_conn, ttmp_ui_msg );

      iLastA2DPErrorCode = A2DP_ERROR_UNKNOWN;
      break;

     default:
      DBG( "Unknown command %d\n", code );
      return;
    }
  } else {
    DBG( "error getting code field from TTMP message\n");
  }
  //TTMPMessageDestroy(msg);
 } else if( ec == TTMP_ERR_CONNEOF ) {
   // ohoh, connection gone

   usleep( 1000000 );
   DBG( "Connection gone, reconnecting !!!\n" );   
   if( !AudioAcceptReconnect( sock_name ) )
   { 
    tmpReconnectCount = RECONNECT_COUNT;
   }
   else
    if( !tmpReconnectCount-- )
    {
     DBG( "Reconnect limit reached, exiting...\n" );

     if( NULL != ui_controller )
     {
      DBG( "Calling Stop of UI controller\n" );
      ui_controller->iStopPlayback() ;
     }
     else
     {
      DBG( "No UI controller set, exiting brutally\n" );
      exit( 1 );
     }
    }
 } else if( ec != TTMP_ERR_TIMEOUT ) {
   // get error
   DBG( "Receive error %i, errno: %d, %s\n", ec, errno, strerror( errno ) );   
 }
}
