/*
** File  : AudioAccept.h
** Author: Koen Martens <koen.martens@tomtom.com>
** Desc. : Abstraction layer for sending sound to the
**	   audio_accept interface
**
*/

#ifndef __AUDIO_ACCEPT_H__
#define __AUDIO_ACCEPT_H__

#include <basetypes.h>

#define SAH_NOTIFY       1
#define SAH_WAKEUP       2
#define SAH_OUTPUT       4
#define SAH_NO_BUF_RESET 8

#ifdef __cplusplus
extern "C" {
#endif

#define DEFAULT_AUDIO_SOCKET_NAME "/var/run/sound_accept"
    
// Command codes
typedef enum { 
 EInitialiseStream  = 0x01,
 EStartPlayback     = 0x02, 
 EStopPlayback      = 0x03,
 EPausePlayback     = 0x04,
 EResumePlayback    = 0x05,
 ESyncPlayback      = 0x06,
 EBTParams          = 0x07,
 ELastCommandResult = 0x08
} ui_command_code_t;

// Feedback codes
typedef enum 
{
  ENoSlotsNotify      = 0x80,
  EConnectionOKNotify = 0x81,
  ESyncNotify         = 0x82,
  ESilenceNotify      = 0x83,
  EStopNotify         = 0x84,
  EResultNotify       = 0x85
} ui_feedback_code_t;

// A2dp modes
typedef enum 
{
 ENormalA2DPMode                 = 0x00,
 EAlwaysSendEmptyPacketsA2DPMode = 0x01
} a2dp_mode_t;

// TTMP message field names
#define MEDIAPLAYER_MESSAGE_CODE_FIELD_NAME        "code"
#define MEDIAPLAYER_MESSAGE_BTADDR_FIELD_NAME      "btaddr"
#define MEDIAPLAYER_MESSAGE_BTQUALITY_FIELD_NAME   "quality"
#define MEDIAPLAYER_MESSAGE_ERRORTEXT_FIELD_NAME   "errortext"
#define MEDIAPLAYER_MESSAGE_RESULT_FIELD_NAME      "result"
#define MEDIAPLAYER_MESSAGE_PIN_FIELD_NAME         "pin"
#define MEDIAPLAYER_MESSAGE_DISCARD_FIELD_NAME     "discard"
#define MEDIAPLAYER_MESSAGE_ERROR_CODE_FIELD_NAME  "errorcode"
#define MEDIAPLAYER_MESSAGE_MODE_FIELD_NAME        "mode"
#define MEDIAPLAYER_MESSAGE_STATUS_FIELD_NAME      "status"

// soundhack options
#define SAH_NOTIFY        1
#define SAH_WAKEUP        2
#define SAH_OUTPUT        4
#define SAH_NO_BUF_RESET  8
#define SAH_NOTIFY_STOP   16
#define SAH_NOTIFY_START  32
#define SAH_NOTIFY_PAUSE  64
#define SAH_NOTIFY_RESUME 128
#define SAH_NOTIFY_SYNC   256
#define SAH_NOTIFY_SILENCE 512

// AA Command fields
#define AA_SOUND_CODE_FIELD_NAME     "code"
#define AA_SOUND_RATE_FIELD_NAME     "rate"
#define AA_SOUND_BPS_FIELD_NAME      "bps"
#define AA_SOUND_CHANNELS_FIELD_NAME "channels"


// Soundhack commands
#define AA_FORMAT_COMMAND 0x09

#define AA_SYNC_NOTIFY 0x82
#define AA_STOP_NOTIFY 0x84

// Output UI control
typedef struct 
{
 // Setup control
 int ( *iSetBlueToothOutput )( const char* aAddress, int aQuality, const char* aPin, int aMode );

 // Playback control 
 int ( *iStartPlayback )( void );
 int ( *iPausePlayback )( int aDiscardBuffers );
 int ( *iResumePlayback )( void );
 int ( *iStopPlayback )( void );
 int ( *iSyncPlayback )( void );

 // Last result
 int ( *iLastResult )( char** aErrorText, int* aStatus );

} ui_output_contoller_t;   
 

void InitUIOutputController( ui_output_contoller_t* aController );
void DoneUIOutputController( ui_output_contoller_t* aController );
void DispatchNextUIOutputMessage( ui_output_contoller_t* aController );

// Audio stuff

int AudioAcceptInit( 
     int         BUFLEN, 
     int         aBufferCount,
     const char* memname, 
     const char* sockname, 
     int         output,
     int         aSamplingRate,
     int         aChannelsNumber,
     int         aBitsPerSample,
     int	     extraOptions );


volatile void *AudioAcceptBuffer(void);

TUint32 AudioAcceptGetLen(int buf);

void AudioAcceptSetLen(int buf,TUint32 newlen);

void AudioAcceptDestroy(void);

int AudioAcceptWait(void);

int AudioAcceptNotify(void);

int AudioAcceptNotifyConditional(void);
    
int AudioAcceptSync(void);

int AudioAcceptSendStart(void);

int AudioAcceptSendStop(void);

int AudioAcceptSendResume( void );

int AudioAcceptSendPause( void );

int AudioAcceptWaitForStop(void);
int AudioAcceptWaitForSync(void);
int AudioAcceptWaitForAck( int aAckCode );

void AudioAcceptDiscardBuffers( void );

int AudioAcceptReconfigure( int aSamplingRate,
                            int aChannelsNumber,
                            int aBitsPerSample );


// Checks if all audio buffers are empty
// Returns -1 if all buffers empty or number of read bytes
int AudioAcceptReadData( void** aptr, int alen );

#ifdef __cplusplus
}
#endif

#endif // __AUDIO_ACCEPT_H__
