/* 
 * File   : Codesc.c
 * Author : Serhiy Tkachenko <Serhiy.Tkachenko@tomtom.com>
 * Desc.  : C++ like abstraction for MP3 decoder and SBC encoders with or
 *          without threading
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <memory.h>
#include <unistd.h>

#include "Codecs.h"
#include "MP3DecoderAbstraction.h"

static TMP3Decoder* iMP3Decoder = NULL;

enum
{
 KNoThreadingMP3StreamBufferHandle = 1,
 KNoThreadingSBCStreamBufferHandle = 2
};

StreamBufferHandle* InitMP3Decoder( int             aInputFile, 
                                    int             aOptions, 
                                    int             aBufferSize, 
                                    EThreadingModel aThreading )
{
 MadOpt opt;

 assert( NULL == iMP3Decoder );

 opt.fd      = aInputFile;
 opt.options = aOptions;

 StreamBufferHandle* tmpResult = NULL; 
 
 if( KUseThreading == aThreading )
 {
  fprintf( stderr, "Creating THREADING MP3 decoder %d %d %d %d...\n", aInputFile, aOptions, aBufferSize, aThreading );

  tmpResult = InitStreamBuffer( aBufferSize, 
                                MadDecoder,
                                &opt,
                                BO_IBLK ); // Fill buffer with samples
 }
 else
  if( KDoNotUseThreading == aThreading )
  {
   fprintf( stderr, "Creating NON-THREADING MP3 decoder...\n" );
   
   if( NULL != ( iMP3Decoder = CreateMP3Decoder( KMADMP3 ) ) )
   {
    if( !iMP3Decoder->iMP3Init( aInputFile, aBufferSize, aOptions ) )
    {
     tmpResult = ( StreamBufferHandle* )KNoThreadingMP3StreamBufferHandle;
    }
    else
    {
     iMP3Decoder->iMP3Finish();
     iMP3Decoder = NULL;
    }
   }
  }
  else
  {
   assert( 0 ); // Bad parameter
  }

 fprintf( stderr, "MP3 decoder creation done, handle %p...\n", tmpResult );

 return tmpResult;
}

void DoneMP3Decoder( StreamBufferHandle* aDecoder )
{
 assert( NULL != aDecoder );

 if( ( StreamBufferHandle* )KNoThreadingMP3StreamBufferHandle == aDecoder )
 {
  assert( NULL != iMP3Decoder );
  DestroyMP3Decoder();
  iMP3Decoder = NULL;
 }
 else
 {
  DoneStreamBuffer( aDecoder );
 }
}

typedef struct 
{
 StreamBufferHandle*  iPCMStreamBuffer;
 TSBCEncoder*         iSBCEncoder;
 sbc_t*               iSBC;
 StreamBufferHandle*  iSBCStreamBuffer;
} TSBCOptions;
 
static TSBCOptions tmpOptions = { NULL, NULL, NULL, NULL };

static void* SBCEncoderThread( void* arg );
static void* GetNextSBCFrameThread();
static void* GetNextSBCFrameNoThread();

TSBCGetNextFrameFunction InitSBCEncoder( 
                            StreamBufferHandle* aPCMBuffer, 
                            ESBCEncoderType     aType, 
                            EThreadingModel     aThreading, 
                            int                 aBufferSize, 
                            sbc_t*              aSBC, 
                            int                 aBitPool )
{
 TSBCGetNextFrameFunction tmpResult = NULL;

 assert( NULL != aPCMBuffer ); 
 assert( NULL != aSBC );
 assert( aBitPool >= 32 );
 assert( NULL == tmpOptions.iPCMStreamBuffer );

 fprintf( stderr, "Creating SBC encoder %d...\n", aType );     

 tmpOptions.iSBCEncoder = CreateSBCEncoder( aType );
 assert( NULL != tmpOptions.iSBCEncoder );

 if( tmpOptions.iSBCEncoder->iSBCInit( aSBC, SBC_NULL, aBitPool ) == -EIO ) //SBC initialization failure
  return NULL;	  

 tmpOptions.iPCMStreamBuffer = aPCMBuffer;
 tmpOptions.iSBC             = aSBC;

 if( KUseThreading == aThreading )
 {
  fprintf( stderr, "Creating THREADING SBC encoder...\n" );     

  tmpOptions.iSBCStreamBuffer = InitStreamBuffer( KDefaultSBCBufferSize, 
                                 SBCEncoderThread,
                                 &tmpOptions,
                                 0 ); 
  if( NULL != tmpOptions.iSBCStreamBuffer )
  {
   tmpResult = GetNextSBCFrameThread;
  }
 }
 else
  if( KDoNotUseThreading == aThreading )
  {
   fprintf( stderr, "Creating NON-THREADING SBC encoder...\n" );
   
   assert( NULL != iMP3Decoder );
  
   tmpOptions.iSBCStreamBuffer = ( StreamBufferHandle* )KNoThreadingSBCStreamBufferHandle;
  
   tmpResult = GetNextSBCFrameNoThread;
  }
  else
  {
   assert( 0 ); // Error
  } 
 
 fprintf( stderr, "SBC encoder creation done, handle %p...\n", tmpResult );
 
 return tmpResult;
}

static void* GetNextSBCFrameNoThread()
{
 assert( ( StreamBufferHandle* )KNoThreadingMP3StreamBufferHandle == tmpOptions.iPCMStreamBuffer );
 assert( NULL != iMP3Decoder );
 
 //fprintf( stderr, "Get next SBC Frame...\n" );

 void* tmpPCMData = iMP3Decoder->iMP3Decode( tmpOptions.iSBC->input_len );
 void* tmpResult  = NULL; 

 if( NULL != tmpPCMData )
 {
  //fprintf( stderr, "Encoding to SBC..." );
  if( tmpOptions.iSBC->input_len == tmpOptions.iSBCEncoder->iSBCEncode( tmpOptions.iSBC, tmpPCMData, tmpOptions.iSBC->size ) )
  { 
   tmpResult = tmpOptions.iSBC->data; 
   //fprintf( stderr, "success\n" );
  }
 }

 return tmpResult;
}

static void* GetNextSBCFrameThread()
{
 int   tmpReadCount = 0;
 void* tmpDummyPtr;
 void* tmpResult;

 assert( NULL != tmpOptions.iSBCStreamBuffer );

 
 while( !tmpReadCount )
 {
  // Reading header
  tmpReadCount = ReadStreamBuffer( tmpOptions.iSBCStreamBuffer, &tmpResult, sizeof( struct sbc_frame_hdr ) );
  if( tmpReadCount < 0 )
  {
   return NULL;
  }
  else
   if( !tmpReadCount )
   {
    usleep( 1000 );
    continue;
   }

  // Reading body
  tmpReadCount = ReadStreamBuffer( tmpOptions.iSBCStreamBuffer, &tmpDummyPtr, tmpOptions.iSBC->len - tmpReadCount );
  if( tmpReadCount < 0 )
  {
   return NULL;
  }

  assert( tmpReadCount );  // Can not happen
 }

 return tmpResult;
}

void DoneSBCEncoder()
{
 assert( NULL != tmpOptions.iSBCStreamBuffer );

 if( ( StreamBufferHandle* )KNoThreadingSBCStreamBufferHandle == tmpOptions.iSBCStreamBuffer )
 {
  DoneStreamBuffer( tmpOptions.iSBCStreamBuffer );
 }

 DestroySBCEncoder();

 memset( &tmpOptions, 0, sizeof( tmpOptions ) );
}

static void* SBCEncoderThread( void* arg )
{
  volatile const unsigned char*      tmpOutputBufferEnd = NULL;
  volatile       unsigned char*      tmpOutputPtr       = NULL;
                 StreamBufferHandle* tmpBuffer          =( StreamBufferHandle* )arg;
                 TSBCOptions*        tmpOptions         = ( TSBCOptions* )GetStreamBufferUdata( tmpBuffer );
                 void*               tmpPCMPtr          = NULL;
                 int                 tmpBufLen          = 0;
                 
  assert( NULL != tmpBuffer );
  assert( NULL != tmpOptions );
  assert( NULL != tmpOptions->iPCMStreamBuffer );
  assert( NULL != tmpOptions->iSBC );
  assert( NULL != tmpOptions->iSBCEncoder );
  
    // set up output buffer
  tmpOutputPtr       = GetStreamBufferPtr( tmpBuffer );
  tmpOutputBufferEnd = GetStreamBufferEnd( tmpBuffer );
  assert( NULL != tmpOutputPtr );
  assert( NULL != tmpOutputBufferEnd );

  do 
  {
   // Reading PCM data from buffer
   int tmpLen = 0;
   
   while( !tmpLen )
   {
    tmpLen = ReadStreamBuffer( tmpOptions->iPCMStreamBuffer, &tmpPCMPtr, tmpOptions->iSBC->input_len );

    tmpOptions->iSBC->data = ( void* )tmpOutputPtr;

    if( tmpLen == tmpOptions->iSBC->input_len )
    {
     tmpOptions->iSBCEncoder->iSBCEncode( tmpOptions->iSBC, tmpPCMPtr, tmpOptions->iSBC->size );
    }
    else
     if( tmpLen < 0 ) // END
     {
      goto end;
     }
     else
      if( !tmpLen )
      {
       usleep( 1000 );       
      }
      else
      {
       assert( 0 ); // Should not happen
      }
   };

    //memcpy( tmpOutputPtr, tmpPCMPtr, tmpOptions->iSBC->len );
    assert( tmpOptions->iSBC->len );
    tmpOutputPtr += tmpOptions->iSBC->len;
    tmpBufLen    += tmpOptions->iSBC->len;
    
    /* Flush the output buffer if it is full. */
    if( tmpOutputBufferEnd < tmpOutputPtr + tmpOptions->iSBC->len ) 
    {
     if( FlushStreamBuffer( tmpBuffer, tmpBufLen ) ) 
     {
      break;
     }
	
     tmpBufLen          = 0;
     tmpOutputPtr       = GetStreamBufferPtr( tmpBuffer );
     tmpOutputBufferEnd = GetStreamBufferEnd( tmpBuffer );  
    }
  } 
  while( !StreamBufferKillThread( tmpBuffer ) );

end:
	// we're done, set up last buffer, including the eos marker
  StreamBufferEos( tmpBuffer );
  FlushStreamBuffer( tmpBuffer, tmpBufLen );

  pthread_exit( 0 );  

  return 0;
}


