/* 
 * File   : Codesc.h
 * Author : Serhiy Tkachenko <Serhiy.Tkachenko@tomtom.com>
 * Desc.  : C++ like abstraction for MP3 decoder and SBC encoders with or
 *          without threading
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#ifndef __CODECS__H__
#define __CODECS__H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
    
#include "StreamBuffer.h"
#include "SBCEncoderAbstraction.h"
#include "MadProducer.h"
    
typedef enum 
{ 
 KDefaultMP3BufferSize = 8192,
 KDefaultSBCBufferSize = 1024
} ECodecBufferSizes;

typedef enum
{
 KUseThreading      = 0,
 KDoNotUseThreading = 1
} EThreadingModel;

/** \brief This function returns the pointer to next encoded SBC frame
           Frame parameters (length, duration, etc.) are stored in 
           sbc_t structure used in \a InitSBCEncoder()

  If return result is NULL, this means end of stream
*/
typedef 
 void*  (* TSBCGetNextFrameFunction )();

/** \Brief Initializes MP3 decoder 

  \param aInputFile  Open file descriptor of an MP3 file
  \param aOptions    Options for MAD decoder, see \a MadProducer.h
  \param aBufferSize Size of MP3 decoder buffer, preferably \a DefaultMP3BufferSize
  \param aThreading  Threading model

  Function returns NULL in case of error
*/
StreamBufferHandle* InitMP3Decoder( int             aInputFile, 
                                    int             aOptions, 
                                    int             aBufferSize, 
                                    EThreadingModel aThreading );
void                DoneMP3Decoder( StreamBufferHandle* aDecoder );

/** \Brief Initializes SBC encoder and returns the pointer to data reading function 

 \param aPCMBuffer  Handle to corresponding PCM StreamBuffer. Input PCM samples
                    have to be 16-bit 44.1kHz LR format
 \param aType       Type of SBC encoder, refer to \a SBCEncoderAbstraction.h
 \param aThreading  Threading model
 \param aSBC        Pointer to sbc_t structure that receives encoder parameters
 \param aBitPool    BitPool to use for SBC encoding. Refer to A2DP spec.
                    Minimum value 32, maximum value 256.
                    However, value of 64 is the maximum BC3 ROM chip can handle

 The only encoder parameter application can control is BitPool.
 All other parameters are fixed in the encoder.
 Function returns NULL in case of error

 After successful the initialization, the application should set the
 sbc->data to a valid buffer pointer, that will be used internally and
 sbc->size to 1024
*/
TSBCGetNextFrameFunction InitSBCEncoder( StreamBufferHandle* aPCMBuffer, 
                                        ESBCEncoderType     aType, 
                                        EThreadingModel     aThreading, 
                                        int                 aBufferSize, 
                                        sbc_t*              aSBC, 
                                        int                 aBitPool );
void                    DoneSBCEncoder();

#ifdef __cplusplus
}
#endif

#endif // __CODECS__H__

