/* 
 * File   : MP3DecoderAbstraction.c
 * Author : Serhiy Tkachenko <Serhiy.Tkachenko@tomtom.com>
 * Desc.  : C++ like abstraction for MP3 dencoder implementation
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdio.h>
#include <memory.h>
#include <assert.h>
#include <malloc.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "MP3DecoderAbstraction.h"

static int            MADMP3Init( int aInputFile, int aBufferSize, unsigned long aFlags );
static unsigned char* MADMP3Decode( int Count );
static void           MADMP3Finish();
 
static TMP3Decoder iMP3Decoder;
static int         iInitialized = 0;

TMP3Decoder* CreateMP3Decoder( EMP3DecoderType aType )
{
 assert( !iInitialized );

 switch( aType )
 {
  case KMADMP3:
   iMP3Decoder.iMP3Decode = MADMP3Decode;
   iMP3Decoder.iMP3Finish = MADMP3Finish;
   iMP3Decoder.iMP3Init   = MADMP3Init;
   break;

  default:
   assert( 0 );
 }

 iInitialized = 1;

 return &iMP3Decoder;
}

void DestroyMP3Decoder()
{
 assert( iInitialized );
 
 iInitialized = 0;
 memset( &iMP3Decoder, 0, sizeof( iMP3Decoder ) );
}

// MAD abstraction
static struct 
{
 struct mad_stream          iStream;
 struct mad_frame           iFrame;
 struct mad_synth           iSynth;
 
        mad_timer_t         iTimer;

        unsigned char       iInputBuffer[ MAD_INTERNAL_INPUT_BUFFER_SIZE + MAD_BUFFER_GUARD ];
        unsigned char*      iGuardPtr;

        unsigned char*      iOutputBuffer;
        unsigned char*      iOutputBufferEnd;
        unsigned char*      iOutputReadPtr; // Pointer to read from buffer
        unsigned char*      iOutputPtr;     // Pointer to fill the buffer in 
        int                 iBufferSize;
        int                 iOccupiedSize;
        int                 iTresholdSize;
        unsigned long       iFrameCount;
        int                 iInputFile;
        int                 iStopped;
        unsigned long       iFlags;
} iMADParameters;

static int  MADDecodeFrame();
static void MADFillBuffer();

static int MADMP3Init( int aInputFile, int aBufferSize, unsigned long aFlags )
{
 assert( aBufferSize );
 assert( !( aBufferSize % 1024 ) );
 assert( iInitialized );

 fprintf( stderr, "Initializing MAD...\n" );

 memset( &iMADParameters, 0, sizeof( iMADParameters ) );

 /* First the structures used by libmad must be initialized. */
 mad_stream_init( &iMADParameters.iStream );
 mad_frame_init( &iMADParameters.iFrame );
 mad_synth_init( &iMADParameters.iSynth );
 mad_timer_reset( &iMADParameters.iTimer );

 // Allocating output buffer
 iMADParameters.iOutputBuffer = malloc( aBufferSize * 2 );
 assert( NULL != iMADParameters.iOutputBuffer );
  
 iMADParameters.iOutputReadPtr   = iMADParameters.iOutputPtr = iMADParameters.iOutputBuffer;
 iMADParameters.iBufferSize      = aBufferSize * 2;
 iMADParameters.iTresholdSize    = aBufferSize;// * 2 / 3;
 iMADParameters.iOutputBufferEnd = iMADParameters.iOutputBuffer + aBufferSize;

 iMADParameters.iInputFile     = aInputFile;
 iMADParameters.iFlags         = aFlags;

 // Buffering data
 fprintf( stderr, "Buffering decoded data...\n" );

 MADFillBuffer();

 iInitialized = 1;

 fprintf( stderr, "MAD init done %d\n", !iMADParameters.iStopped );
 
 return iMADParameters.iStopped;
}

static void MADFillBuffer()
{
 int tmpRes = 1;

 while( iMADParameters.iOccupiedSize < iMADParameters.iTresholdSize && tmpRes )
 {
  tmpRes &= MADDecodeFrame();
 }

 iMADParameters.iStopped = !tmpRes;
}

static unsigned char* MADMP3Decode( int Count )
{
 unsigned char* tmpResult = iMADParameters.iOutputReadPtr;

 assert( Count > 0 );
 assert( iInitialized );
 assert( !( iMADParameters.iBufferSize % Count ) );
 
 //fprintf( stderr, "MAD MP3 decode %d...", Count );
 
 if( iMADParameters.iOccupiedSize >= Count )
 { // Normal case
   iMADParameters.iOccupiedSize  -= Count;  
   iMADParameters.iOutputReadPtr += Count;

   // Wrapping up
   if( iMADParameters.iOutputBufferEnd <= iMADParameters.iOutputReadPtr )
   {
    iMADParameters.iOutputReadPtr = iMADParameters.iOutputBuffer;
   }
 }
 else
 {
  if( !iMADParameters.iStopped )
  {
   MADFillBuffer();
   tmpResult = MADMP3Decode( Count );
  }
  else
   if( iMADParameters.iOccupiedSize )
   { // Giving away the rest
    memset( tmpResult + iMADParameters.iOccupiedSize - Count, 0,  Count - iMADParameters.iOccupiedSize );
    iMADParameters.iOccupiedSize = 0;
   }
   else
   { 
    tmpResult = NULL;
   }
 }

 //fprintf( stderr, "end %p\n", tmpResult );

 return tmpResult;
}

static void MADMP3Finish()
{
 assert( iInitialized );

 mad_synth_finish( &iMADParameters.iSynth );
 mad_frame_finish( &iMADParameters.iFrame );
 mad_stream_finish( &iMADParameters.iStream );

 free( iMADParameters.iOutputBuffer );
 close( iMADParameters.iInputFile );

 memset( &iMADParameters, 0, sizeof( iMADParameters ) );

 iInitialized = 0;
}

#define WriteSample( a )
#if 0
void WriteSample( signed short aSample )  
{
 static FILE* tmpFile = NULL;

 if( NULL == tmpFile )
 {
  tmpFile = fopen( "test.cdr", "w+b" );
 }

 aSample = ( ( aSample >> 8 ) & 0xFF ) | ( ( aSample << 8 ) &0xFF00 );
 
 fwrite( &aSample, 1, 2, tmpFile );
}

#endif

static int MADDecodeFrame()
{
 int i;

 assert( iInitialized );
 
 // Reading data
 if( NULL             == iMADParameters.iStream.buffer || 
     MAD_ERROR_BUFLEN == iMADParameters.iStream.error ) 
 {
   size_t         ReadSize, ReadResult, Remaining;
   unsigned char* ReadStart;

   if( NULL != iMADParameters.iStream.next_frame ) 
   {
    Remaining = iMADParameters.iStream.bufend - iMADParameters.iStream.next_frame;
    memmove( iMADParameters.iInputBuffer, iMADParameters.iStream.next_frame,Remaining );

    ReadStart = iMADParameters.iInputBuffer + Remaining;
    ReadSize  = MAD_INTERNAL_INPUT_BUFFER_SIZE - Remaining;
   } 
   else 
   {
    ReadSize  = MAD_INTERNAL_INPUT_BUFFER_SIZE;
    ReadStart = iMADParameters.iInputBuffer;
    Remaining = 0;
   }
    
   ReadResult = read( iMADParameters.iInputFile, ReadStart, ReadSize );

   if( !ReadResult && iMADParameters.iGuardPtr ) 
   {
    return 0;
   }

   if( ReadResult < ReadSize ) 
   {
    if( ReadStart == iMADParameters.iInputBuffer ) 
    {
	  // no new data, just flush and exit
	  return 0;
	} 
    else 
    {
	  // possibly partial frame, pad and decode
	  fprintf( stderr, "Got stream end, padding..\n");

      iMADParameters.iGuardPtr = ReadStart + ReadResult;
      memset( iMADParameters.iGuardPtr, 0, MAD_BUFFER_GUARD );
      ReadResult += MAD_BUFFER_GUARD;
      iMADParameters.iStopped = 1;
	}
   }
   
   mad_stream_buffer( &iMADParameters.iStream, 
                      iMADParameters.iInputBuffer, 
                      ReadResult + Remaining );
   iMADParameters.iStream.error = 0;
  }

 // Decoding
 if( mad_frame_decode( &iMADParameters.iFrame, &iMADParameters.iStream ) ) 
 {
  if( MAD_RECOVERABLE( iMADParameters.iStream.error ) ) 
  {
   if( MAD_ERROR_LOSTSYNC != iMADParameters.iStream.error || 
       iMADParameters.iStream.this_frame != iMADParameters.iGuardPtr ) 
   {
    fprintf( stderr,"%s: recoverable frame level error (%s)\n", 
             "decoder","x"); //MadErrorString(&Stream));
    fflush(stderr);
   }   
  } 
  else 
   if( MAD_ERROR_BUFLEN != iMADParameters.iStream.error ) 
    {
     fprintf( stderr,"%s: unrecoverable frame level error (%s).\n",
              "decoder","x"); //MadErrorString(&Stream));
      return 0;
     }
 }

 //if(FrameCount==0) PrintFrameInfo(stderr,&Frame.header);

 iMADParameters.iFrameCount++;
    
 mad_timer_add( &iMADParameters.iTimer, iMADParameters.iFrame.header.duration );

 mad_synth_frame( &iMADParameters.iSynth, &iMADParameters.iFrame );

 // copy to output buffer
 unsigned char* tmpPrevPtr = iMADParameters.iOutputPtr;
 signed  short  Sample     = 0;

 //fprintf( stderr, "s %d\n", iMADParameters.iSynth.pcm.length );
 for( i = 0; i < iMADParameters.iSynth.pcm.length; i++ ) 
 {
  /* Left channel */
  PUT_SAMPLE( 0, iMADParameters.iFlags, iMADParameters.iSynth, iMADParameters.iOutputPtr );

  WriteSample( MadFixedToSshort( iMADParameters.iSynth.pcm.samples[ 0 ][ i ] ) );

  /* Right channel. If the decoded stream is monophonic then
   * the right output channel is the same as the left one.
   */
  
  if( MAD_NCHANNELS( &iMADParameters.iFrame.header ) == 2 ) 
   { // stereo
	PUT_SAMPLE( 1, iMADParameters.iFlags, iMADParameters.iSynth, iMADParameters.iOutputPtr );
    WriteSample( MadFixedToSshort( iMADParameters.iSynth.pcm.samples[ 1 ][ i ] ) );
   } 
  else 
  {	// mono
   PUT_SAMPLE( 0, iMADParameters.iFlags, iMADParameters.iSynth, iMADParameters.iOutputPtr );
   WriteSample( MadFixedToSshort( iMADParameters.iSynth.pcm.samples[ 0 ][ i ] ) );
  }

  // Wrapping up, should always be OK
  assert( iMADParameters.iOutputBufferEnd >= iMADParameters.iOutputPtr );

  if( iMADParameters.iOutputBufferEnd == iMADParameters.iOutputPtr )
  {
   //fprintf( stderr, "%d %d\n", iMADParameters.iOccupiedSize, iMADParameters.iBufferSize );
   iMADParameters.iOccupiedSize += ( iMADParameters.iOutputPtr - tmpPrevPtr );
   tmpPrevPtr = iMADParameters.iOutputPtr = iMADParameters.iOutputBuffer;
   //fprintf( stderr, "%d %d\n", iMADParameters.iOccupiedSize, iMADParameters.iBufferSize );
  } 
 }

 //fprintf( stderr, "%d %d\n", iMADParameters.iOccupiedSize, iMADParameters.iBufferSize );
 
 iMADParameters.iOccupiedSize += ( iMADParameters.iOutputPtr - tmpPrevPtr );
 assert( iMADParameters.iOccupiedSize <= iMADParameters.iBufferSize ); 
 assert( iMADParameters.iOccupiedSize >= 0 ); 

 //fprintf( stderr, "%d %d\n", iMADParameters.iOccupiedSize, iMADParameters.iBufferSize );
 
 return !iMADParameters.iStopped;
}

