/* 
 * File   : UIController.c
 * Author : Serhiy Tkachenko <Serhiy.Tkachenko@tomtom.com>
 * Desc.  : UI controller implementeation for TomTom sound player
 *      
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *      
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>

#include "Debug.h"
#include "UIController.h"
#include "UICommand.h"
#include "TTMP.h"

#define RECONNECT_COUNT 50

static TTMPMessage feedback_msg      = NULL;
static TTMPMessage command_msg       = NULL;
static ui_contoller_t* ui_controller = NULL;

static void SigtermHandler( int aSigNum )
{
 DBG( "Received SIGTERM\n" );

 if( NULL != ui_controller )
 {
  DBG( "Calling stop of UIController\n" );

  ui_controller->iStopPlayback( 1 ); // Sudden death
 }
 else
 { 
  DBG( "No UIController set, just exiting\n" );

  exit( 0 );
 }
}

void DoneUIController( ui_contoller_t* aController )
{
 signal( SIGTERM, NULL );

 MediaPlayerDoneMessageServer();

 TTMPMessageDestroy( feedback_msg );

 if( NULL != command_msg )
 {
  TTMPMessageDestroy( command_msg );
 }

 feedback_msg  = NULL;
 command_msg   = NULL;
 ui_controller = NULL;
}

void InitUIController( ui_contoller_t* aController )
{
 int ec;

 assert( NULL != aController );

 if( TTMP_ERR_NONE == MediaPlayerInitMessageServer() )
 {
  ec = TTMPMessageCreate( &feedback_msg );
  if(ec!=TTMP_ERR_NONE) {
    fprintf(stderr,"Fatal error: can not create feedback message\n");
    exit(1);
  }

  /*ec = TTMPMessageCreate( &command_msg );
  if(ec!=TTMP_ERR_NONE) {
    fprintf( stderr,"Fatal error: can not create command message\n" );
    exit(1);
  }*/
  DBG("message server initialized\n");
 }
 else
 {
  fprintf( stderr,"Fatal error: can not initialise message server\n" );
  exit( 1 );
 }

 ui_controller = aController;

 if( SIG_ERR != signal( SIGTERM, SigtermHandler ) )
 {
  DBG( "Installed SIGTERM handler\n" );
 }
 else
 {
  DBG( "Error %d installing SIGTERM handler, %s\n", errno, strerror( errno ) );
 }
}

static int tmpLastResult       = 0;
static int tmpLastCommandError = 0;
static int tmpReconnectCount   = RECONNECT_COUNT;

static void ProcessFeedback( ui_contoller_t* aController, int aCode )
{
 int ec;
 const char *tmpStr;

 ec = TTMPMessageClear( feedback_msg );
 if(ec!=TTMP_ERR_NONE) {
   fprintf(stderr,"FATAL: unable to clear message\n");
   return;
 }

 ec=TTMPSetFieldUint8( feedback_msg, MEDIAPLAYER_MESSAGE_CODE_FIELD_NAME, aCode );
 if(ec!=TTMP_ERR_NONE) {
   fprintf(stderr,"FATAL: unable to set feedback code\n");
   return;
 }

 switch( aCode )
 {
  case EGetPlayPos:
   ec=TTMPSetFieldUint32( feedback_msg, MEDIAPLAYER_MESSAGE_TIMESTAMP_FIELD_NAME,
		aController->iGetCurrentPlaybackPosition());
   if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback timestamp\n");
   tmpLastCommandError = 1;
     return;
   }

    ec=TTMPSetFieldUint32(feedback_msg, MEDIAPLAYER_MESSAGE_PERCENTAGE_FIELD_NAME,
		aController->iGetCurrentPlayPercentage());
    if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback percentage\n");

    tmpLastCommandError = 1;
     return;
   }

    ec=TTMPSetFieldUint32(feedback_msg, MEDIAPLAYER_MESSAGE_OFFSET_FIELD_NAME,
		aController->iGetCurrentByteOffset());
    if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback byte offset\n");

    tmpLastCommandError = 1;
     return;
   }

    ec=TTMPSetFieldUint32(feedback_msg, MEDIAPLAYER_MESSAGE_INDEX_FIELD_NAME,
		aController->iGetCurrentChapter());
    if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback %s\n", MEDIAPLAYER_MESSAGE_INDEX_FIELD_NAME);

    tmpLastCommandError = 1;
     return;
   }

   tmpLastCommandError = 0; 
   break;

  case EGetCurrentSong:
   DBG( "Get current song name...\n" );
   tmpStr=aController->iGetCurrentSongName();
   if(tmpStr==NULL) {
     ec=TTMPSetFieldStr(feedback_msg, MEDIAPLAYER_MESSAGE_CURRENTFILE_FIELD_NAME, "");
   } else {
     ec=TTMPSetFieldStr(feedback_msg,MEDIAPLAYER_MESSAGE_CURRENTFILE_FIELD_NAME, tmpStr );
   }
   if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback currentfile\n");
     tmpLastCommandError = 1;
     return;
   }
     
   DBG( "Get next song name...\n" );
   tmpStr=aController->iGetNextSongName();
   if(tmpStr==NULL) {
     ec=TTMPSetFieldStr(feedback_msg, MEDIAPLAYER_MESSAGE_NEXTFILE_FIELD_NAME, "");
   } else {
     ec=TTMPSetFieldStr(feedback_msg, MEDIAPLAYER_MESSAGE_NEXTFILE_FIELD_NAME, tmpStr);
   }
   if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback nextfile\n");
     tmpLastCommandError = 1;
     return;
   }
   tmpLastCommandError = 0;
   break;

  case EGetStatus:
   DBG( "Get last error text...\n" );
   if( tmpLastCommandError )
   {
    tmpStr = "Incorrect command or command parameters";
   }
   else
   {
    tmpStr=aController->iGetLastError();
   }

   ec=TTMPSetFieldStr(feedback_msg, MEDIAPLAYER_MESSAGE_ERRORTEXT_FIELD_NAME,(tmpStr==NULL)?"":tmpStr);
    if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback errortext\n");
     tmpLastCommandError = 1;
     return;
    }   

   ec=TTMPSetFieldUint8(feedback_msg, MEDIAPLAYER_MESSAGE_RESULT_FIELD_NAME,( (tmpStr==NULL) || (!strlen(tmpStr)) )? 1: 0 );
   if(ec!=TTMP_ERR_NONE) {
     fprintf(stderr,"FATAL: unable to set feedback errortext\n");
     tmpLastCommandError = 1;
     return;
   }
   tmpLastCommandError = 0;
   break;

  default:
   DBG( "Unknown feedback command code 0x%02x\n", aCode );
   aCode               = 0;
   tmpLastCommandError = 1;
   break;
 }

 if( aCode )
 {
#if 0
#ifdef DEBUG
   DBG( "Media Player response:\n" );
   TTMPMessageDump(feedback_msg, stderr);
#endif
#endif
  MediaPlayerSendUIFeedback( feedback_msg );
 }
}

int DispatchNextUIMessage( ui_contoller_t* aController, EErrorBehaviour aExitMode )
{
 //TTMPMessage msg;
 int ec;
 assert( NULL != aController );

 // Deleting previous message
 if( NULL != command_msg )
 {
  TTMPMessageDestroy( command_msg );
  command_msg = NULL;
 }

 tmpLastCommandError = 0;

 ec = MediaPlayerGetUICommand( &command_msg );
 if( ec == TTMP_ERR_NONE )
 {
  TUint8  code;
  TUint8  int8Arg;
  TUint32 int32Arg, tmpByteOffset;
  TChr    strArg[2048], strArg2[ 2048 ], *codecNamePtr = NULL;

#ifdef DEBUG
  TTMPGetFieldUint8( command_msg, MEDIAPLAYER_MESSAGE_CODE_FIELD_NAME, &code );
  if( EFeedback != code )
  {
   DBG( "Media Player command:\n" );
   TTMPMessageDump( command_msg, stderr);
  }
#endif

  ec=TTMPGetFieldUint8( command_msg, MEDIAPLAYER_MESSAGE_CODE_FIELD_NAME, &code);
  if(ec==TTMP_ERR_NONE) {
    switch( code )
    {
     case EStartPlayback:
      ec=TTMPGetFieldStr( command_msg, MEDIAPLAYER_MESSAGE_FILENAME_FIELD_NAME, strArg, 2048);
      if(ec!=TTMP_ERR_NONE) {
        fprintf(stderr,"error %i getting filename from command_cmd\n",ec);
	    //TTMPMessageDestroy( command_msg );
        tmpLastCommandError = 1;
        return tmpLastCommandError;
      }

      if( TTMP_ERR_NONE == TTMPGetFieldStr( command_msg, MEDIAPLAYER_MESSAGE_CODECNAME_FIELD_NAME, strArg2, 2048 ) )
      {
       codecNamePtr = strArg2;
      }
      else
      {
       codecNamePtr = NULL;
      }

      if( TTMP_ERR_NONE != TTMPGetFieldUint32( command_msg, MEDIAPLAYER_MESSAGE_OFFSET_FIELD_NAME, &tmpByteOffset ) )
      {
       tmpByteOffset = 0;
      }

      DBG( "Start playback %s, codec %s, offset %u\n", strArg, codecNamePtr, tmpByteOffset );
      tmpLastResult       = aController->iStartPlayback( strArg, codecNamePtr, tmpByteOffset );
      tmpLastCommandError = 0;
      break;

     case EPausePlayback:
      ec = TTMPGetFieldInt32( command_msg, MEDIAPLAYER_MESSAGE_MODE_FIELD_NAME, &int32Arg );
      if( TTMP_ERR_NONE != ec ) 
      {
       int32Arg = -1;
      }

      DBG( "Pause/resume playback, mode %d\n", int32Arg );
      tmpLastResult       = aController->iPausePlayback( int32Arg );
      tmpLastCommandError = 0;
      break;

     case EStopPlayback:
      ec = TTMPGetFieldInt32( command_msg, MEDIAPLAYER_MESSAGE_MODE_FIELD_NAME, &int32Arg );
      if( TTMP_ERR_NONE != ec ) 
      {
       int32Arg = EStopNormal;
      }
      DBG( "Stop playback, mode %d\n", int32Arg );

      tmpLastResult       = aController->iStopPlayback( int32Arg );
      tmpLastCommandError = 0;
      break;
    
     case ESkip:
      DBG("Skip\n");
      ec=TTMPGetFieldStr( command_msg, MEDIAPLAYER_MESSAGE_FILENAME_FIELD_NAME, strArg, 2048);
      if(ec!=TTMP_ERR_NONE) {
        fprintf(stderr,"error %i getting filename from command_cmd\n",ec);
        tmpLastCommandError = 1;
        return tmpLastCommandError;
      }
      tmpLastResult       = aController->iSkip(strArg);
      tmpLastCommandError = 0;
      break;

     case ESetChapter:
      DBG("Set chapter\n");
      ec=TTMPGetFieldUint8( command_msg, MEDIAPLAYER_MESSAGE_INDEX_FIELD_NAME, &int8Arg);
      if(ec!=TTMP_ERR_NONE) 
      {
       fprintf(stderr,"error %i getting %s field from command_cmd\n",ec, MEDIAPLAYER_MESSAGE_INDEX_FIELD_NAME );
       tmpLastCommandError = 1;
       return tmpLastCommandError;
      }
      
      tmpLastResult       = aController->iGoToChapter( int8Arg, 1 );
      tmpLastCommandError = 0;
      break;

     case ENextChapter:
      DBG("Next chapter\n");

      tmpLastResult       = aController->iGoToChapter( 1, 0 );
      tmpLastCommandError = 0;
      break;

     case EPrevChapter:
      DBG("Prev chapter\n");

      tmpLastResult       = aController->iGoToChapter( -1, 0 );
      tmpLastCommandError = 0;
      break;

     case EFeatureMode:
      DBG("Feature mode\n");
      ec = TTMPGetFieldUint32( command_msg, MEDIAPLAYER_MESSAGE_SPEED_FIELD_NAME, &int32Arg);

      if(ec!=TTMP_ERR_NONE) {
        DBG( "error %i getting %s field from command_cmd, setting speed to 1\n",ec, MEDIAPLAYER_MESSAGE_SPEED_FIELD_NAME );        
        int32Arg = 1;
      }

      ec = TTMPGetFieldUint8( command_msg, MEDIAPLAYER_MESSAGE_MODE_FIELD_NAME, &int8Arg);
      if(ec!=TTMP_ERR_NONE) {
        fprintf(stderr,"error %i getting %s field from command_cmd\n",ec, MEDIAPLAYER_MESSAGE_MODE_FIELD_NAME );        
        tmpLastCommandError = 1;
        return tmpLastCommandError;
      }

      tmpLastResult       = aController->iFeatureMode( int8Arg, int32Arg );
      tmpLastCommandError = 0;
      break; 

     case EFeedback:
      ec=TTMPGetFieldUint8( command_msg, MEDIAPLAYER_MESSAGE_FEEDBACK_FIELD_NAME, &int8Arg);
      if(ec!=TTMP_ERR_NONE) {
        fprintf(stderr,"error %i getting %s from command_cmd\n",ec, MEDIAPLAYER_MESSAGE_FEEDBACK_FIELD_NAME);
        //TTMPMessageDestroy(msg);
        tmpLastCommandError = 1;
        return tmpLastCommandError;
      }
      //DBG( "Feedback code %d\n", int8Arg );
      ProcessFeedback( aController, int8Arg );
      break;

     default:
      fprintf(stderr,"Unknown command %d\n", code );
      //TTMPMessageDestroy( msg );
      tmpLastCommandError = 1;
      return tmpLastCommandError;
    }
  } else {
    fprintf(stderr,"error getting code field from command_msg\n");
    tmpLastCommandError = 1;
  }
  //TTMPMessageDestroy(msg);
 } else if( ec == TTMP_ERR_CONNEOF && EExitIfError != aExitMode ) {
   // ohoh, connection gone, try and accept new one...
   usleep( 1000000 );
   if( TTMP_ERR_NONE != MediaPlayerReinitMessageServer() )
   {
    if( !tmpReconnectCount-- )
    {
     if( NULL != ui_controller )
     {
      DBG( "Reconnect attempts limit reached, exiting...\n" );
      ui_controller->iStopPlayback( 1 ); // Sudden death
     }
    }
   }
   else
   {
    tmpReconnectCount = RECONNECT_COUNT;
   }
 } else if( ec!=TTMP_ERR_TIMEOUT ) {
   // get error
   DBG( "Receive error %i, errno: %d, %s\n", ec, errno, strerror( errno ) );
   tmpLastCommandError = 1;
 }

 return tmpLastCommandError;
}

