/* 
 * File   : MadProducer.c
 * Author : Koen Martens <koen.martens@tomtom.com>
 * Desc.  : mp3 decoder with synchronised output using libmad, threading 
 *          version
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <pthread.h>
#include <assert.h>

#include <mad.h>

#ifdef NMP3DEBUG
#undef DEBUG
#endif

#include "Debug.h"
#include "StreamBuffer.h"
#include "MadProducer.h"
#include "AudioAccept.h"

#define NON_MP3_FRAME_READ_COUNT 4

typedef enum { EFlipOK, EFlipTerminate, EFlipSkip, EFlipDiscard } mad_flip_t;


static inline signed short MadFixedToSshort( mad_fixed_t Fixed )
{
        if(Fixed>=MAD_F_ONE)
                return(SHRT_MAX);
        if(Fixed<=-MAD_F_ONE)
                return(-SHRT_MAX);
        Fixed=Fixed>>(MAD_F_FRACBITS-15);
        return((signed short)Fixed);
}


// returns 1 in case of error/thread termination, 2 in case of skip, 3 in case of discard and wait
static mad_flip_t FlipOutputBuffer( StreamBufferHandle* h, MadOpt* opt, int* buffIdx, int delta )
{
    assert( NULL != h );
    assert( NULL != opt );
    assert( NULL != buffIdx );

    mad_flip_t tmpResult = EFlipOK;

    if( GetStreamBufferLength( h ) > delta )
    {
     //DBG( "Flush buffer, buf size %d, delta %d\n", GetStreamBufferLength( h ), delta );
     
     if(opt->haveAudioAccept) {
		// this is a very dirty hack
	  *buffIdx = GetStreamBufferIndex( h );

	  if( FlushStreamBufferAsync( h, GetStreamBufferLength( h ) - delta ) )
      {
       DBG( "Flip result %d\n", EFlipTerminate );
       return EFlipTerminate;
      }

      //DBG( "MAD: Flush buf %d, res %d\n", buffIdx, fec );
	  

      //DBG( "MAD: Set buf %d, len %d\n", buffIdx, GetStreamBufferLength( h ) );
      AudioAcceptSetLen( *buffIdx, GetStreamBufferLength( h ) - delta );

      if( StreamBufferIsEos( h ) )
      {
       DBG( "End of stream, syncing...\n" );
       AudioAcceptSync();
       AudioAcceptWaitForSync();      
       DBG( "End of stream done\n" );
      }
      else
      {
       h->currBuff = 1 - h->currBuff;	// yuck!

	   *buffIdx = GetStreamBufferIndex( h );

       //DBG( "MAD: New buf %d, len %d\n", buffIdx, AudioAcceptGetLen( buffIdx ) );

	   while( !h->killThread && AudioAcceptGetLen(*buffIdx) && !opt->skip && !opt->discardAndWait) {
	    // TODO: killThread should notify the sysv sem..
	    AudioAcceptWait();
       }
      }
	} else {
	  if( FlushStreamBuffer( h, GetStreamBufferLength(h) - delta ) ) 
      {
       tmpResult = EFlipTerminate;
      }
	}
   }
   else
   {
    DBG( "MAD: Out buffer empty, not flipped !\n" );
   }

 //DBG( "Out buffer flipped, new id %d, delta %d, cur b %d\n", *buffIdx, delta, h->currBuff );

 if( opt->discardAndWait )
 {
  tmpResult = EFlipDiscard;
 }
 else      
  if( h->killThread ) 
   {
    tmpResult = EFlipTerminate;
   }
   else
    if( opt->skip ) 
    {
     tmpResult = EFlipSkip;
    }
    
#ifdef DEBUG
 if( EFlipOK != tmpResult )
 {
  DBG( "Flip result %d\n", tmpResult );
 }
#endif
 
 return tmpResult;
}

static void ResetAudioAccept( StreamBufferHandle* h, int* buffInd )
{
  AudioAcceptSetLen( 0, 0 );
  AudioAcceptSetLen( 1, 0 );

  h->currBuff = 1; 

  *buffInd = GetStreamBufferIndex( h );

  DBG( "AudioAccept reset, buf index %d\n", *buffInd );
}

static void ReconfigureStream( const MadOpt* aOpt, const stream_parameters_t* aData, StreamBufferHandle* h, int* buffInd ) 
{
 DBG( "Reconfiguring...\n" );
 
 if(aOpt->haveAudioAccept) {    
        DBG( "Setting up AudioAccept, size %d, readsize %d, sr %d, ch %d, bps %d, type %d, fb %d...\n", 
            aData->iBufferSize,
            aData->iMinReadSize,
            aData->iSamplingRate,
            aData->iNumChannels,
            aData->iBitsPerSample,
            aData->iBufferType,
            aData->iFeatureBufferSize );

         if( AudioAcceptReconfigure( 
             aData->iSamplingRate,
             aData->iNumChannels,
             aData->iBitsPerSample ) )
        {
         DBG( "Error sending Stream Format command\n" );   
        }
        
        ResetAudioAccept( h, buffInd );
 }
 else
 {
  DBG( "No audio accept, doing nothing\n" );
 }
}

void *MadDecoder(void *arg) {
  struct mad_stream       Stream;
  struct mad_frame        Frame;
  struct mad_synth        Synth;
  mad_timer_t             Timer;
  unsigned char           InputBuffer[MAD_INTERNAL_INPUT_BUFFER_SIZE+MAD_BUFFER_GUARD],
                                                *GuardPtr=NULL;
  volatile const unsigned char    *OutputBufferEnd;
  volatile unsigned char          *OutputPtr;
  void                            *readPtr;
  int                     PCMCount   = 0;
  int                     Status     = 0,
                          i,
			              readRetry  = 0,
			              skip       = 0;
  unsigned long           FrameCount = 0;
  int	                  ReadResult = 0;
  StreamBufferHandle     *h=(StreamBufferHandle *)arg;
  MadOpt                 *opt=(MadOpt*)GetStreamBufferUdata(h);
  int			         buffIdx=-1; // for use in shared mem len setting
  stream_parameters_t    tmpStreamData;
  const stream_parameters_t* tmpCurStreamData = &tmpStreamData;
  
  memcpy( ( stream_parameters_t* )tmpCurStreamData, opt->fileReader->iGetStreamParameters( opt->fileReader ), sizeof( tmpStreamData ) );

  assert( NULL != tmpCurStreamData );

#ifdef DEBUG
  if(opt->options & MO_PCM) 
  {
    DBG("PCM samples\n");
  }
  else 
  {
    DBG("MAD samples\n");
  }

  if(opt->options & MO_SWP) 
  {
      DBG("swap samples\n");
  }
  else
  {
      DBG("DONT swap samples\n");
  }
#endif

  if(opt->haveAudioAccept) {
    h->currBuff=1;
  }

  /* First the structures used by libmad must be initialized. */
  mad_stream_init(&Stream);
  mad_frame_init(&Frame);
  mad_synth_init(&Synth);
  mad_timer_reset(&Timer);

  //opt->bytesRead = 0;
  opt->timeStamp = 0;
  opt->fileSize  = opt->fileReader->fileSize;

    // set up output buffer
  OutputPtr=GetStreamBufferPtr(h);
  OutputBufferEnd=GetStreamBufferEnd(h);

  DBG("entering main loop\n");
  do {
    //DBG( "Dis %d, cb %d, kill %d, sk %d\n", opt->discardAndWait, buffIdx, StreamBufferKillThread(h), opt->skip ); 

    if(StreamBufferKillThread(h)) goto flush_and_kill;    

	// check discard
    if(opt->discardAndWait) {  // <discard>
    
discard_and_wait:
      DBG("discardAndWait\n"); 
      DiscardStreamBufferFull(opt->h);
      DiscardStreamBufferFull( h );

      if( opt->haveAudioAccept )
      {
       ResetAudioAccept( h, &buffIdx );
      }
      
      opt->discardAndWait = 1;

     mad_synth_finish(&Synth);
     mad_frame_finish(&Frame);
     mad_stream_finish(&Stream);

     mad_stream_init(&Stream);
     mad_frame_init(&Frame);
     mad_synth_init(&Synth);

      DBG( "Enter wait, File Buf %d, AA buf %d\n", 
          GetStreamBufferAvailable( opt->h ),
          GetStreamBufferAvailable( h ) );
      
      pthread_mutex_lock(&(opt->resumeMutex));
      //DBG("Resume Mutex locked\n");

      while( opt->discardAndWait && !StreamBufferKillThread(h) ) {
	    pthread_cond_wait(&(opt->resumeCond),&(opt->resumeMutex));
        //DBG("Waiting for Resume cond\n");
      }
      pthread_mutex_unlock(&(opt->resumeMutex));
      //DBG("Resume Mutex unlocked\n");

      if(StreamBufferKillThread(h)) goto kill_thread;

      DBG("resuming\n"); 

      if(opt->haveAudioAccept) {
		// special case, when skipping stop to flush buffer, then restart
//        DBG("resetting audioaccept buffer, start with buff 0\n");
//        ResetAudioAccept( h, &buffIdx ); 
        if( !opt->paused )
        {
         AudioAcceptSendStart();
        }
      }
	 OutputPtr       = GetStreamBufferPtr( h );
	 OutputBufferEnd = GetStreamBufferEnd( h );
     memset( InputBuffer, 0, MAD_INTERNAL_INPUT_BUFFER_SIZE + MAD_BUFFER_GUARD );
     
      DBG( "Done wait, File Buf %d, AA buf %d\n", 
          GetStreamBufferAvailable( opt->h ),
          GetStreamBufferAvailable( h ) );
    }	// </discard>

    if(Stream.buffer==NULL || Stream.error==MAD_ERROR_BUFLEN) {
      int              ReadSize  = 0,
                       Remaining = 0;
      unsigned char   *ReadStart = InputBuffer;
      int              nextStart = 1;

      ReadResult = readRetry = 0;

      switch( tmpCurStreamData->iBufferType )
      {
       case EPCMBuffer:
        assert( tmpCurStreamData->iFrameSize );
        assert( tmpCurStreamData->iUncompressedFrameSize );

        ReadSize = tmpCurStreamData->iUncompressedFrameSize * NON_MP3_FRAME_READ_COUNT;
        break;

       case ECustomEcodedBuffer:
        assert( tmpCurStreamData->iFrameSize );
        assert( tmpCurStreamData->iUncompressedFrameSize );

        ReadSize = tmpCurStreamData->iFrameSize * NON_MP3_FRAME_READ_COUNT;
        break;

       case EMP3Buffer:
        if(Stream.next_frame!=NULL) {
          Remaining=Stream.bufend-Stream.next_frame;
          memmove(InputBuffer,Stream.next_frame,Remaining);
          ReadStart=InputBuffer+Remaining;
          ReadSize=MAD_INTERNAL_INPUT_BUFFER_SIZE-Remaining;
        } else 
          ReadSize=MAD_INTERNAL_INPUT_BUFFER_SIZE;//,
                  //ReadStart=InputBuffer,
                  //Remaining=0;
        break;
      }

read_again:
      ReadResult=ReadStreamBuffer(opt->h,&readPtr,ReadSize);

      if( StreamBufferKillThread( h ) ||
          StreamBufferKillThread( opt->h ) ) goto kill_thread;
      
      if( opt->discardAndWait )
      {
       DBG( "Switching to discard state\n" );
       goto discard_and_wait;
      }
      //DBG( "MAD Read %d, ptr %p, size %d, fs %d\n", ReadResult, readPtr, ReadSize, tmpCurStreamData->iFrameSize );

      if( FlippedStreamBuffer( opt->h ) && ReadResult > 0 )
      {
       DBG( "StreamBuffer flipped, checking parameters...\n" );

       const stream_parameters_t* tmpStreamParameters = opt->fileReader->iGetStreamParameters( opt->fileReader );
       assert( NULL != tmpStreamParameters );
 
       if( memcmp( tmpStreamParameters, tmpCurStreamData, sizeof( stream_parameters_t ) ) )
       {
        DBG( "Parameters changed, flushing buffers...\n" );

        FlipOutputBuffer( h, opt, &buffIdx, OutputBufferEnd - OutputPtr );

        DBG( "Buffer flushed\n" );

        OutputPtr       = GetStreamBufferPtr( h );
	    OutputBufferEnd = GetStreamBufferEnd( h );

        memcpy( ( stream_parameters_t* )tmpCurStreamData, tmpStreamParameters, sizeof( stream_parameters_t ) );

        DBG( "Reinitialising MAD\n" );

        ReadStart = InputBuffer;
        ReadSize  = MAD_INTERNAL_INPUT_BUFFER_SIZE;
        memset( InputBuffer, 0, MAD_INTERNAL_INPUT_BUFFER_SIZE + MAD_BUFFER_GUARD );

        mad_synth_finish(&Synth);
        mad_frame_finish(&Frame);
        mad_stream_finish(&Stream);

        mad_stream_init(&Stream);
        mad_frame_init(&Frame);
        mad_synth_init(&Synth);
        mad_timer_reset(&Timer);

        DBG( "Reconfiguring stream\n" );
        ReconfigureStream( opt, tmpCurStreamData, h, &buffIdx );
       }
      }

skip:
	// check (and update) the skip variable safely
      pthread_mutex_lock(&(opt->skipMutex));
      skip=opt->skip;
      if(skip) { DBG("resetting mad skip\n"); opt->skip=0; }
      pthread_mutex_unlock(&(opt->skipMutex));

      if( ( ReadResult==-1 || skip ) && !opt->fileReader->currentFeature ) {
        DBG("got eof, rr %d, skip %d\n", ReadResult, skip );
		// eof
        if(GuardPtr || skip) {
		// guard ptr was set, so really eof, or we got a
		// skip result
          DBG("got eof, guard ptr set (skip=%i)\n",skip);
          if(skip) {
              DiscardStreamBufferFull(opt->h);
              DiscardStreamBufferFull( h );

            if(opt->haveAudioAccept && skip) {
		// special case, when skipping stop to flush buffer, then restart
              DBG("resetting audioaccept buffer, start with buff 0\n");              

              ResetAudioAccept( h, &buffIdx );

              if( !opt->paused )
              {
               AudioAcceptSendStart();
              }
            }
	       OutputPtr       = GetStreamBufferPtr( h );
	       OutputBufferEnd = GetStreamBufferEnd( h );

           DBG( "Re-initalising MAD decoder...\n" );

           ReadStart       = InputBuffer;
           ReadSize        = MAD_INTERNAL_INPUT_BUFFER_SIZE;
           memset( InputBuffer, 0, MAD_INTERNAL_INPUT_BUFFER_SIZE + MAD_BUFFER_GUARD );
           mad_synth_finish(&Synth);
           mad_frame_finish(&Frame);
           mad_stream_finish(&Stream);

           mad_stream_init(&Stream);
           mad_frame_init(&Frame);
           mad_synth_init(&Synth);
           mad_timer_reset(&Timer);
          }
	  if(opt->fileReader->hasNext) {
            DBG("got eof, hasNext\n");
		// set up shop for next mp3
	    StreamBufferResetEos(opt->h);
		// signal file reader to start feeding us again
            pthread_mutex_lock(&(opt->fileReader->nextMutex));
	    opt->fileReader->next=1;
	    pthread_cond_broadcast(&(opt->fileReader->nextCond));
	    pthread_mutex_unlock(&(opt->fileReader->nextMutex));
	    GuardPtr=NULL;
	    while(opt->fileReader->next && !h->killThread) usleep(1000);
	    if(h->killThread) goto kill_thread;
	    readRetry=0; 
        opt->bytesRead = 0;
        opt->timeStamp = 0;
        FrameCount = 0;
        opt->fileSize=opt->fileReader->fileSize;
            nextStart=1;
	    if(skip) opt->skipped=1;
	    goto read_again;
          } else {
            DBG("got eof, no hasNext\n");
		// be done with it
            if( opt->discardAndWait )
            {
             DBG( "go to discard and wait\n" );
             goto discard_and_wait;
            }

	    goto flush_and_kill;
          }
	} else {
		// guard ptr was not set, not really eof
          DBG("got eof, no guardPtr\n");
		// possibly partial frame, pad and decode
	  DBG("got stream end, padding..\n");
          GuardPtr=ReadStart;
          memset(GuardPtr,0,MAD_BUFFER_GUARD);
          ReadResult+=MAD_BUFFER_GUARD;
        }
      } else if(ReadResult==0) {
		// nothing available, try again later
        if( nextStart ) {
          usleep(10000);
          goto read_again;
        }
        if( ++readRetry > 100 ) {
	  DBG("%s: read failure after retry\n",__func__);
	  goto flush_and_kill;
	}
	usleep(10000);
	goto read_again;
      } else {
        nextStart=0;

        //if( opt->fileReader->iIsMP3( opt->fileReader ) )
        if( tmpCurStreamData->iBufferType == EMP3Buffer )
        {
         memcpy(ReadStart,readPtr,ReadResult);
        }
        
      }

      switch( tmpCurStreamData->iBufferType )
      {
       case EMP3Buffer:
       case ECustomEcodedBuffer:
        opt->bytesRead += ReadResult;
        break;

       case EPCMBuffer:
        opt->bytesRead += ReadResult * tmpCurStreamData->iFrameSize / tmpCurStreamData->iUncompressedFrameSize;
        break;
      }

      //if( opt->fileReader->iIsMP3( opt->fileReader ) )
      if( EMP3Buffer == tmpCurStreamData->iBufferType )
      {
       mad_stream_buffer(&Stream,InputBuffer,ReadResult+Remaining);
       Stream.error = 0;
      }
    }
  
    //if( opt->fileReader->iIsMP3( opt->fileReader ) ) 
    if( EMP3Buffer == tmpCurStreamData->iBufferType )
    {
    
     if(mad_frame_decode(&Frame,&Stream)) {
      if(MAD_RECOVERABLE(Stream.error)) {
        if(Stream.error!=MAD_ERROR_LOSTSYNC || Stream.this_frame!=GuardPtr) {
          DBG("recoverable frame level error\n");
        }
        continue;
      } else 
        if(Stream.error==MAD_ERROR_BUFLEN) continue;
        else {
          DBG("unrecoverable frame level error.\n");
          Status=1;
          break;
        }
     }

     FrameCount++;
     mad_timer_add(&Timer,Frame.header.duration);

     opt->timeStamp = mad_timer_count(Timer,MAD_UNITS_SECONDS);

     mad_synth_frame(&Synth,&Frame);

     PCMCount = Synth.pcm.length;

        // Checking stream parameters change
        if( Frame.header.samplerate != tmpCurStreamData->iSamplingRate )
        {
         DBG( "Parameters changed, flushing buffers...\n" );
         
         FlipOutputBuffer( h, opt, &buffIdx, OutputBufferEnd - OutputPtr );

         DBG( "Buffer flushed\n" );
         
         OutputPtr       = GetStreamBufferPtr( h );
	     OutputBufferEnd = GetStreamBufferEnd( h );

         DBG( "Frame sampling rate %d, stream reports %d, reconfiguring...\n", Frame.header.samplerate, tmpCurStreamData->iSamplingRate );

         const stream_parameters_t* tmpStreamParameters = opt->fileReader->iGetStreamParameters( opt->fileReader );
         assert( NULL != tmpStreamParameters );

         ( ( stream_parameters_t* )tmpCurStreamData )->iSamplingRate = Frame.header.samplerate;

         if( !tmpCurStreamData->iFrameSize )
         { // Only change for mp3
          ( ( stream_parameters_t* )tmpCurStreamData )->iFeatureBufferSize = DEFAULT_FEATURE_BYTES_READ * Frame.header.samplerate / 44100;
         }

         if( tmpStreamParameters->iSamplingRate != Frame.header.samplerate )
          {
           ( ( stream_parameters_t* )tmpStreamParameters )->iSamplingRate = Frame.header.samplerate;
           if( !tmpStreamParameters->iFrameSize )
           { 
            ( ( stream_parameters_t* )tmpStreamParameters )->iFeatureBufferSize = DEFAULT_FEATURE_BYTES_READ * Frame.header.samplerate / 44100;
           }
          }         
 
         ReconfigureStream( opt, tmpCurStreamData, h, &buffIdx );   
        }

    }
    else
    { // Decoding using external decoder directly into output buffer
     if( ReadResult > 0 && NULL != readPtr )
     {
      if( EPCMBuffer == tmpCurStreamData->iBufferType )
      {
       memcpy( ( void* )OutputPtr, readPtr, ReadResult );
       OutputPtr += ReadResult;
      }
      else
      {
       assert( tmpCurStreamData->iFrameSize );
      
       PCMCount = opt->fileReader->iDecodeFrame( opt->fileReader, readPtr, ReadResult, ( void* )OutputPtr );

       //DBG( "MAD Decoded %d, ptr %p, len %d, out ptr %p, unc %d\n", PCMCount, readPtr, ReadResult, OutputPtr, tmpCurStreamData->iUncompressedFrameSize );
       OutputPtr += PCMCount;
      }
     }
    }

	// copy to output buffer
    if( EMP3Buffer == tmpCurStreamData->iBufferType )
    {
     signed short Sample = 0;

     for( i=0 ; i < PCMCount; i++ ) {

     //if( opt->fileReader->iIsMP3( opt->fileReader ) ) 
      //if( tmpCurStreamData->iIsMP3  )
      {
        /* Left channel */
       PUT_SAMPLE( 0, opt->options, Synth, OutputPtr );

        /* Right channel. If the decoded stream is monophonic then
         * the right output channel is the same as the left one.
         */
       if(MAD_NCHANNELS(&Frame.header)==2) 
       { // stereo
	    PUT_SAMPLE( 1, opt->options, Synth, OutputPtr );
       } 
       else 
       {	// mono
	    PUT_SAMPLE( 0, opt->options, Synth, OutputPtr );
       }
      }    
     
        /* Flush the output buffer if it is full. */
       if(OutputPtr==OutputBufferEnd) {

          switch( FlipOutputBuffer( h, opt, &buffIdx, OutputBufferEnd - OutputPtr ) )
          {
           case EFlipDiscard:
	        goto discard_and_wait;

           case EFlipTerminate:
            goto kill_thread;

           case EFlipSkip:
            goto skip;

           default: 
	        OutputPtr       = GetStreamBufferPtr(h);
	        OutputBufferEnd = GetStreamBufferEnd(h);
          };           

      }
     }
    }
    else
    { // Non MP3
     assert( OutputPtr <= OutputBufferEnd );

     if( OutputBufferEnd - OutputPtr < tmpCurStreamData->iUncompressedFrameSize * NON_MP3_FRAME_READ_COUNT )
     {
          switch( FlipOutputBuffer( h, opt, &buffIdx, OutputBufferEnd - OutputPtr ) )
          {
           case EFlipTerminate:
            goto kill_thread;

           case EFlipSkip:
            goto skip;
          
           case EFlipDiscard:
	        goto discard_and_wait;
            
           default:
	        OutputPtr       = GetStreamBufferPtr( h );
	        OutputBufferEnd = GetStreamBufferEnd( h );
          }
     }
    }
  } while(1);

flush_and_kill:
  DBG("flush_and_kill\n");
	// we're done, set up last buffer, including the eos marker
  StreamBufferEos( h );

  FlipOutputBuffer( h, opt, &buffIdx, OutputBufferEnd - OutputPtr );

  /*if(opt->haveAudioAccept) {
    buffIdx=GetStreamBufferIndex(h);
  }

  if(opt->haveAudioAccept) {
    FlushStreamBufferAsync(h,OutputPtr-GetStreamBufferPtr(h));
    AudioAcceptSetLen(buffIdx,GetStreamBufferLength(h));
    AudioAcceptSync();
    AudioAcceptWaitForSync();
  } else {
    FlushStreamBuffer(h,OutputPtr-GetStreamBufferPtr(h));
  }*/

kill_thread:
  DBG("MAD kill thread\n");
  mad_synth_finish(&Synth);
  mad_frame_finish(&Frame);
  mad_stream_finish(&Stream);

  opt->terminated=1;
  pthread_exit(0);  
}

/* EOF */
