#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <string.h>

#include "Debug.h"
#include "StreamBuffer.h"
#include "AudioPlugIn.h"
#include "Encoder.h"
#include "Packager.h"
#include "A2dp.h"
#include "A2dpPlayer.h"
#include "DynamicSBC.h"
#include "A2DPPlugIn.h"

#define SBCBUF_SIZE	    ( 8    * 1024 )	    // size of the encoded, packaged SBC buffer
#define PCMBUF_SIZE	    ( 16   * 1024 )     // size of the decoded PCM sample buffer for PUSH model

typedef struct {
  // bluetooth related
  char                bdAddr[ 50 ];
  int                 bitpool;

  // a2dp related
  encoder_t          *encoder;
  packager_t         *packager;
  a2dp_packager_t    *a2dpPackager;
  StreamBufferHandle *inputBuffer;
  StreamBufferHandle *a2dpBuffer;
  a2dp_player_t      *a2dpPlayer;
  unsigned long       PREBUFFER_PACKETS;		// number of packets to send extra in the
  EPauseMode          aPauseMode;

  audio_output_model_t outputModel;

} a2dp_config_t;

static a2dp_config_t iA2DPConfig;
static int           iIsInitialised = 0;

static int   PlaybackActive( void );
static void  FreeSBCEncoder( void );

int PlugInStopPlayback( void );

int PlugInProcessPeerEvents( void )
{
 int tmpResult = 0;
 
 if( iIsInitialised && NULL != iA2DPConfig.packager )
 {
  assert( NULL != iA2DPConfig.packager->processpeerevents );
   
  //if( PlaybackActive() )
  {
   tmpResult = iA2DPConfig.packager->processpeerevents( iA2DPConfig.packager );
  }
 }

 return tmpResult;
}

int PlugInPlaybackFinished( int* aPlaybackInitialised, int* aPlaybackConnected )
{
 int tmpResult = 0;

 if( NULL != iA2DPConfig.a2dpPlayer )
 {
  tmpResult = iA2DPConfig.a2dpPlayer->eos;
 }

 if( NULL != aPlaybackInitialised ) 
 {
  *aPlaybackInitialised = ( NULL != iA2DPConfig.a2dpPlayer );
 }

 if( NULL != aPlaybackConnected )
 {
  *aPlaybackConnected = 0;

  if( NULL != iA2DPConfig.packager )
   if( NULL != iA2DPConfig.packager->udata )
   {
    a2dp_packager_t* h = ( a2dp_packager_t * )( iA2DPConfig.packager->udata );

    *aPlaybackConnected = ( h->cmdfd != -1 );
   }
 }

 return tmpResult;
}

int PlugInInit( audio_output_model_t aModel, void* aData )
{
 DBG( "A2DP: PlugIn Init...\n" );

 char tmpErrorText[ 2048 ];
 int  tmpResult = 1;

 assert( NULL != aData );

 a2dp_audio_output_plugin_data* tmpData = ( a2dp_audio_output_plugin_data* )( aData );

 DBG( "A2DP: bitpool %d, addr %s\n", tmpData->iBitPool, tmpData->iBTAddress );
 
 if( iIsInitialised )
 {
  iA2DPConfig.bitpool = tmpData->iBitPool;
  if( PlaybackActive() )
  { // playback active
   return 0; // Dynamic changing of bitpool or model not supported yet
  }
  else
  { // No playback, change address and model also
   strncpy( iA2DPConfig.bdAddr, tmpData->iBTAddress, 50 );
   iA2DPConfig.outputModel = aModel;

   if( NULL != iA2DPConfig.packager )
   {
    DBG("free> packager\n");
    iA2DPConfig.packager->destroy( iA2DPConfig.packager );
    iA2DPConfig.packager = NULL;
   }   
  }
 }
 else
 {
  memset( &iA2DPConfig, 0, sizeof( iA2DPConfig ) );
  
  assert( NULL != tmpData->iBTAddress );
  strncpy( iA2DPConfig.bdAddr, tmpData->iBTAddress, 50 );
  iA2DPConfig.bitpool     = tmpData->iBitPool;
  iA2DPConfig.outputModel = aModel;

  iIsInitialised = 1;
 }

 FreeSBCEncoder();

 DBG( "Setting up SBC encoder...\n" );
 iA2DPConfig.encoder = InitEncoder( iA2DPConfig.bitpool, tmpErrorText );

 if( NULL == iA2DPConfig.encoder )
 {
  DBG( "Can not initialise SBC encoder: %s\n", tmpErrorText );
  return 0;
 }

 DBG( "Initialising A2DP...\n" );
 if( NULL == iA2DPConfig.packager )
 {
  iA2DPConfig.packager = InitPackager(
                      iA2DPConfig.encoder,
					  iA2DPConfig.bdAddr,
                      NULL,
                      tmpErrorText );
  if( NULL == iA2DPConfig.packager )
  {
   tmpResult = 0;
  }
 }

 DBG( "End, pack %p, enc %p, res %d\n", iA2DPConfig.packager, iA2DPConfig.encoder, tmpResult );
 
 return tmpResult;
}

int PlugInGetInputFormat( void )
{
    switch( iA2DPConfig.encoder->functions->iSBCInputType() ) 
    {
      case SBC_IN_MAD:
        return MO_MAD;
        break;

      case SBC_IN_PCMS:
        return MO_PCM | MO_SWP;
        break;

      case SBC_IN_PCM:
        return MO_PCM | MO_NOSWP;
        break;

      default:
       DBG( "A2DP: Unknown SBC encoder type\n" );
       assert( 0 );
       return MO_PCM | MO_SWP;
    }
}

int PlugInSetVolume( int aVolume )
{ // A2DP does not support volume
 return 1;
}

int PlugInGetVolume( void )
{
 // A2DP does not support volume
 return 0;
}

int PlugInPushData( void* aData, int aLength )
{
 int tmpResult = -1;

 assert( EPushModel == iA2DPConfig.outputModel );
 
 if( NULL != iA2DPConfig.inputBuffer )
 {
  tmpResult = WriteStreamBuffer( iA2DPConfig.inputBuffer, aData, aLength );
 }

 return tmpResult;
}

#if 0
static void* DummyReadFunction( void* aData )
{
 DBG("Dummy read start...\n");

 assert( NULL != aData );

 StreamBufferHandle* h = ( StreamBufferHandle * )aData;

 do 
 {
   usleep( 1000 );
 }
 while( !StreamBufferKillThread( h ) );
  
 DBG( "Dummy read end...\n");

 // we're done, set up last buffer, including the eos marker
 StreamBufferEos( h );

 if( !!StreamBufferKillThread( h ) )
 {
  FlushStreamBuffer( h, GetStreamBufferLength( h ) - h->bufferFreeSize );
 }

 pthread_exit( 0 );  
}
#endif

static void SetPauseMode( int aPaused, EPauseMode aMode )
{
 if( NULL != iA2DPConfig.encoder )
 {
  iA2DPConfig.encoder->discarded = 0;
  iA2DPConfig.encoder->paused    = aPaused;
 }

 if( NULL != iA2DPConfig.a2dpPlayer )
 {
  iA2DPConfig.a2dpPlayer->paused = aPaused;
 }
 
 iA2DPConfig.aPauseMode         = aMode;            
}

int PlugInStartPlayback( int ( *aReadFunction )( void**ptr, int len ), char* aErrorText )
{
 DBG( "A2DP: Start playback...\n" );

 assert( NULL != aErrorText );

 int tmpResult = 0;

 *aErrorText = 0;

 if( iIsInitialised && !PlaybackActive() )
 {
  if( NULL != iA2DPConfig.encoder && NULL != iA2DPConfig.packager )
  {
   DBG( "Setting up input buffering, buffering model %d...\n", iA2DPConfig.outputModel );

   if( EPushModel == iA2DPConfig.outputModel )
   {
     assert( NULL == aReadFunction );

     iA2DPConfig.inputBuffer = InitStreamBuffer( 
                              PCMBUF_SIZE,
                              NULL, //DummyReadFunction, // Empty reader, synchronous buffer
                              NULL,
                              BO_NOIBLK );
   }
   else
   { // Pull model
     assert( NULL != aReadFunction );

     DBG( "PULL model, Read func addr %p\n", aReadFunction );
     
     iA2DPConfig.inputBuffer = InitStreamBuffer( 
                              0,
                              (void *(*)(void *))aReadFunction,
                              NULL,
                              BO_IFAKE );
   }

   if( NULL != iA2DPConfig.inputBuffer )
   {
    DBG( "Setting up A2DP packager, InputBuffer %p, Enc %p...\n", iA2DPConfig.inputBuffer, iA2DPConfig.packager->encoder );

    iA2DPConfig.packager->in = iA2DPConfig.inputBuffer;

    DBG( "InputBuffer set to %p, connecting...\n", iA2DPConfig.packager->in );
    
    if( iA2DPConfig.packager->connect( iA2DPConfig.packager, iA2DPConfig.bdAddr, aErrorText ) )
    { 
     iA2DPConfig.a2dpPackager = ( ( a2dp_packager_t * )( iA2DPConfig.packager->udata ) );

     DBG( "Init A2DP buffer...\n" );
     iA2DPConfig.a2dpBuffer = InitStreamBuffer( 
                                SBCBUF_SIZE,
                                packagerProducer,
			                    ( void* )( iA2DPConfig.packager ), 
                                BO_NOIBLK );

     if( NULL != iA2DPConfig.a2dpBuffer )
     {
      iA2DPConfig.packager->out = iA2DPConfig.a2dpBuffer;

      DBG( "Allocated A2DP packet buffer, size %d, addr %p\n", SBCBUF_SIZE, iA2DPConfig.packager->out );

      // start a2dp player
      DBG( "Starting A2DP player...\n" );
      iA2DPConfig.a2dpPlayer = InitA2dpPlayer( iA2DPConfig.packager, aErrorText );

      //DBG( "Started %p\n", iA2DPConfig.a2dpPlayer );

      if( NULL == iA2DPConfig.a2dpPlayer ) 
      {
       strcat( aErrorText, "Can not start A2DP player\n" );     
      }
      else
      {
       //DBG( "Started 2\n" );
       SetPauseMode( 0, KNoPause );
       //DBG( "Started 3\n" );
       tmpResult = 1;
      }
     }
     else
     {
      strcat( aErrorText, "Can not allocate Bluetooth streaming buffer\n" );
     }
    }
    else
    {
     strcat( aErrorText, "Can not connect to remote device\n" ); 
    }
   }
   else
   {
    strcat( aErrorText, "Can not initialize input buffer\n" ); 
   }     
  }
  else
  {
   strcat( aErrorText, "A2DP PlugIn not initialised\n" );
  }

  if( !tmpResult )
  { // Error stopping playback
   DBG( "Start error, stopping...\n" );
   //PlugInStopPlayback();
 
   PlugInDone();

   DBG( "Reinit...\n" );
   a2dp_audio_output_plugin_data tmpData;

   tmpData.iBitPool = iA2DPConfig.bitpool;
   strcpy( tmpData.iBTAddress, iA2DPConfig.bdAddr );

   PlugInInit( iA2DPConfig.outputModel, &tmpData );
  }
 }

 /*if( tmpResult )
 { // If we could start, do not return until a kind of link is established :)
  while( iA2DPConfig.encoder->empty_count < 200 )
  {
   usleep( 100 );
  }

  if( iA2DPConfig.a2dpPlayer->eos )
  {
   DBG( "Error sending empty packets, returning error\n" );
   PlugInStopPlayback();
   tmpResult = 0;
  }
 }*/

 DBG( "End, result %d\n", tmpResult );
 
 return tmpResult;
}

int PlugInGetPosition( void )
{
 DBG( "A2DP: Get position...\n" );

 int tmpResult = 0;

 if( iIsInitialised && PlaybackActive() )
 {
   tmpResult = iA2DPConfig.a2dpPlayer->currentPos / 1000000ULL; // Converting to seconds
 }

 return tmpResult;
}

int PlugInStopPlayback( void )
{
 DBG( "A2DP: Stop playback...\n" );

 int tmpResult = 0;

 if( iIsInitialised )
 {
  DBG( "Freeing resources...\n" );

  SetPauseMode( 0, KNoPause );

  if( NULL != iA2DPConfig.a2dpPlayer )
  {
   DBG("free> a2dpPlayer\n");
   DestroyA2dpPlayer( iA2DPConfig.a2dpPlayer );
  }

  // a2dpBuffer 
  if( NULL != iA2DPConfig.a2dpBuffer )
  {
    DBG("free> a2dpBuffer\n");
    DoneStreamBuffer( iA2DPConfig.a2dpBuffer );
  }

  // packager
  //  +--> a2dpPackager
  if( NULL != iA2DPConfig.packager )
  {
    DBG("free> packager\n");
    iA2DPConfig.packager->destroy( iA2DPConfig.packager );
  }

  // a2dpBuffer 
  if( NULL != iA2DPConfig.inputBuffer )
  {
    DBG("free> inputBuffer\n");
    DoneStreamBuffer( iA2DPConfig.inputBuffer );
  }

  // Resetting all fields
  iA2DPConfig.packager          = NULL;
  iA2DPConfig.a2dpBuffer        = NULL;
  iA2DPConfig.a2dpPlayer        = NULL;
  iA2DPConfig.a2dpPackager      = NULL;
  iA2DPConfig.inputBuffer       = NULL;

  tmpResult = 1;
 }

 DBG( "A2DP: Stop playback done\n" );
 
 return tmpResult;
}

int PlugInPausePlayback( EPauseMode aPauseMode, char* aErrorText )
{
 DBG( "A2DP: Pause playback, mode %d...\n", aPauseMode );

 assert( NULL != aErrorText );

 int tmpResult = 0;
 
 *aErrorText = 0;

 if( PlaybackActive() )
 {
  if( !iA2DPConfig.a2dpPlayer->paused )
  {
   // Pausing sender thread
   if( KSync == aPauseMode )
   {
    iA2DPConfig.a2dpPlayer->sync = 1;

    // Busy loop, waiting for sync
    while( !iA2DPConfig.a2dpPlayer->paused )
    {
     DBG( "Waiting for A2DP player to sync...\n" );
     usleep( 100000 );
    }
   }
   else
    if( KDiscard == aPauseMode ) 
    {
     DBG( "Discarding A2DP buffer...\n" );
     DiscardStreamBufferFull( iA2DPConfig.a2dpBuffer );

     iA2DPConfig.encoder->discarded = 1;
     while( !iA2DPConfig.encoder->paused )
     {
      DBG( "Waiting for encoder thread to pause...\n" );
      usleep( 100000 );
     }
    }
    
   SetPauseMode( 1, aPauseMode );

   tmpResult = iA2DPConfig.packager->pause( iA2DPConfig.packager, aErrorText );
  }
  else
  { // Already paused
   tmpResult = 1;
  }
 } 

 return tmpResult;
}

int PlugInResumePlayback( char* aErrorText )
{
 DBG( "A2DP: Resume playback...\n" );

 assert( NULL != aErrorText );

 int tmpResult = 0;
 
 *aErrorText = 0;

 if( PlaybackActive() )
 {
  if( iA2DPConfig.a2dpPlayer->paused )
  {
//   if( KDiscard == iA2DPConfig.aPauseMode )
   {
    DBG( "Discarding A2DP buffer...\n" );
    DiscardStreamBufferFull( iA2DPConfig.a2dpBuffer );    
   }

    // Resuming encoder thread
   //if( KDiscard == iA2DPConfig.aPauseMode )    
    {
     iA2DPConfig.encoder->resumed = 1;
    }

   iA2DPConfig.aPauseMode = KNoPause;
   
   tmpResult = iA2DPConfig.packager->resume( iA2DPConfig.packager, aErrorText );

   if( tmpResult )
   {    
    // Resuming sender thread
    iA2DPConfig.a2dpPlayer->resumed = 1;

    SetPauseMode( 0, KNoPause );
   }
  }
  else
  { // Already playing back
   tmpResult = 1;
  }
 } 

 return tmpResult;
}

int PlugInDone( void )
{
 DBG( "A2DP: Done plugin...\n" );
 
 int tmpResult = 0;

 if( iIsInitialised )
 {
  PlugInStopPlayback();
  FreeSBCEncoder();

  iIsInitialised = 0;
  tmpResult      = 1;
 }

 return tmpResult;
}

static int PlaybackActive( void )
{
 return ( iIsInitialised && 
          NULL != iA2DPConfig.a2dpPackager &&
          NULL != iA2DPConfig.a2dpPlayer ) ? 1: 0;
}


static void  FreeSBCEncoder( void )
{
 // Nobody should use SBC if it's deleted
 assert( NULL == iA2DPConfig.a2dpPackager );
 assert( NULL == iA2DPConfig.a2dpPlayer );

 if( NULL != iA2DPConfig.encoder )
 {
  DBG( "A2DP: Free SBC encoder...\n" );
  iA2DPConfig.encoder->destroy( iA2DPConfig.encoder );
  iA2DPConfig.encoder = NULL;
 }
}
