
#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <dlfcn.h>
#include <assert.h>
#include <string.h>

#include "Debug.h"
#include "AudioPlugIn.h"

static void* GetDLLSymbol( void* aDLL, const char* aSymbolName, char* aErrorText )
{
 assert( NULL != aDLL );
 assert( NULL != aSymbolName );
 assert( NULL != aErrorText );

 void* tmpAddr = dlsym( aDLL, aSymbolName );
 char* sbc_dl_err;
 
 if( ( sbc_dl_err = dlerror() ) != NULL )  {
    sprintf( aErrorText, "Unable to load symbol %s: %s\n", aSymbolName, sbc_dl_err );
    fprintf( stderr, "%s", aErrorText );
    tmpAddr = NULL;
  }

 return tmpAddr;
}

static int DefaultPushData( void* aData, int aLength )
{
 return -1; // Error
}

audio_input_plugin_t* LoadAudioInputPlugIn( const char* aFileName, char* aErrorText )
{
 DBG( "Loading Audio input PlugIn %s...\n", aFileName );

 assert( NULL != aFileName );
 assert( NULL != aErrorText );

 audio_input_plugin_t* tmpResult = NULL;
 
 *aErrorText = 0;

 tmpResult = malloc( sizeof( audio_input_plugin_t ) );

 if( NULL != tmpResult )
 {
  memset( tmpResult, 0, sizeof( audio_input_plugin_t ) );

  // Loading DLL
  DBG( "Loading dynamic object \"%s\"\n", aFileName );
  
  tmpResult->iInternalData = dlopen( aFileName, RTLD_NOW );

  if( !tmpResult->iInternalData ) 
  {
    sprintf( aErrorText, "Audio input plugin library %s not loaded: %s\n", aFileName, dlerror() ); 
    fprintf( stderr,"%s", aErrorText );
    goto free_and_exit;
  }

  tmpResult->iPlugInInit             = GetDLLSymbol( tmpResult->iInternalData, "PlugInInit",             aErrorText );
  tmpResult->iPlugInPlayFile         = GetDLLSymbol( tmpResult->iInternalData, "PlugInPlayFile",         aErrorText );
  tmpResult->iPlugInSkipToNextFile   = GetDLLSymbol( tmpResult->iInternalData, "PlugInSkipToNextFile",   aErrorText );
  tmpResult->iPlugInReadStreamBuffer = GetDLLSymbol( tmpResult->iInternalData, "PlugInReadStreamBuffer", aErrorText );
  tmpResult->iPlugInDone             = GetDLLSymbol( tmpResult->iInternalData, "PlugInDone",             aErrorText );

  if( NULL == tmpResult->iPlugInInit             ||
      NULL == tmpResult->iPlugInPlayFile         ||
      NULL == tmpResult->iPlugInSkipToNextFile   ||
      NULL == tmpResult->iPlugInReadStreamBuffer || 
      NULL == tmpResult->iPlugInDone             )
  {
   dlclose( tmpResult->iInternalData );
   goto free_and_exit;
  }
 }
 else
 {
  strcpy( aErrorText, "Can not allocate memory" );
 }

 return tmpResult;

free_and_exit:
  free( tmpResult );

 return NULL;
}

void UnLoadAudioInputPlugIn( audio_input_plugin_t* aPlugIn )
{
 DBG( "Unloading Audio input plugin...\n" );
 
 assert( NULL != aPlugIn ); 

 aPlugIn->iPlugInDone();

 dlclose( aPlugIn->iInternalData );

 free( aPlugIn );
}

audio_output_plugin_t* LoadAudioOutputPlugIn( const char* aFileName, char* aErrorText )
{
 DBG( "Loading Audio output PlugIn %s...\n", aFileName );

 assert( NULL != aFileName );
 assert( NULL != aErrorText );

 audio_output_plugin_t* tmpResult = NULL;
 
 *aErrorText = 0;

 tmpResult = malloc( sizeof( audio_output_plugin_t ) );

 if( NULL != tmpResult )
 {
  memset( tmpResult, 0, sizeof( audio_output_plugin_t ) );

  // Loading DLL
  DBG( "Loading dynamic object \"%s\"\n", aFileName );
  
  tmpResult->iInternalData = dlopen( aFileName, RTLD_NOW );

  if( !tmpResult->iInternalData ) 
  {
    sprintf( aErrorText, "Audio output plugin library %s not loaded: %s\n", aFileName, dlerror() ); 
    fprintf( stderr,"%s", aErrorText );
    goto free_and_exit;
  }

  tmpResult->iPlugInInit           = GetDLLSymbol( tmpResult->iInternalData, "PlugInInit",           aErrorText );
  tmpResult->iPlugInSetVolume      = GetDLLSymbol( tmpResult->iInternalData, "PlugInSetVolume",      aErrorText );
  tmpResult->iPlugInGetVolume      = GetDLLSymbol( tmpResult->iInternalData, "PlugInGetVolume",      aErrorText );
  tmpResult->iPlugInStartPlayback  = GetDLLSymbol( tmpResult->iInternalData, "PlugInStartPlayback",  aErrorText );
  tmpResult->iPlugInStopPlayback   = GetDLLSymbol( tmpResult->iInternalData, "PlugInStopPlayback",   aErrorText );
  tmpResult->iPlugInResumePlayback = GetDLLSymbol( tmpResult->iInternalData, "PlugInResumePlayback", aErrorText );
  tmpResult->iPlugInPausePlayback  = GetDLLSymbol( tmpResult->iInternalData, "PlugInPausePlayback",  aErrorText );
  tmpResult->iPlugInDone           = GetDLLSymbol( tmpResult->iInternalData, "PlugInDone",           aErrorText );
  tmpResult->iPlugInGetPosition    = GetDLLSymbol( tmpResult->iInternalData, "PlugInGetPosition",    aErrorText );
  tmpResult->iPlugInPushData       = GetDLLSymbol( tmpResult->iInternalData, "PlugInPushData",       aErrorText );
  tmpResult->iPlugInGetInputFormat = GetDLLSymbol( tmpResult->iInternalData, "PlugInGetInputFormat", aErrorText );

  if( NULL == tmpResult->iPlugInPushData )
  {
   DBG( "PlugInPushData() not found, setting up default\n" );
   
   tmpResult->iPlugInPushData = DefaultPushData;
  }

  if( NULL == tmpResult->iPlugInInit             ||
      NULL == tmpResult->iPlugInGetVolume        ||
      NULL == tmpResult->iPlugInPausePlayback    ||
      NULL == tmpResult->iPlugInResumePlayback   ||
      NULL == tmpResult->iPlugInSetVolume        ||
      NULL == tmpResult->iPlugInStartPlayback    ||
      NULL == tmpResult->iPlugInStopPlayback     ||
      NULL == tmpResult->iPlugInGetPosition      ||
      NULL == tmpResult->iPlugInGetInputFormat   ||
      NULL == tmpResult->iPlugInDone             )
  {
   dlclose( tmpResult->iInternalData );
   goto free_and_exit;
  }
 }
 else
 {
  strcpy( aErrorText, "Can not allocate memory" );
 }

 return tmpResult;

free_and_exit:
  free( tmpResult );

 return NULL;
}

void UnLoadAudioOutputPlugIn( audio_output_plugin_t* aPlugIn )
{
 DBG( "Unloading Audio output plugin...\n" );
 
 assert( NULL != aPlugIn ); 

 aPlugIn->iPlugInDone();
 
 dlclose( aPlugIn->iInternalData );

 free( aPlugIn );
}
