
#ifndef _AUDIO_PLUGIN_H_
#define _AUDIO_PLUGIN_H_

#ifdef __cplusplus
extern "C" {
#endif

/** \file Audio PlugIn.h

 Defines the interface for audio input/output
*/

/** \brief Audio output format, those values could be OR'ed

*/
#define MO_PCM    (1<<0)                        /**< output PCM samples */
#define MO_MAD    (0<<0)                        /**< output MAD samples (4 bytes fiexd-point)*/
#define MO_SWP    (1<<1)                        /**< swap PCM samples (ignored 
                                                     for non-pcm output) */
#define MO_NOSWP  (0<<1)                        /**< don't swap */

/** \brief Audio input, all functions return 1 in case of 
           success and 0 in case of error

 Audio always has 44.1kHz sampling rate, 2 channels, 2 bytes per sample
*/
typedef struct
{
 /** \brief Initializes Audio PlugIn library, aData is PlugIn specific */
 int ( *iPlugInInit )( int aOutputFormat, void* aData );

 /** \brief Plays or enqueues a file for playback */
 int ( *iPlugInPlayFile )( const char* aFileName, char* aErrorText );

 /** \brief Skips current playback to given file */
 int ( *iPlugInSkipToNextFile ) ( const char* aFileName, char* aErrorText );

 /** \brief  Reads next PCM data chunk from MP3 PlugIn

 \param ptr Pointer to a pointer where the data location are stored
 \param len Length of requested data block

 Return values:
 Positive value - Number of read bytes
 0              - Data not ready, try again
 -1             - EOF or hard error, playback should stop
 */
 int ( *iPlugInReadStreamBuffer )( void**ptr, int len );

 /** \brief Stops the playback and releases all PlugIn resources */
 int ( *iPlugInDone )( void );

 void* iInternalData;
} audio_input_plugin_t;


/** \brief Audio output models */
typedef enum { EPullModel, EPushModel } audio_output_model_t;

/** \brief Audio output, all functions return 1 in case of 
           success and 0 in case of error

 Audio always has 44.1kHz sampling rate, 2 channels, 2 bytes per sample */
typedef struct 
{
 /** \brief Initialize Audio Output PlugIn, aData is PlugIn specific.
 
   
 Audio output model can not be changed after initialization */
 int ( *iPlugInInit )( audio_output_model_t aModel, void* aData );

 /** \brief Retuns format of input data in MO_XXX options */
 int ( *iPlugInGetInputFormat )( void );

 /** \brief Sets the volume
 
 Volume in range 0-100 */
 int ( *iPlugInSetVolume )( int aVolume );

 /** \brief Returns current volume */
 int ( *iPlugInGetVolume )( void );

 /** \brief Starts the playback 
 
 aReadFunction should be NULL for PUSH model and NOT NULL otherwise 
 
 \note Start playback should perform device initialization. In case of BlueTooth,
 it can take a long time, depending from the device used and its avaiability */ 
 int ( *iPlugInStartPlayback )( int ( *aReadFunction )( void**ptr, int len ), char* aErrorText );

 /** \brief Pushes the data for playback. 
 
 Should be called only in PUSH mode
 Returns number of pushed bytes or -1 in case of error  
 
 \param aData   Pointer to PCM data to play
 \param aLength Length of PCM block to play. Must be kByte aligned. 
                Minimum recommended value - 16k, 
                Maximum recommended value - 64k

 \note Because of real-time nature of audio, in order to avoid audio breaks,
       this function should be called at regular intervals.  
       User has to take into account the possible delay for particular PlugIn, caused by
       internal buffering and processing
 */
 int ( *iPlugInPushData )( void* aData, int aLength );

 /** \brief Stops the playback */
 int ( *iPlugInStopPlayback )( void );

 /** \brief Pauses the playback */
 int ( *iPlugInPausePlayback )( char* aErrorText );

 /** \brief Resumes paused playback */
 int ( *iPlugInResumePlayback )( char* aErrorText );

 /** \brief Returns currentplayback position in seconds */
 int ( *iPlugInGetPosition )( void );

 /** \brief Stops the playback and releases all PlugIn resources */
 int ( *iPlugInDone )( void );

 void* iInternalData;
} audio_output_plugin_t;

/** \brief A2DP Audio Output PlugIn data definition 

 Should be sent to A2DP PlugIn upon initialization or at any other moment
 if set during playback, only bitpool value is changed */
typedef struct 
{
 int  iBitPool;           /**< BitPool (Quality), value in range 10 - 96, use 56 for medium quality */
 char iBTAddress[ 50 ];   /**< Address of BlueTooth device to connect to */
} a2dp_audio_output_plugin_data;

/** \brief Loads an audio input PlugIn dynamically from given .so file */
audio_input_plugin_t* LoadAudioInputPlugIn( const char* aFileName, char* aErrorText );

/** \brief Releases given audio input PlugIn. 

 After this call, aPlugIn pointer becomes invalid and should not be used any more.
 Prior calling this function, iPlugInDone() function of this PlugIn should be called */
void                  UnLoadAudioInputPlugIn( audio_input_plugin_t* aPlugIn );

/** \brief Loads an audio output PlugIn dynamically from given .so file */
audio_output_plugin_t* LoadAudioOutputPlugIn( const char* aFileName, char* aErrorText );

/** \brief Releases given audio output PlugIn. 

 After this call, aPlugIn pointer becomes invalid and should not be used any more.
 Prior calling this function, iPlugInDone() function of this PlugIn should be called */
void                   UnLoadAudioOutputPlugIn( audio_output_plugin_t* aPlugIn );

#ifdef __cplusplus
}
#endif

#endif // _AUDIO_PLUGIN_H_

