/* 
 * File   : FileReader.c
 * Author : Koen Martens <koen.martens@tomtom.com>
 * Desc.  : Buffering file reader, simply fills the
 *          StreamBuffer with raw file data to be 
 *          consumed by the decoder. Keep the file 
 *          buffer large, to protect against hd spinup
 *          delays and such.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * Documentation of the statemachine in function fileProducer:
 * 
 * State 1 , READING      Reading of 1MB blocks from current file
 * State 2 , FLUSHING     Flushing block from current file
 * State 3 , EXIT         Exit fileReader thread
 * State 4 , EOF          Flush last block from current file
 *                        and set up shop for pre-buffering
 * State 6 , PREREADING   Read first 1MB block from next file
 * State 7 , WAITNEXT     Wait for next flag from consumer
 * State 8 , FLUSHINGPRE  Flush pre-read block from next file
 * State 9 , SKIP         Just falls through to 4
 * State 10, IDLE         Wait for next song or termination
 *
 * State transitions (Default=none of the other conditions apply):
 *
 *        From       |        To        |         Condition
 * ------------------+------------------+-----------------------------
 *  1 , READING      | 4 , EOF          | End-of-file on current file
 *                   | 2 , FLUSHING     | Buffer full
 *                   | 1 , READING      | Default
 *                   | 9 , SKIP         | Skip
 *  2 , FLUSHING     | 3 , EXIT         | killThread==1
 *                   | 1 , READING      | Default
 *  4 , EOF          | 6 , PREREADING   | Next file initialized
 *                   | 10, IDLE         | No next file queued
 *                   | 10, IDLE         | Next file initialisation failed
 *                   | 3 , EXIT         | killThread==1
 *  6 , PREREADING   | 7 , WAITNEXT     | End-of-file on next file
 *                   | 7 , WAITNEXT     | Not end-of-file on next file
 *                   | 9 , SKIP         | Skip
 *  7 , WAITNEXT     | 4 , EOF          | End-of-file on next file
 *                   | 8 , FLUSHINGPRE  | Not end-of-file on next file
 *  8 , FLUSHINGPRE  | 3 , EXIT         | killThread==1
 *                   | 1 , READING      | Default
 *  9 , SKIP         | 4 , EOF          | Default
 *  10, IDLE         | 6 , PREREADING   | Next file was queued & is initialized
 *                   | 3 , EXIT         | killThread==1
 *                   | 10, IDLE         | Default
 *  11, DISCARD      | 11, EXIT         | killThread==1
 *                   |  1, READING      | Default
 * 
 *
 */

#include <stdio.h>
#include <assert.h>
#include <malloc.h>
#include <pthread.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifdef NMP3DEBUG
#undef DEBUG
#endif

#include "Debug.h"
#include "StreamBuffer.h"
#include "ReadFile.h"
#include "MadProducer.h"

#define MP3_SAMPLING_RATE   44100
#define MP3_NUM_CHANNELS    2
#define MP3_BITS_PER_SAMPLE 16

#define MAX_READ_0_RETRIES  20

// states:

#define S_READING          1   // Reading of 1MB blocks from current file
#define S_FLUSHING         2   // Flushing block from current file
#define S_EXIT             3   // Exit fileReader thread
#define S_EOF              4   // Flush last block from current file
                               // and set up shop for pre-buffering
#define S_PREREADING       6   // Read first 1MB block from next file
#define S_WAITNEXT         7   // Wait for next flag from consumer
#define S_FLUSHINGPRE      8   // Flush pre-read block from next file
#define S_SKIP             9   // Just falls through to 4
#define S_IDLE             10  // Wait for next song or termination
#define S_DISCARD          11  // Discard and wait

#define DEFAULT_FILE_SIZE_CORRECTION 1024


static stream_parameters_t mp3_stream_parameters =
{
 MP3_SAMPLING_RATE, 
 MP3_NUM_CHANNELS, 
 MP3_BITS_PER_SAMPLE, 
 EMP3Buffer, 
 0, 0, 
 DEFAULT_FILEBUF_SIZE, 
 DEFAULT_FEATURE_BYTES_READ, 
 DEFAULT_MIN_READ_SIZE,
 0
};

static size_t GetPrimaryBufferLen( StreamBufferHandle* h, fileReaderData* data ) 
{
 return ( ENoFeature == data->currentFeature ) ? 
         MIN( GetStreamBufferLength( h ), ( int )( data->iGetStreamParameters( data )->iBufferSize ) ): 
         data->iGetStreamParameters( data )->iFeatureBufferSize;
}

static int GetFileSize( FILE* fp )
{
 assert( NULL != fp );

 struct stat tmpStat;
 int         tmpResult = 0;

 if( !fstat( fileno( fp ), &tmpStat ) )
 {
  tmpResult = tmpStat.st_size;
 }

 return tmpResult; 
}

static long GetEncodedDataSize( fileReaderData *fileReader, FILE* fp )
{
 long tmpResult = fileReader->iGetStreamParameters( fileReader )->iEncodedDataSize;

 if( !tmpResult )
 {
  tmpResult = GetFileSize( fp );
  if( tmpResult > DEFAULT_FILE_SIZE_CORRECTION )
  {
   tmpResult -= DEFAULT_FILE_SIZE_CORRECTION; // Average length of all tags, etc.
  }
 }

 return tmpResult;
}

static void* fileProducer(void *arg);
static int   setFileProducerNextFile( fileReaderData *fileReader, const char * aName );

static void DestroyFileProducer( fileReaderData *fileReader) 
{
  assert( NULL != fileReader );
  
  pthread_mutex_destroy( &fileReader->nextMutex );
  pthread_mutex_destroy( &fileReader->wakeupMutex );
  pthread_mutex_destroy( &fileReader->skipMutex );
  pthread_mutex_destroy( &fileReader->resumeMutex );

  pthread_cond_destroy( &fileReader->nextCond );
  pthread_cond_destroy( &fileReader->wakeupCond );
  pthread_cond_destroy( &fileReader->resumeCond );
  
  if( NULL != fileReader->internalData )
  {
   free( fileReader->internalData );
  }

  if( NULL != fileReader->internalData2 )
  {
   free( fileReader->internalData2 );
  }

  free( fileReader );
}

static int StartFeatureMode( struct __fileReaderData *fileReader, int aMode, int aSpeed, char* aErrorText )
{
 assert( NULL != fileReader );
 
 if( NULL != aErrorText )
 {
  *aErrorText = 0;
 }
 
 const int tmpStartFeature = ( ( int )( fileReader->currentFeature ) != aMode );

 fileReader->currentFeature       = ENoFeature;
 fileReader->currentFeatureSpeed  = 0;
 fileReader->startFeature         = 0;

 if( ENoFeature != aMode )
 {
  fileReader->currentFeatureSpeed  = aSpeed;
  fileReader->currentFeature       = aMode;
  fileReader->startFeature         = tmpStartFeature;
 }
 
 DBG( "Feature mode set to %d, speed %d, start %d\n", fileReader->currentFeature, fileReader->currentFeatureSpeed, fileReader->startFeature );
 
 return 1;
}

static int ReadFileData( struct __fileReaderData *fileReader, void* aBuffer, int aBufferLen )
{
 assert( NULL != fileReader );
 assert( NULL != aBuffer );  
 
 if( fileReader->currentFeature && fileReader->currentFeatureSpeed )
 { // Feature mode
  DBG( "Feature mode %d, speed %d\n", fileReader->currentFeature, fileReader->currentFeatureSpeed );
  
  // Calculating new jump position
  long tmpCurrentJumpOffset = ( fileReader->fileSize * fileReader->currentFeatureSpeed / 100 + 
                                      ( ( EFastReverseFeature == fileReader->currentFeature ) ? fileReader->iGetStreamParameters( fileReader )->iFeatureBufferSize: 0 ) ) *
                                    ( ( EFastReverseFeature == fileReader->currentFeature ) ? -1: 1 );
   
  if( fileReader->startFeature )
   {
    /*int tmpCorr = ( int )( fileReader->iGetStreamParameters( fileReader )->iBufferSize * ( ( EFastReverseFeature == fileReader->currentFeature ) ? 1.5: 1.8 ) );
    
    DBG( "Feature just started, correcting file pos %ld for %d bytes\n", ftell( fileReader->fp ), tmpCorr );

    if( tmpCorr > ftell( fileReader->fp ) )
    {
     tmpCorr = ftell( fileReader->fp );
    }

    fseek( fileReader->fp, -tmpCorr, SEEK_CUR );*/

    const int tmpCorr = ( ( MadOpt *)fileReader->udata )->bytesRead;
    
    DBG( "Feature just started file pos %ld, jumping to %d\n", ftell( fileReader->fp ), tmpCorr );

    if( fseek( fileReader->fp, tmpCorr, SEEK_SET ) )
    { // Error
     goto error;   
    }

    fileReader->startFeature = 0;
   }

   DBG( "Jumping %ld bytes from current position %ld, file size %d...\n", tmpCurrentJumpOffset, ftell( fileReader->fp ), fileReader->fileSize );
   
   if( fseek( fileReader->fp, tmpCurrentJumpOffset, SEEK_CUR ) )
   { // Error
error:
    DBG( "Error %d (%s), cancelling feature mode\n", errno, strerror( errno ) );
    
    fileReader->iStartFeatureMode( fileReader, ENoFeature, 0, NULL );
   }    
  
  if( ftell( fileReader->fp ) > ( int )( fileReader->fileSize ) )
  { // Seek to EOF
   DBG( "EOF, seeking to the end of file %d (%s)\n", errno, strerror( errno ) );

   fseek( fileReader->fp, 0, SEEK_END );

   fileReader->iStartFeatureMode( fileReader, ENoFeature, 0, NULL );
  }

  // Setting file position for Mad
  ( ( MadOpt *)fileReader->udata )->bytesRead = ftell( fileReader->fp );

  DBG( "New file pos %ld\n", ftell( fileReader->fp ) ); 
 }

 return fread( aBuffer, 1, aBufferLen, fileReader->currfp );
}


static int InitReadFile( struct __fileReaderData *fileReader, FILE* aFile , const char* aFileName, char *errorTxt )
{
 assert( NULL != fileReader );

 fileReader->openFilePos = 0;

 return 0;
}

static void CloseReadFile( struct __fileReaderData *fileReader )
{
}

static int MP3EOF( struct __fileReaderData *fileReader )
{
 assert( NULL != fileReader );

 return feof( fileReader->currfp );
}

static const stream_parameters_t*  GetStreamParameters( struct __fileReaderData *fileReader )
{
 return &mp3_stream_parameters;
}

static int BytePosToChapter( struct __fileReaderData *fileReader, unsigned long aBytePos, unsigned long* aChapterNumber, unsigned long* aTimeMS )
{
 assert( NULL != aChapterNumber );
 assert( NULL != aTimeMS );

 *aChapterNumber = 0;
 *aTimeMS        = 0;

 return 0; // Failed
}

fileReaderData *InitMP3FileProducer(const char *fname, unsigned long aByteOffset ) {
  fileReaderData *data;

  data=malloc(sizeof(fileReaderData));
  if(!data) 
  {
      DBG( "Can not allocate memory for MP3 File Producer\n" );

      return NULL;
  }

  memset( data, 0, sizeof( fileReaderData ) );

  data->fp = fopen( fname,"rb" );
  if(!data->fp) 
  { 
      DBG( "Can not open file %s\n", fname );

      free(data); 
      return NULL; 
  }

  data->fileSize = GetFileSize( data->fp );
  data->nextfp   = NULL;
  strncpy( data->curr_file_name, fname, MAX_FNAME_LEN );
  strncpy( data->file_name, fname, MAX_FNAME_LEN );
  
  if( aByteOffset )
  {
   fseek( data->fp, aByteOffset, SEEK_SET );
  }

  data->openFilePos = aByteOffset;
  
  pthread_mutex_init( &(data->nextMutex), NULL );
  pthread_cond_init( &(data->nextCond), NULL );

  pthread_mutex_init( &(data->wakeupMutex), NULL );
  pthread_cond_init( &(data->wakeupCond), NULL );

  pthread_mutex_init(&(data->skipMutex),NULL);

  data->next            = 0;
  data->hasNext         = 0;
  data->terminated      = 0;
  data->nextInitialised = 0;

  // Setting methods
  data->iFileProducer            = fileProducer;
  data->iSetFileProducerNextFile = setFileProducerNextFile;
  data->iDecoderProducer         = MadDecoder;
  data->iDestroyFileProducer     = DestroyFileProducer;
  data->iReadFileData            = ReadFileData;
  data->iInitReadFile            = InitReadFile;
  data->iCloseReadFile           = CloseReadFile;
  data->iEOF                     = MP3EOF;
  data->iGetStreamParameters     = GetStreamParameters;
  data->iStartFeatureMode        = StartFeatureMode;
  data->iBytePosToChapter        = BytePosToChapter;
  
  DBG( "MP3 PlugIn initialised, address %p\n", data );
  
  return data;
}


static int setFileProducerNextFile( fileReaderData *fileReader, const char * aName )
{
 assert( NULL != aName );
 assert( NULL != fileReader );

 DBG( "Setting next song to %s...\n", aName );

 pthread_mutex_lock( &( fileReader->nextMutex ) );

 if(fileReader->nextfp!=NULL) fclose((FILE *)(fileReader->nextfp));

 fileReader->nextfp          = fopen( aName, "rb" );
 fileReader->hasNext         = ( NULL != fileReader->nextfp );
 fileReader->nextInitialised = 0;

 strncpy( fileReader->next_file_name, aName, MAX_FNAME_LEN );
 
 DBG("hasNext set to %i, name %s\n",fileReader->hasNext, fileReader->next_file_name);

 pthread_mutex_unlock( &( fileReader->nextMutex ) );

 // Waking up
 if( fileReader->hasNext )
 {
  pthread_mutex_lock(&(fileReader->wakeupMutex));
  pthread_cond_signal( &( fileReader->wakeupCond ) );
  pthread_mutex_unlock(&(fileReader->wakeupMutex));
  
  pthread_mutex_lock(&(fileReader->nextMutex));
  pthread_cond_signal( &( fileReader->nextCond ) );
  pthread_mutex_unlock(&(fileReader->nextMutex));

  DBG( "Success\n" );
 }
 else
 {
  DBG( "Error %s\n", strerror( errno ) );
 }

 return fileReader->hasNext;
}

static int InitNextFile( fileReaderData* data )
{
 assert( NULL != data );

 DBG( "Initialising next file %s\n", data->next_file_name );

 const int tmpResult = ( data->nextInitialised ) ? 0: ( data->iInitReadFile( data, ( FILE* )data->nextfp, data->next_file_name, NULL ) );

 if( !tmpResult )
 {
  data->currfp = ( FILE * )( data->nextfp );
  strncpy( data->file_name, data->curr_file_name, MAX_FNAME_LEN );
  strncpy( data->curr_file_name, data->next_file_name, MAX_FNAME_LEN );
  *data->next_file_name = 0;
  data->nextInitialised = 1;
 }

 return tmpResult;
}

static void UnInitNextFile( fileReaderData* data )
{
 assert( NULL != data );

 if( data->hasNext )
 {
  DBG( "Uninit next file %s\n", data->curr_file_name );

  data->currfp = data->fp;
  strncpy( data->next_file_name, data->curr_file_name, MAX_FNAME_LEN );
  strncpy( data->curr_file_name, data->file_name, MAX_FNAME_LEN ); 
  fseek( ( FILE* )data->nextfp, data->openFilePos, SEEK_SET );
 }
}

static void *fileProducer(void *arg) {
  StreamBufferHandle *h=(StreamBufferHandle *)arg;
  fileReaderData     *data=(fileReaderData *)(h->udata);
  size_t              bufferLen;
  unsigned char      *bufferStart;
  size_t              bytesLeft;
  size_t              result=0;
  void               *rv=NULL;
  int                 skip=0;
  int                 state=0;
  int		      eof=0;
  int                 ec;
  int                 read_retries = 0;

  DBG( "File producer starting, open pos %d...\n", data->openFilePos ); 

  bufferLen   = GetPrimaryBufferLen( h, data );
  bufferStart = GetStreamBufferPtr(h);
  bytesLeft   = bufferLen;
  data->currfp= data->fp;
  
  data->fileSize = GetEncodedDataSize( data, data->currfp );
 ( ( MadOpt *)data->udata )->bytesRead = data->openFilePos;

  state = S_READING;

  while(!StreamBufferKillThread(h)) {
     DBG("state=%i\n",state);
    
    switch(state) {
       
      case S_DISCARD: 
	if(data->discardAndWait) {	// <discard>
	  DBG("discardAndWait\n");
      
      UnInitNextFile( data );

          StreamBufferResetEos( h );
	  DiscardStreamBuffer( h );

	  data->discardAndWait=1;

          pthread_mutex_lock(&(data->resumeMutex));
          while( data->discardAndWait && !StreamBufferKillThread(h) ) {
            pthread_cond_wait(&(data->resumeCond),&(data->resumeMutex));
          }
          pthread_mutex_unlock(&(data->resumeMutex));

	  DBG("Resuming\n");
	  if(StreamBufferKillThread(h)) state=S_EXIT;
	  else {
            bufferLen   = GetPrimaryBufferLen( h, data );
            bufferStart = GetStreamBufferPtr(h);
            bytesLeft   = bufferLen;
	        state       = S_READING;
	  }
	} // </discard>
	break;
    
      case S_READING:
	if(data->discardAndWait) {
          state=S_DISCARD;
        } else {
          DBG("bytesLeft=%i\n",bytesLeft);

          result = data->iReadFileData( data, bufferStart + bufferLen - bytesLeft, bytesLeft );
          DBG("file producer read result %i\n",result); 

          pthread_mutex_lock(&(data->skipMutex));

	  if(data->skip) {
	     data->skip = 0; 
             skip       = 1;
	     state      = S_SKIP;
          } else {
	     bytesLeft -= result;
             if( data->iEOF( data ) ) 
               state = S_EOF;
	     else if( bytesLeft < data->iGetStreamParameters( data )->iMinReadSize ) 
               state = S_FLUSHING;
          }
          pthread_mutex_unlock(&(data->skipMutex));

          if( !result && S_READING == state )
          {
           if( read_retries++ > MAX_READ_0_RETRIES )
           {
            DBG( "Too many read retries, assuming EOF\n" );
            read_retries = 0;            
            state        = S_EOF;
           }
          }
          else
          {
           read_retries = 0;
          }
        }
        break;

      case S_FLUSHING:
	// flush normally
        DBG( "File producer flushing, buf len %d, left %d, type %d\n", bufferLen, bytesLeft, data->iGetStreamParameters( data )->iBufferType ); 

        // Set buffer type
        //SetStreamBufferType( h, data->iIsMP3( data ) ? EMP3Buffer: EPCMBuffer );

	    // flush, but exit if thread kill request set
        if(!data->discardAndWait && FlushStreamBuffer(h, bufferLen - bytesLeft ) ) { 
          if(h->killThread) { // if rv==1 & killThread, exit, otherwise 
  			      // it is interruptFlush so we discard any data
            DBG("flushkill\n"); 
	    state=S_EXIT;
            break; 
          }
        }
	if(data->discardAndWait) {
            state=S_DISCARD;
            DBG( "Switching to Discard state (%d)\n", state );
	    break;
        }
		// reset reading
        bufferLen   = GetPrimaryBufferLen( h, data );
        bufferStart = GetStreamBufferPtr(h);
        bytesLeft   =  bufferLen;

        DBG( "File producer flushed, buflen %d, ptr %p\n", bufferLen, bufferStart );

	state=S_READING;
	if(data->discardAndWait) 
    {
        state=S_DISCARD;
        DBG( "Switching to Discard state (%d)\n", state );
    }

        break;

      case S_EXIT:
	// kill
        DBG( "File producer terminated\n" );
        data->terminated=1;
        pthread_exit(rv);
	break;
    
      case S_EOF:
        DBG( "Entering EOF state...\n" );

        ec=0;
        if(!skip) {
          DBG( "Flushing Stream buffer...\n" );

          if( !data->currentFeature || !data->currentFeatureSpeed )
          { // Do not signal EOS in feature mode
           StreamBufferEos( h );
          }

          //SetStreamBufferType( h, data->iIsMP3( data ) ? EMP3Buffer: EPCMBuffer );
	  // don't flush if we are asked to discard because doing so may cause a race condition!
	  // (note: race condition still exists: thr1:check_dAW ; thr2:set_dAW ; thr1:flush)
	  // (note: low probability race condition, handled by timeouts in main thread)
	  if(data->discardAndWait) {
            ec=1;
          } else {
            ec=FlushStreamBuffer(h,bufferLen-bytesLeft);
          }
	}
        DBG( "Stream buffer flushed...\n" );

        /*if( data->currentFeature && data->currentFeatureSpeed )
        {
         DBG( "Feature mode on, swithing to read state back\n" );

         bufferLen   = GetPrimaryBufferLen( h, data );
         bufferStart = GetStreamBufferPtr( h );
         bytesLeft   =  bufferLen;

         state = S_READING;
         break;
        }*/

        if(ec && h->killThread) {
	  DBG("killThread received\n");
	  state=S_EXIT;
	} else if(ec && data->discardAndWait) {
	  state=S_DISCARD;
       DBG( "Switching to Discard state (%d)\n", state );
        } else if(!data->hasNext) {
	  state=S_IDLE; 
	} else {
		// restart reading next song
          if( !InitNextFile( data ) ) {

            DBG( "Starting reading file %s, initial pos %d\n", data->curr_file_name, data->openFilePos );

            bufferLen   = GetPrimaryBufferLen( h, data );
            bufferStart = GetStreamBufferPtr(h);
            bytesLeft   = bufferLen;
    	    state = S_PREREADING;
	  } else {
	    fprintf(stderr,"error initialising next song\n");
	    state=S_EXIT;
	  }
	}
	skip=0;
	break;

      case S_PREREADING:
          if(data->discardAndWait) {
              state=S_DISCARD;
              DBG( "Switching to Discard state (%d)\n", state );
          }
        else {
	  // read but wait for next after flush
          result = data->iReadFileData( data, bufferStart + bufferLen - bytesLeft, bytesLeft );
          DBG("state 6 file producer read result %i on fp %p\n",result,data->currfp); 

          pthread_mutex_lock(&(data->skipMutex));
	  if(data->skip) {
	    data->skip=0; skip=1;
	    state=S_SKIP;
	  } else {
	    bytesLeft-=result;
            if(data->iEOF(data)) { eof=1; state=S_WAITNEXT; }
	    else if(bytesLeft < data->iGetStreamParameters( data )->iMinReadSize ) { eof=0; state=S_WAITNEXT; }
	  }
          pthread_mutex_unlock(&(data->skipMutex));
        }
	break;
    
      case S_WAITNEXT:
	// wait for next flag (meaning consumer is ready for more data)
        DBG("waiting for next flag\n");
        pthread_mutex_lock(&(data->nextMutex));
        while(!data->next && !data->discardAndWait) {
          pthread_cond_wait(&(data->nextCond),&(data->nextMutex));
        }
	pthread_mutex_unlock(&(data->nextMutex));

    if(data->discardAndWait) {
        state=S_DISCARD;
        DBG( "Switching to Discard state (%d)\n", state );
    }
        else {
          DBG("got next flag\n");
	  state=eof?S_EOF:S_FLUSHINGPRE; // TODO: fix up eof==1 case

          // update data, close old song
          data->iCloseReadFile( data );

          fclose( data->fp );
          data->fp              = NULL;
          data->fileSize        = 0;
          data->hasNext         = 0;
          data->nextInitialised = 0;

          data->fp = ( FILE *)data->nextfp; 
          data->fileSize = GetEncodedDataSize( data, data->fp );
          strcpy( data->file_name, data->curr_file_name );

          ( ( MadOpt *)data->udata )->bytesRead = data->openFilePos;

          data->nextfp   = NULL;        
          data->next     = 0;
          //data->switchedToNext = 1;
	}
	break;
    
      case S_FLUSHINGPRE:
        DBG("file producer flushing in state 8, buffer type %d\n", data->iGetStreamParameters( data )->iBufferType ); 

        //SetStreamBufferType( h, data->iIsMP3( data ) ? EMP3Buffer: EPCMBuffer );

	    // flush, but exit if thread kill request set
        if(!data->discardAndWait && FlushStreamBuffer(h,bufferLen-bytesLeft)) { 
          if(h->killThread) { // if rv==1 & killThread, exit, otherwise 
			    // it is interruptFlush so we discard any data
            DBG("flushkill\n"); 
	    state=S_EXIT;
            break; 
          }
        }
	if(data->discardAndWait) {
	  state=S_DISCARD;
          DBG( "Switching to Discard state (%d)\n", state );
	  break;
        }
        DBG( "File producer flushed in state 8, len %d, ptr %p\n", bufferLen-bytesLeft, bufferStart );

        bufferLen   = GetPrimaryBufferLen( h, data );
        bufferStart = GetStreamBufferPtr(h);
        bytesLeft   = bufferLen;

	    state=S_READING;
        data->switchedToNext = 1;
        data->nextInitialised = 0;
	break;
    
      case S_SKIP:
	// skip (or: do eof pretending nothing has been read)
//        pthread_mutex_lock(&(data->skipMutex));
//	skip=data->skip; data->skip=0;
 //       pthread_mutex_unlock(&(data->skipMutex));
	state=S_EOF;
	break;

      case S_IDLE:
	// idle
          // idle until either kill or next song
        DBG("idling\n");

        data->isIdle = 1;

        pthread_mutex_lock(&(data->wakeupMutex));
        //DBG("Wakeup Mutex locked\n");

        while( !StreamBufferKillThread(h) && !data->hasNext && !data->discardAndWait) {
          pthread_cond_wait(&(data->wakeupCond),&(data->wakeupMutex));
          //DBG("Waiting cond var wakeupCond\n");
        }
        pthread_mutex_unlock(&(data->wakeupMutex));
        //DBG("Wakeup Mutex unlocked\n");

	// unconditionally reset skip flag, since if it is not
        // set, we don't care, and if it is set, we are already
        // skipping to the next song anyway
	pthread_mutex_lock(&(data->skipMutex));
    //DBG("Skip Mutex locked\n");
        data->skip=skip=0;
        pthread_mutex_unlock(&(data->skipMutex));
        //DBG("Skip Mutex unlocked\n");
       
        data->isIdle = 0;

	if(data->discardAndWait) {
	  state=S_DISCARD;
      DBG( "Switching to Discard state (%d)\n", state );
	} else if(StreamBufferKillThread(h)) {
	  state=S_EXIT;
	} else if(data->hasNext) {
		DBG( "Restart reading next song\n" );
          if( !InitNextFile( data ) ) {
            DBG( "Starting reading file %s, initial pos %d\n", data->curr_file_name, data->openFilePos );

            bufferLen   = GetPrimaryBufferLen( h, data );
            bufferStart = GetStreamBufferPtr(h);
            bytesLeft   = bufferLen;
    	    state = S_PREREADING;
	  } else {
	    fprintf(stderr,"error initialising next song\n");
	    state=S_EXIT;
	  }
	}
	break;
      default:
	fprintf(stderr,"OUT OF BOUNDS state %i, DYING\n",state);
	state=S_EXIT;
	break;
    }
  }


  DBG( "File producer terminated\n" );
  data->terminated=1;
  pthread_exit(rv);
}
