/* 
 * File   : FileReader.h
 * Author : Koen Martens <koen.martens@tomtom.com>
 * Desc.  : FileReader data structures and function prototypes.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#ifndef __READFILE_H__
#define __READFILE_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <pthread.h>

#ifndef MIN
#define MIN( a, b ) ( ( ( a ) < ( b ) ) ? ( a ): ( b ) )
#endif

#ifndef MAX
#define MAX( a, b ) ( ( ( a ) > ( b ) ) ? ( a ): ( b ) )
#endif

#define MAX_FNAME_LEN              1024
#define DEFAULT_FEATURE_BYTES_READ ( 32 * 1024 )
#define DEFAULT_FILEBUF_SIZE	   ( 1024 * 1024 )	 // size of the mp3 buffer
#define DEFAULT_MIN_READ_SIZE      1    
    
typedef enum { ENoFeature = 0x00, EFastForwardFeature = 0x01, EFastReverseFeature = 0x02 } filereader_feature_t;
typedef enum { EMP3Buffer = 0x00, EPCMBuffer = 0x01, ECustomEcodedBuffer = 0x03 } primary_cashe_buffer_t;

typedef struct 
{
 unsigned int iSamplingRate;
 int iNumChannels;
 int iBitsPerSample;
 primary_cashe_buffer_t iBufferType;
 int iFrameSize;                   // Compressed frame size in bytes
 int iUncompressedFrameSize;       // Uncompressed frame size in bytes
 unsigned int iBufferSize;         // File reader buffer size
          int iFeatureBufferSize;  // File reader buffer size for feature mode
 unsigned int iMinReadSize;        // File reader number of empty bytes that can be befor buffer flip 
 long         iEncodedDataSize;    // Size of encoded data
} stream_parameters_t;

typedef struct __fileReaderData {
  FILE             *fp;	        // current song
  volatile FILE    *nextfp;	    // next song (NULL -> no next song)
  FILE             *currfp;     // real current song (for prereading)
  char              next_file_name[ MAX_FNAME_LEN ];
  char              curr_file_name[ MAX_FNAME_LEN ];
  char              file_name[ MAX_FNAME_LEN ];
  pthread_mutex_t   nextMutex;	// to let reader wait for madproducer ..
  pthread_cond_t    nextCond;	//  .. to ask for next file
  volatile int      next;	    // and the actual flag
  volatile int      hasNext;
  volatile int      nextInitialised; // Set to 1 when next file is initialized
  pthread_mutex_t   wakeupMutex;// to wakeup reader for killthread 
  pthread_cond_t    wakeupCond;	//  .. or next mp3
  volatile int      terminated; // set to 1 when thread terminates
  volatile unsigned int fileSize;   // The size of current file
  volatile int      skip;
  pthread_mutex_t   skipMutex;
  volatile void*    udata;      // Holds MAD opt, ugly, but life is hard...
  volatile int      switchedToNext;
  volatile int      discardAndWait; // 1=discard everything and wait for resumeCond;
  pthread_mutex_t   resumeMutex;
  pthread_cond_t    resumeCond;
  volatile unsigned int openFilePos;   // Position in file while opening
  volatile int      isIdle;

  // Fast forward/reverse
  volatile filereader_feature_t currentFeature;
  volatile int                  currentFeatureSpeed;
  volatile int                  startFeature;
  
  // Internal data
  void*             internalData;  // Differs for different file readers
  void*             internalData2; // Differs for different file readers

  // Methods
  int    ( *iSetFileProducerNextFile )( struct __fileReaderData *fileReader, const char * ); // Set next file
  void*  ( *iFileProducer )( void * );                                                       // File reader thread func
  void*  ( *iDecoderProducer )( void * );                                                    // Decoder thread func

  // Reading support
  int    ( *iInitReadFile )( struct __fileReaderData *fileReader, FILE* aFile, const char* aFileName, char* aErrorText ); // Returns 0 in case of success, initializes internal structures
  void   ( *iCloseReadFile )( struct __fileReaderData *fileReader ); // Should deinitialize internal structures but not close file 
  int    ( *iReadFileData )( struct __fileReaderData *fileReader, void* aBuffer, int aBufferLen ); // Encoded data reader function
  int    ( *iEOF )( struct __fileReaderData *fileReader ); 

  // decoding support, returns number of decoded bytes
  int    ( *iDecodeFrame )( struct __fileReaderData *fileReader, const void* aBuffer, unsigned long aInputLen, void* aOutbuffer ); 

  // Stream parameters, do not destroy pointer
  const stream_parameters_t*   ( *iGetStreamParameters )( struct __fileReaderData *fileReader );

  // Advanced navigation
 int ( *iStartFeatureMode )( struct __fileReaderData *fileReader, int aMode, int aSpeed, char* aErrorText );
 int ( *iGoToChapter )( struct __fileReaderData *fileReader, int aChapterNumber, int aAbsolute, char* aErrorText );
 int ( *iBytePosToChapter )( struct __fileReaderData *fileReader, unsigned long aBytePos, unsigned long* aChapterNumber, unsigned long* aTimeMS );

  // Destruction
  void  ( *iDestroyFileProducer )( struct __fileReaderData *fileReader );
  
} fileReaderData;

// Shoulr return 1 in case of success
typedef int  ( *InitPlugInFunc )( void );

fileReaderData* InitMP3FileProducer( const char *aFileName, unsigned long aByteOffset );
fileReaderData* InitPlugInFileProducer( const char * aFileName, const char* aPlugInName, unsigned long aByteOffset, char* aErrorText );

// Internal plugin data
typedef struct
{
 void* iDLLHandle;
 
 void  ( *iDestroyOldFileProducer )( struct __fileReaderData *fileReader );
 void  ( *iDestroyNewFileProducer )( struct __fileReaderData *fileReader );

 const stream_parameters_t* ( *iOldGetStreamParameters )( struct __fileReaderData *fileReader );

} file_reader_plugin_internal_data_t;

#ifdef __cplusplus
}
#endif

#endif //  __READFILE_H__

/* EOF */
