/* 
 * File   : TomTomProjects/AudioStreaming/StreamBuffer.c
 * Author : Koen Martens <koen.martens@tomtom.com>
 * Desc.  : mp3 decoder with synchronised output
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <pthread.h>
#include <assert.h>
#include <stdarg.h>
#include "../AudioAccept.h"
#include "StreamBuffer.h"
#include "Debug.h"

#define MIN( a, b ) ( ( a ) < ( b ) ) ? ( a ): ( b );

inline int GetStreamBufferIndex(StreamBufferHandle *h) {
  return 1 - h->currBuff;
} 
	// return ptr to producer buffer
inline unsigned char *GetStreamBufferPtr(StreamBufferHandle *h) { 
  return h->bufferStart[1 - h->currBuff];
}

inline unsigned char *GetStreamBufferEnd(StreamBufferHandle *h) {
  return h->bufferStart[1 - h->currBuff] + h->bufferSize;
}

inline int GetStreamBufferLength(StreamBufferHandle *h) {
  return h->bufferSize;
}

inline int GetStreamBufferAvailable(StreamBufferHandle *h)
{
 return h->available[ GetStreamBufferIndex( h ) ];
}

inline void *GetStreamBufferUdata(StreamBufferHandle *h) {
  return h->udata;
}

inline int StreamBufferKillThread(StreamBufferHandle *h) {
  return h->killThread;
}

inline void StreamBufferEos(StreamBufferHandle *h) {
  h->eos[1-(h->currBuff)]=1;
}

inline int  StreamBufferIsEos( StreamBufferHandle *h )
{
 return h->eos[ 0 ] || h->eos[ 1 ];
}

void StreamBufferResetEos(StreamBufferHandle *h) {
  h->eos[0]=h->eos[1]=0;
}

int StreamBufferWaitFlush(StreamBufferHandle *h) {
  assert( NULL != h );

  if( h->fake )
  {
   //usleep( 1000 ); No sleep !!!
  }
  else
  {
   pthread_mutex_lock(&(h->flushMutex)); 
   while( !(h->flush) && !(h->killThread) ) {
     pthread_cond_wait(&(h->flushCond),&(h->flushMutex));
   }
   if(h->flush) h->flush=0;
   pthread_mutex_unlock(&(h->flushMutex)); 
  };
   
  return h->killThread;  
}

 // returns 0 if ok, 1 if thread kill requested
 // blocks until reader performed flip
int FlushStreamBufferInternal(StreamBufferHandle *h, int size, int sync) {
  int rv=0;

  h->available[1-(h->currBuff)]=size;
  h->offset[1-(h->currBuff)]=0;

  pthread_mutex_lock(&(h->flipMutex));

  assert(h->flip==0);

  if(sync) h->flip=1;

	// wake up any waiting thread
  pthread_mutex_lock(&(h->flushMutex));
  h->flush=1;
  pthread_cond_broadcast(&(h->flushCond)); // Serg ?
  pthread_mutex_unlock(&(h->flushMutex));

  
  if(h->debug) fprintf(stderr,"%s: waiting on %p\n",__func__,h);
  while(h->flip && rv==0 && sync) {
    pthread_cond_wait(&(h->flipCond),&(h->flipMutex));
    if(h->killThread) rv=1;
    if(h->interruptFlush) { rv=1; h->interruptFlush=0; }
  }
  if(h->debug) fprintf(stderr,"%s: got back from %p\n",__func__,h);

  pthread_mutex_unlock(&(h->flipMutex));
  
  return rv;
}


int FlushStreamBuffer(StreamBufferHandle *h, int size) {
  return FlushStreamBufferInternal(h,size,1);
}

int FlushStreamBufferAsync(StreamBufferHandle *h, int size) {
  return FlushStreamBufferInternal(h,size,0);
}

void UnflushStreamBuffer(StreamBufferHandle *h) {
  pthread_mutex_lock(&(h->flipMutex));
  h->flip=0;
  pthread_mutex_unlock(&(h->flipMutex));
}


StreamBufferHandle *InitStreamBuffer_internal( size_t bufferSize, void *(*producer)(void *), void *arg, int options, void *buffer) 
{
  StreamBufferHandle *handle;

  handle=malloc(sizeof(StreamBufferHandle));

  if(!handle) return NULL;

  memset( handle, 0,  sizeof( StreamBufferHandle ) );
  
  handle->udata = arg;

  if( BO_IFAKE & options )
  { // Fake buffer
    handle->fake = 1;
    handle->FakeReadFunction = (int ( * )( void**ptr, int len ))producer;	// Serhiy, PLEASEFIX
    return handle;
  }

  if(buffer==NULL) { 
    handle->buffer=malloc(bufferSize*2);
  } else {
    handle->buffer=buffer;
  }

  if(!handle->buffer) { free(handle); return NULL; }

  handle->bufferStart[0] = handle->buffer;
  handle->bufferStart[1] = handle->buffer + bufferSize;

  handle->available[0] = handle->available[1] = 0;
  handle->offset[0] = handle->offset[1] = 0;
  handle->eos[0] = handle->eos[1] = 0;

  handle->bufferSize = bufferSize;

  handle->flip           = 0;
  handle->flush          = 0;
  handle->currBuff       = 0;
  handle->killThread     = 0;
  handle->interruptFlush = 0;
  handle->bufferFreeSize = bufferSize;

  handle->debug=0;

  pthread_mutex_init(&(handle->flipMutex),NULL);
  pthread_cond_init(&(handle->flipCond),NULL);

  pthread_mutex_init(&(handle->flushMutex),NULL);
  pthread_cond_init(&(handle->flushCond),NULL);

  pthread_mutex_init(&(handle->flushMutex),NULL);

  pthread_attr_init(&(handle->thread_attr));
  pthread_attr_setschedpolicy(&(handle->thread_attr),SCHED_RR);

  if( NULL != producer )
  {
   //if(pthread_create( &(handle->thread) , &(handle->thread_attr) , producer , (void*)handle)) {
   if(pthread_create( &(handle->thread) , NULL , producer , (void*)handle)) {
    free(handle->buffer);
    free(handle);
    return NULL;
   }

	// block until data available requested?
   if(options & BO_IBLK) {
    pthread_mutex_lock(&(handle->flipMutex));
    while(!handle->flip) {
      pthread_mutex_unlock(&(handle->flipMutex));
      usleep(1000);
      pthread_mutex_lock(&(handle->flipMutex));
    }
	// signal thread to continue producing
    pthread_cond_broadcast(&(handle->flipCond));
    pthread_mutex_unlock(&(handle->flipMutex));
   }
  }

  handle->external_buffer = ( options & BO_EXTERNALBUFFER ) ? 1: 0;

  return handle;
}


//StreamBufferHandle *InitStreamBuffer( size_t bufferSize, void *(*producer)(void *), void *arg, int options) 
// initialize, but use provided memory area for buffer
StreamBufferHandle *InitStreamBufferSet( size_t bufferSize, void *(*producer)(void *), 
						void *arg, int options, void *buffer) {
  return InitStreamBuffer_internal( bufferSize, producer, arg, options | BO_EXTERNALBUFFER, buffer );
}

StreamBufferHandle *InitStreamBuffer( size_t bufferSize, void *(*producer)(void *), 
						void *arg, int options) {
  return InitStreamBuffer_internal(bufferSize,producer,arg,options,NULL);
}

// note, can not ask for new read buffer after returning data from
// current buffer. 
// returns: >=0 -> bytes read (0 means try again)
//          -1  -> end of stream
int ReadStreamBuffer(StreamBufferHandle *h,void* *ptr,int len) 
{
  assert( NULL != h );

  if( h->fake )
  {
   assert( NULL != h->FakeReadFunction );

   return h->FakeReadFunction( ptr, len );
  }

  int howmany;

  *ptr=NULL;

  //fprintf(stderr,"%s, %p: acquiring lock\n",__func__,h);
  pthread_mutex_lock(&(h->flipMutex));
  //fprintf(stderr,"%s, %p: acquired lock\n",__func__,h);

  if( h->offset[h->currBuff] < h->available[h->currBuff]) {
	// determine how much bytes to give
    howmany = h->available[h->currBuff] - h->offset[h->currBuff];
    if(len<howmany) howmany=len;

	// set pointer
    *ptr=(void*)(h->bufferStart[h->currBuff]+h->offset[h->currBuff]);
    h->offset[h->currBuff]+=howmany;
  } else {
    if(h->eos[h->currBuff]) {
      howmany=-1;
    } else {
      if(h->flip) {
	// flip
        h->currBuff=1-(h->currBuff);
	h->flip=0;
	// set flipped flag
        pthread_mutex_lock(&(h->flippedMutex));
        h->flipped=1;
        pthread_mutex_unlock(&(h->flippedMutex));
        // signal waiting threads
	if(h->debug) fprintf(stderr,"%s: sending broadcast on %p\n",__func__,h); fflush(stderr);
        pthread_cond_broadcast(&(h->flipCond));
	if(h->debug) fprintf(stderr,"%s: sent broadcast on %p\n",__func__,h); fflush(stderr);
	// do read
        if( h->offset[ h->currBuff ] < h->available[ h->currBuff ] ) {
	    // determine how much bytes to give
          howmany=h->available[h->currBuff] - h->offset[h->currBuff];
          if(len<howmany) howmany=len;

	    // set pointer
          *ptr=(void*)(h->bufferStart[h->currBuff] + h->offset[h->currBuff]);
          h->offset[h->currBuff]+=howmany;
        } else {
          if(h->eos[h->currBuff]) howmany=-1; else howmany=0;
        }
      } else {
	howmany=0;
      }
    }
  }
  //fprintf(stderr,"%s, %p: releasing lock\n",__func__,h); fflush(stderr);
  pthread_mutex_unlock(&(h->flipMutex));
  //fprintf(stderr,"%s, %p: released lock\n",__func__,h); fflush(stderr);

  return howmany;
}

int FlippedStreamBuffer(StreamBufferHandle *h) {
  int rv;

  pthread_mutex_lock(&(h->flippedMutex));
  rv=h->flipped;
  h->flipped=0;
  pthread_mutex_unlock(&(h->flippedMutex));

  return rv;
}

// discards what the reader currently has as current buffer
void DiscardStreamBuffer(StreamBufferHandle *h)
{
  h->offset[h->currBuff]=h->available[h->currBuff]=0;
}

void DiscardStreamBufferFull( StreamBufferHandle *h )
{
 assert( NULL != h );

 // discards everything in sream buffer
 pthread_mutex_lock(&(h->flipMutex));

 h->offset[ 0 ] = h->available[ 0 ] = 0;
 h->offset[ 1 ] = h->available[ 1 ] = 0;

 pthread_mutex_unlock(&(h->flipMutex));
}

// cancels the flush operation
void InterruptFlushStreamBuffer(StreamBufferHandle *h) 
{
  pthread_mutex_lock(&(h->flipMutex));
  h->flip=0;
  h->interruptFlush=1;
  pthread_cond_broadcast(&(h->flipCond));
  pthread_mutex_unlock(&(h->flipMutex));
}

static void FreeStreamBufferHandle( StreamBufferHandle *h )
{
  DBG("Freeing handle for buffer %p\n",h );

  assert( NULL != h );

  free(h);
}

void DoneStreamBuffer( StreamBufferHandle *h ) 
{
  if( !h->fake )  
  {
   DoneStreamBufferWakeUp(h,0,0);
  }
  else
  {
   FreeStreamBufferHandle( h );
  }
}

void DoneStreamBufferWithAudioAccept( StreamBufferHandle *h )
{
  if( !h->fake )  
  {
   DoneStreamBufferWakeUp(h,1,0);
  }
  else
  {
   FreeStreamBufferHandle( h );
  }
}


void DoneStreamBufferWakeUp(StreamBufferHandle *h, int wakeAudioAccept, int wakeUps, ...)
{
  va_list args;
  int i;
  pthread_mutex_t *mutex[wakeUps];
  pthread_cond_t  *cond[wakeUps];

  DBG("killing thread for buffer %p\n",h);

  assert( NULL != h );

  if( h->fake )
  {
   DoneStreamBuffer( h );
   return;
  }

  // set boolean
  h->killThread = 1;

  va_start(args,wakeUps);
  for(i=0;i<wakeUps;i++) {
    mutex[i]=va_arg(args,pthread_mutex_t *);
    cond[i]=va_arg(args,pthread_cond_t *);
  }

  if(wakeAudioAccept) AudioAcceptNotifyConditional();

	// lock mutexes
  pthread_mutex_lock(&(h->flipMutex));
  pthread_mutex_lock(&(h->flushMutex));
  for(i=0;i<wakeUps;i++) pthread_mutex_lock(mutex[i]);

  // 
  pthread_cond_broadcast(&(h->flipCond));
  pthread_cond_broadcast(&(h->flushCond));
  for(i=0;i<wakeUps;i++) pthread_cond_broadcast(cond[i]);

  for(i=0;i<wakeUps;i++) pthread_mutex_unlock(mutex[i]);
  pthread_mutex_unlock(&(h->flushMutex));
  pthread_mutex_unlock(&(h->flipMutex));  

  DBG( "Interrupting stream buffer flush...\n" );
  InterruptFlushStreamBuffer( h );

  // wait for thread completion
  if( h->thread )
  {
   DBG("joining thread for buffer %p, kill thread %d\n",h, h->killThread);
   pthread_join(h->thread,0);
  }

  if( !h->external_buffer )
  {
   DBG( "Freeing underlaying buffer\n" );
   free( h->buffer );
  }
  else
  {
   DBG( "Buffer managed externally, not deleting here!!!\n" );
  }

  FreeStreamBufferHandle( h );
 
  return;
}

int WriteStreamBuffer( StreamBufferHandle *h,void* ptr, int len )
{
 assert( NULL != h );
 assert( NULL != ptr );

/* static FILE* tmpFile = NULL;

 if( NULL == tmpFile )
 {
  tmpFile = fopen( "pcm1.cdr", "w+b" );
  assert( NULL != tmpFile );
 }*/

 int tmpResult = -1;
       
/* if( !StreamBufferKillThread( h ) )
 {
  const size_t         bufferLen   = GetStreamBufferLength( h );
  const int            writeCount  = MIN( h->bufferFreeSize, len );         
  const size_t         writeOffset = bufferLen - h->bufferFreeSize;
  unsigned char*       bufferStart = GetStreamBufferPtr( h );

  //fwrite( ptr, writeCount, 1, tmpFile );
  
  memcpy( bufferStart + writeOffset, ptr, writeCount );

  h->bufferFreeSize -= writeCount;

  if( !h->bufferFreeSize )
  {
   FlushStreamBuffer( h, bufferLen );
   h->bufferFreeSize = bufferLen; 
  }

  //DBG( "len %d, wc %d\n", len, writeCount );
  
  tmpResult = writeCount;
 }*/

 if( !StreamBufferKillThread( h ) )
 {
  const int writeCount  = MIN( GetStreamBufferLength( h ), len );         

  memcpy( GetStreamBufferPtr( h ), ptr, writeCount );

  FlushStreamBuffer( h, writeCount );

  tmpResult = writeCount;
 }

 return tmpResult;
}

/* EOF */
