#include <assert.h>

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/time.h> 
#include <sys/poll.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/l2cap.h>

#include <netinet/in.h>

#include "Packager.h"
#include "A2dp.h"
#include "A2DPListen.h"
#include "Debug.h"

#define ACCEPT_TIMEOUT 100

extern void SetRoleSwitch( a2dp_packager_t *a2dp, int aEnable );

static void ExitListen( a2dp_listen_t* aA2DP, int aExitCode )
{
 assert( NULL != aA2DP );

 aA2DP->iResult     = ( aA2DP->iStopListen ) ? EErrorInterrupted: aExitCode;
 aA2DP->iStopListen = 1;
 aA2DP->RemoteConnected( aA2DP );

 pthread_exit( 0 );
 assert( 0 ); // Should never get here
}

void* ListenRemoteBTDevice( void* aListenStruct )
{
 a2dp_listen_t* tmpA2DP = ( a2dp_listen_t* )aListenStruct;
 bdaddr_t bdaddr;
 int sock;

 assert( NULL != tmpA2DP );
 
 if( tmpA2DP->iStopListen )
 {
  DBG( "Exit signalled, exiting...\n" );
  ExitListen( tmpA2DP, EErrorInterrupted );
 }

 assert( NULL != tmpA2DP->RemoteConnected ); 
 assert( !tmpA2DP->iFile );
 assert( tmpA2DP->iChannel );

 DBG( "Addr %s, PSM %d\n", tmpA2DP->iBTAddr, tmpA2DP->iChannel );
 
 //Translate bluetooth address
 if( strlen( tmpA2DP->iBTAddr ) )
  str2ba(tmpA2DP->iBTAddr, &bdaddr);
 else
  bacpy(&bdaddr, BDADDR_ANY);
 

 // Setup a socket in listen mode
 {
  bdaddr_t local;
  struct sockaddr_l2 addr;

  if( ( ( sock = socket( PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP ) ) < 0 ) || tmpA2DP->iStopListen )
  {
   DBG( "Can not create socket, stop state %d\n", tmpA2DP->iStopListen );

   ExitListen( tmpA2DP, EErrorFailed );
  }

  memset(&addr, 0, sizeof(addr));
  addr.l2_family = AF_BLUETOOTH;
  hci_devba(0, &local);
  bacpy(&addr.l2_bdaddr, &local);
  addr.l2_psm = htobs( tmpA2DP->iChannel );

  if( ( bind( sock, ( struct sockaddr *)&addr, sizeof( addr ) ) < 0 ) || tmpA2DP->iStopListen )
  {
   DBG( "Can not bind socket, stop state %d\n", tmpA2DP->iStopListen );

   close( sock );

   ExitListen( tmpA2DP, EErrorFailed );  
  }

  if( ( listen( sock, 1 ) < 0 ) || tmpA2DP->iStopListen )
  {
   DBG( "Can not listen socket, stop state %d\n", tmpA2DP->iStopListen );
   
   close(sock);

   ExitListen( tmpA2DP, EErrorFailed );
  }
 }

 DBG( "LISTEN L2CAP socked configured, entering Listen loop...\n" );
 
 // Accept incoming connection
 while( !tmpA2DP->iStopListen )
 {
  fd_set rset;
  struct timeval tv;
  socklen_t socklen;
  struct sockaddr_l2 addr;
  int res;

  tv.tv_sec = 0;
  tv.tv_usec = ACCEPT_TIMEOUT;
  FD_ZERO(&rset);
  FD_SET(sock, &rset);
  res = select(sock+ 1, &rset, NULL, NULL, &tv);
  if (res < 0 || res > 1)
  {
   DBG( "Select failed\n" );

   tmpA2DP->iResult = EErrorFailed;
   close(sock);
   tmpA2DP->iStopListen = 1;
   tmpA2DP->RemoteConnected( tmpA2DP );

   pthread_exit( 0 );
  }
  else if (res == 0)
   continue;

  SetRoleSwitch( ( a2dp_packager_t* )( tmpA2DP->iUserData ), 1 );

  memset(&addr, 0, sizeof(addr));
  socklen = sizeof(addr);
  res = accept(sock, (struct sockaddr *) &addr, &socklen);
  if( res < 0 )
  {
   DBG( "Can not accept\n" );

   close(sock);
   ExitListen( tmpA2DP, EErrorFailed );
  }

  /* check if the connect is coming from the right bt device */
  if( bacmp( &bdaddr, BDADDR_ANY ) && bacmp( &bdaddr, &addr.l2_bdaddr ) )
  {
   DBG( "Wrong device connected, refusing\n" );

   close(res);
   SetRoleSwitch( ( a2dp_packager_t* )( tmpA2DP->iUserData ), 0 );
   continue;
  }

  bacpy(&bdaddr, &addr.l2_bdaddr);
  close(sock);
  sock = res;
  break;
 }

 if( tmpA2DP->iStopListen )
 {  
  close(sock);
  ExitListen( tmpA2DP, EErrorInterrupted );
 }

 // If connected, set FD, ADDR, call callback and exit
 tmpA2DP->iFile = sock;
 //if( !strlen(tmpA2DP->iBTAddr) )
 {
   ba2str( &bdaddr, tmpA2DP->iBTAddr);
 }

 DBG( "Connection succeeded\n" );
 ExitListen( tmpA2DP, EErrorNone );
 pthread_exit( 0 );
}

int StartListen(a2dp_listen_t *aListenStruct)
{
 assert( NULL != aListenStruct );

 aListenStruct->iResult     = EErrorNone;
 aListenStruct->iStopListen = 0;

 if (pthread_create(&aListenStruct->iThread, NULL, ListenRemoteBTDevice, aListenStruct) == 0)
 {
  pthread_detach(aListenStruct->iThread);
  return EErrorNone;
 }
 return EErrorFailed;
}

void StopListen(a2dp_listen_t *aListenStruct)
{
 assert( NULL != aListenStruct );

 if( !aListenStruct->iStopListen && EErrorNone == aListenStruct->iResult )
 {
  int tmpRetry = 50;

  aListenStruct->iStopListen = 1;
  while( EErrorNone == aListenStruct->iResult && tmpRetry-- )
  {
   DBG( "Waiting for listen to stop, retry %d\n", tmpRetry );

   usleep( 10000 );
  }
 }
 else
 {
  DBG( "Listening not started\n" );
 }
}
