/* 
 * File   : A2dp.c
 * Author : Koen Martens <koen.martens@tomtom.com>
 * Desc.  : A2dp protocol implementation, packs frames (calling the 
 *          encoder) into a StreamBuffer, consuming PCM/MAD_FIX from
 *          incoming StreamBuffer. Largely based on bluez.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */  

#include <assert.h>
#include <stdio.h>
#include <errno.h> 
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>
#include <assert.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/soundcard.h>
#include <sys/time.h> 
#include <sys/poll.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/l2cap.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#include <netinet/in.h>

#include "Debug.h"
#include "Encoder.h"
#include "Packager.h"
#include "A2dpTypes.h"
#include "A2dp.h"
#include "A2DPErrors.h"
#include "A2DPListen.h"

// Uncomment this to support BlueWalker
//#define ENABLE_ROLE_SWITCH

// Uncomment to enable fragmented packet read/write
#define ENABLE_PACKET_FRAGMENTATION

static char* iErrorText          = NULL;

static void InitError( char* aData )
{
 assert( NULL != aData );

 iErrorText  = aData;
 *iErrorText = 0;
}

static void AppendError( const char* aErrorText, int aAddErrNo )
{
 assert( NULL != aErrorText );
 assert( NULL != iErrorText );

 DBG( "%s\n", aErrorText );

 strcat( iErrorText, aErrorText );

 if( aAddErrNo )
 {
  char tmpErr[ 200 ];

  sprintf( tmpErr, " err %d, %s\n", errno, strerror( errno ) );
  strcat( iErrorText, tmpErr );
  
  DBG( "%s", tmpErr );
 }
}

void SetRoleSwitch( a2dp_packager_t *a2dp, int aEnable );
//static int A2DPSwitchRole( packager_t* p, int aRole );

static void init_request(packager_t *packager,struct avdtp_header * header, int request_id)
{       
        packager_t *p=(packager_t *)packager;
        a2dp_packager_t *h=(a2dp_packager_t *)(p->udata);

        header->packet_type = PACKET_TYPE_SINGLE;
        header->message_type = MESSAGE_TYPE_COMMAND;
        header->transaction_label = h->transaction;
        header->signal_id = request_id;
        
        // clear rfa bits
        header->rfa0 = 0;
        
        h->transaction = (h->transaction + 1) & 0xf;
}

#ifdef ENABLE_PACKET_FRAGMENTATION

static int SendAVDTPSignalPacket( int mtu, int aFile, char* aBuffer, int aLength )
{
 int tmpResult = 0;
 
 assert( NULL != aBuffer );
 assert( mtu > 1 );
 assert( aLength > 1 );

 errno = 0;

 if( mtu >= aLength )
 {
  /* Packet fits in MTU, send it */
  tmpResult = write( aFile, aBuffer, aLength );
 }
 else
 {
  const int            fragments  = ( aLength + mtu - 2 ) / ( mtu - 1 );
  uint8_t*             ptr        = malloc( aLength + ( fragments - 1 ) * sizeof( struct avdtp_header ) + sizeof( uint8_t ) );  
  struct avdtp_fragment_start_header* header     = ( struct avdtp_fragment_start_header* )ptr; 
  uint8_t*             cur_ptr    = ptr;
  const int            tmpPrevLen = aLength;
  int                  i;

  DBG( "MTU %d, Len %d, Fragments %d\n", mtu, aLength, fragments );

  assert( NULL != ptr );

  /* Copy the data */
  memcpy( ptr, aBuffer, aLength );
  
  /* Make room for number of packets */
  memmove( ptr + sizeof( struct avdtp_fragment_header ) + sizeof( uint8_t ),       
           ptr + sizeof( struct avdtp_fragment_header ), 
           aLength - sizeof( struct avdtp_fragment_header ) );
        
  /* Set number of fragments and packet type */
  ptr[ 1 ]            = fragments;  
  header->packet_type = PACKET_TYPE_START;

  DBG( "START packet label %d, type %d, message %d, signal 0x%x, fragments %d\n", 
       header->transaction_label, 
       header->packet_type,
       header->message_type,
       header->signal_id,
       header->number_of_fragments );

  /* Write first packet */
  tmpResult = write( aFile, header, mtu );  

  /* Adjust remaining length and data */
  aLength -= ( mtu - sizeof( uint8_t ) );
  cur_ptr += mtu;
  memmove( ptr + sizeof( struct avdtp_fragment_header ), cur_ptr, aLength );

  DBG( "START packet sent, total bytes %d, left %d\n", tmpResult, aLength );  

  /* Continuation Packets */
  for( i = 0; i < fragments - 2; i++ )
   {
    header->packet_type = PACKET_TYPE_CONTINUE;

    tmpResult += write( aFile, header, mtu );    

    /* Adjust remaining length and data */
    aLength -= ( mtu - sizeof( struct avdtp_fragment_header ) );
    cur_ptr += mtu;
    memmove( ptr + sizeof( struct avdtp_fragment_header ), cur_ptr, aLength );

    DBG( "CONT packet sent, total bytes %d, left %d\n", tmpResult, aLength );  
  }

  /* End Packet */
  header->packet_type = PACKET_TYPE_END;
  tmpResult += write( aFile, header, aLength + 1 );

  DBG( "END packet sent, total bytes written %d\n", tmpResult );  

  if( tmpPrevLen + 
       ( fragments - 1 ) * ( int )sizeof( struct avdtp_fragment_header ) + 
       ( int )sizeof( uint8_t ) == 
      tmpResult )
  { // ALL OK 
   tmpResult = tmpPrevLen;
  }
  else
  {
   DBG( "Write count does not match\n" );
  }

  free( ptr );
 }

 return tmpResult;
}

#define REASSEMBLE_MORE_DATA 0
#define REASSEMBLE_DISCARD   -1

static int ReassembleAVDTPSignalPacket( a2dp_packager_t* a2dp, struct avdtp_header* aHeader, int aMinBytesNeed )
{
 assert( NULL != a2dp );
 assert( a2dp->signalling_mtu > 1 );
 assert( NULL != aHeader );
 assert( aMinBytesNeed > 0 );

 int tmpResult = REASSEMBLE_MORE_DATA;
 errno = 0;

 // Reading packet header
 if( ( tmpResult = read( a2dp->cmdfd, 
                         aHeader, 
                         /*aMinBytesNeed*/ a2dp->signalling_mtu ) ) < ( int )sizeof( struct avdtp_header ) )
 {
  //DBG( "Can not read AVDTP packet header, res %d, err %d, %s\n", tmpResult, errno, strerror( errno ) );

  return REASSEMBLE_DISCARD;
 }

 DBG( "MTU %d, MinBytes %d, Read %d\n", a2dp->signalling_mtu, aMinBytesNeed, tmpResult );

 switch( aHeader->packet_type )
 {
   case PACKET_TYPE_SINGLE:
    DBG( "SINGLE packet, label %d, type %d, signal 0x%x\n", aHeader->transaction_label, aHeader->message_type, aHeader->signal_id );

#ifdef DEBUG
    if( a2dp->reassembled_packet_size )
    {
     DBG( "Prev Fragmented packet not finished, discarding\n" );
    }
#endif

    a2dp->reassembled_packet_size = 0; // Discarding current assembling packet

    // Reading the rest
/*    tmpResult = sizeof( struct avdtp_header ) + 
                read( a2dp->cmdfd, 
                      ( ( uint8_t *)aHeader ) + sizeof( struct avdtp_header ), 
                      aMinBytesNeed - sizeof( struct avdtp_header ) );

    DBG( "Read %d bytes\n", tmpResult );*/
    break;

   case PACKET_TYPE_START:
   {    
    int tmpReadCount = 0;
    int tmpNeedCount = a2dp->signalling_mtu - tmpResult;

    if( tmpResult < ( int )( sizeof( struct avdtp_fragment_start_header ) ) )
    {
     DBG( "Too small START packet, discarding\n" );
     memset( aHeader, 0, sizeof( *aHeader ) );
     tmpResult = REASSEMBLE_DISCARD;
     break;
    }

    DBG( "START packet, label %d, type %d, num packets %d, signal %d\n", 
         aHeader->transaction_label, 
         aHeader->message_type, 
         ( ( struct avdtp_fragment_start_header * )aHeader )->number_of_fragments, 
         ( ( struct avdtp_fragment_start_header * )aHeader )->signal_id );

#ifdef DEBUG
    if( a2dp->reassembled_packet_size )
    {
     DBG( "Prev Fragmented packet not finished, discarding\n" );
    }
#endif

    a2dp->reassembled_packet_size = 0; // Discarding current assembling packet                

    // Copying read header to reassemble buffer
    memcpy( a2dp->reassembled_packet, aHeader, sizeof( struct avdtp_fragment_header ) );

    // Copy rest except of number of packets
    memcpy( a2dp->reassembled_packet + sizeof( struct avdtp_fragment_header ), 
            ( ( uint8_t* )aHeader ) + sizeof( struct avdtp_header ),
            tmpResult - sizeof( uint8_t ) - sizeof( struct avdtp_fragment_header ) );

    a2dp->reassembled_packet_size = tmpResult - sizeof( uint8_t );

    while( tmpNeedCount && tmpReadCount >= 0 )
    {
     // Read the rest
     tmpReadCount = read( a2dp->cmdfd, 
                          a2dp->reassembled_packet + a2dp->reassembled_packet_size,
                          tmpNeedCount );
    
     DBG( "Read %d bytes extra, need %d\n", tmpReadCount, tmpNeedCount );

     tmpNeedCount                  -= tmpReadCount;
     a2dp->reassembled_packet_size += tmpReadCount;
    }

    if( tmpNeedCount || tmpReadCount < 0 )
    {
     DBG( "Can not read rest of START packet, discarding\n" );

     memset( aHeader, 0, sizeof( *aHeader ) );
     tmpResult = REASSEMBLE_DISCARD;
     break;
    }

    DBG( "START packet, label %d, type %d, signal %d\n", 
         aHeader->transaction_label, 
         aHeader->message_type, 
         aHeader->signal_id );

    a2dp->reassembled_packet_size = a2dp->signalling_mtu - sizeof( uint8_t );    
    
    // Recursively reading data
    tmpResult = ReassembleAVDTPSignalPacket( a2dp, aHeader, a2dp->signalling_mtu );
   }
   break;

   case PACKET_TYPE_END:
   case PACKET_TYPE_CONTINUE:
   {
    DBG( "CONT or END packet detected, label %d, type %d\n", aHeader->transaction_label, aHeader->message_type );
    
    if( !a2dp->reassembled_packet_size )
    {
     DBG( "No START of fragmented packet received, discarding\n" );

     memset( aHeader, 0, sizeof( *aHeader ) );
     tmpResult = REASSEMBLE_DISCARD;
     break;
    }

    if( aHeader->transaction_label != ( ( struct avdtp_fragment_header* )( a2dp->reassembled_packet ) )->transaction_label )
    {
     DBG( "Transaction label does not match, discarding all\n" );

     a2dp->reassembled_packet_size = 0;
     memset( aHeader, 0, sizeof( *aHeader ) );
     tmpResult = REASSEMBLE_DISCARD;
     break;
    }

    if( tmpResult != a2dp->signalling_mtu && PACKET_TYPE_CONTINUE == aHeader->packet_type )
    {
     DBG( "Wrong CONT packet size, discarding\n" );

     a2dp->reassembled_packet_size = 0;
     memset( aHeader, 0, sizeof( *aHeader ) );
     tmpResult = REASSEMBLE_DISCARD;
     break;
    }

    // Copy the rest, except header
    memcpy( a2dp->reassembled_packet + a2dp->reassembled_packet_size,
            aHeader + sizeof( struct avdtp_fragment_header ),
            tmpResult - sizeof( struct avdtp_fragment_header ) );

    a2dp->reassembled_packet_size += ( tmpResult - sizeof( struct avdtp_fragment_header ) );

    if( PACKET_TYPE_END == aHeader->packet_type )
    {
     DBG( "END packet detected, copying restored %d bytes to output\n", a2dp->reassembled_packet_size );

     memcpy( aHeader, a2dp->reassembled_packet, a2dp->reassembled_packet_size );
     tmpResult                     = a2dp->reassembled_packet_size;
     a2dp->reassembled_packet_size = 0;
    }
    else
    {
     // Recursively reading data
     tmpResult = ReassembleAVDTPSignalPacket( a2dp, aHeader, a2dp->signalling_mtu );
    }
   }
   break;

   default:
    DBG( "Wrong packet type, discarding all\n" );

    tmpResult                     =  REASSEMBLE_DISCARD;
    a2dp->reassembled_packet_size = 0;
    memset( aHeader, 0, sizeof( *aHeader ) );    
	break;
  }

 return tmpResult;
}

#endif // ENABLE_PACKET_FRAGMENTATION

static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t *mtu, uint16_t *imtu)
{
        struct sockaddr_l2   addr;
        struct l2cap_options opts;
        int                  sk;
        unsigned int         opt;
        char                 errorText[ 2048 ];

        sk = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
        if (sk < 0){
                AppendError( "Can't create socket", 1 );
                return -1;
        }

        memset(&addr, 0, sizeof(addr));
        addr.l2_family = AF_BLUETOOTH;
        bacpy(&addr.l2_bdaddr, src);
        if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
                
                AppendError( "Can't bind socket", 1 );
                close( sk );
                return -1;
        }

        /* Get default options */
        opt = sizeof(opts);
        if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, &opt) < 0) {
                AppendError( "Can't get default L2CAP options", 1 );
                close( sk );
                return -1;
        }

        /* Set new options */
        if(mtu && *mtu) {
                DBG( "Setting OMTU to %d\n", *mtu );
                opts.omtu = *mtu;
                //opts.imtu = *mtu;
        }
        
        if(imtu && *imtu) 
        {
         if( *imtu > 0 )
          {
           DBG( "Setting IMTU to %d\n", *imtu );
           opts.imtu = *imtu;
          }
        }

        if (setsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, opt) < 0) {
                AppendError( "Can't set L2CAP options", 1 );
                close( sk );
                return -1;
        }

        memset(&addr, 0, sizeof(addr));
        addr.l2_family = AF_BLUETOOTH;
        bacpy(&addr.l2_bdaddr, dst);
        addr.l2_psm = htobs(psm);
        if (connect(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
                
                sprintf( errorText, "Can't connect to %s", batostr( &addr.l2_bdaddr ) );
                close(sk);
                AppendError( errorText, 1 );
                return -1;
        }

        opt = sizeof(opts);
        if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, &opt) < 0) {

                close(sk);
                AppendError( "Can't get L2CAP options", 1 );
                return -1;
        }

        DBG("Connected [imtu %d, omtu %d, flush_to %d]\n",
                        opts.imtu, opts.omtu, opts.flush_to);

        if (mtu)
                *mtu = opts.omtu;

        if( NULL != imtu )
        {
         *imtu = opts.imtu;
        }

        return sk;
}



static void fill_sbc_capabilities( struct sbc_codec_specific_elements* sbc_elements, packager_t* p, int min_bitpool, int max_bitpool )
{
 assert( NULL != sbc_elements );
 assert( NULL != p );

 int v = 0x02; // default to stereo

 switch(p->encoder->channel_mode) {
                case MONO:
                           v = 0x08;
                           break;

                case DUAL_CHANNEL:
                           v = 0x04;
                           break;

                case STEREO:
                           v = 0x02;
                           break;

                case JOINT:
                           v = 0x01;
                           break;

                default:
                        assert( 0 );
                        break;
        }
 sbc_elements->channel_mode = v;

 // Rate
 switch(p->encoder->rate) 
        {
                case 16000:
                        v = 0x08;
                        break;

                case 32000:
                        v = 0x04;
                        break;

                case 44100:
                        v = 0x02;
                        break;

                case 48000:
                        v = 0x01;
                        break;
                
                default:
                        fprintf( stderr, "Unknown sampling rate %d\n", p->encoder->rate );
                        fflush( stderr );
                        assert( 0 );
                        break;
        }
 sbc_elements->frequency = v;

 // Allocation method
 switch( p->encoder->allocation_method )
        {
          case SNR:
                    v = 0x2;
                    break;

          case LOUDNESS:
                    v = 0x1;
                    break;


          default:
              assert( 0 );
              break;
        }
 sbc_elements->allocation_method = v;

 // Subbands
 switch(p->encoder->subbands) {
                case 4:
                        v = 0x02;
                        break;

                case 8:
                        v = 0x01;
                        break;

                default:
                        assert( 0 );
                        break;
        }
 sbc_elements->subbands = v;

 // Blocks
 switch(p->encoder->blocks) {
                case 4:
                        v = 0x8;
                        break;

                case 8:
                        v = 0x4;
                        break;

                case 12:
                        v = 0x2;
                        break;

                case 16:
                        v = 0x1;
                        break;

                default:
                    assert( 0 );
                    break;
        }
 sbc_elements->block_length = v;
 sbc_elements->min_bitpool = min_bitpool;
 sbc_elements->max_bitpool = max_bitpool;
}

#ifdef DEBUG
static void DumpCapResp( int aReadSize, struct getcap_resp const* cap_resp )    
{
            int i;
            FILE* tmpF = NULL;

            assert( NULL != cap_resp );

#ifdef A2DP_DEBUG_FILE
            tmpF = fopen( A2DP_DEBUG_FILE_NAME, "a+b" );
#endif
            fprintf( stderr, "ReadSize %d, CP size %d, data:", aReadSize, sizeof( *cap_resp ) );
            if( NULL != tmpF )
            {
             fprintf( tmpF, "ReadSize %d, CP size %d, data:", aReadSize, sizeof( *cap_resp ) );
            }

            for( i = 0; i < aReadSize; ++i )
            {
             fprintf( stderr, "0x%x ", ( ( char * )( cap_resp ) )[ i ] );
             if( NULL != tmpF )
             {
              fprintf( tmpF, "0x%x ", ( ( char * )( cap_resp ) )[ i ] ); 
             }
            }

            fprintf( stderr, "\n" );
            if( NULL != tmpF )
            {
             fprintf( tmpF, "\n" );
             fclose( tmpF );
            }
}
#endif

struct generic_capability {
    // Media codec capability
	uint8_t cap_type;
	uint8_t length;
} __attribute__ ((packed));

struct media_codec_capability
{
	uint8_t cap_type;
	uint8_t length;
    
	uint8_t rfa1:4;
    uint8_t media_type:4;
	uint8_t media_codec_type;

	struct sbc_codec_specific_elements sbc_elements;
} __attribute__ ((packed));

// Returns 1 if SBC audio capabilities are found
static int process_cap_resp( int aReadSize, struct getcap_resp* cap_resp )
{
 int                        tmpResult = 0;
 struct generic_capability* tmpCurCap = ( struct generic_capability* )( &( cap_resp->serv_cap ) );
 
 assert( NULL != cap_resp );
 
 // Looking for Codec capability
 while( aReadSize && !tmpResult )
 {
  DBG( "Processing cap cat 0x%x, lenght %d, bytes left %d, need cat %d...\n", tmpCurCap->cap_type, tmpCurCap->length, aReadSize, MEDIA_CODEC_CATEGORY );

  if( MEDIA_CODEC_CATEGORY == tmpCurCap->cap_type )
  { // Found
   if( tmpCurCap->length >= 6 && aReadSize >= 8 )
   {
    struct media_codec_capability* tmpMediaCap = ( struct media_codec_capability * )( tmpCurCap );
 
    DBG( "Found MEDIA_CODEC_CATEGORY, media type %d, codec type %d, need media type %d, codec type %d\n", tmpMediaCap->media_type, tmpMediaCap->media_codec_type, AUDIO_MEDIA_TYPE, SBC_MEDIA_CODEC_TYPE );

    if( tmpMediaCap->media_type != AUDIO_MEDIA_TYPE ||
        tmpMediaCap->media_codec_type != SBC_MEDIA_CODEC_TYPE ) 
    {
     DBG( "MEDIA_CODEC_CATEGORY is of wrong type, continuing\n" );
     goto next_cap;
    }
    else
    { // Great, found right sink
     if( ( void* )( tmpMediaCap ) != ( void * )&( cap_resp->cap_type ) ) 
     { // Copying to right place, USE MEMMOVE !
      DBG( "Copying found cap to right place...\n" );
      memmove( &( cap_resp->cap_type ), tmpMediaCap, sizeof( *tmpMediaCap ) );
     }
     tmpResult = 1;
    }
   }
   else
   {
    DBG( "Wrong MEDIA_CODEC_CATEGORY length or total stuct length, skipping\n" );
    goto next_cap;
   }
  }
  else
  {
   int tmpOffset = 0;
   
next_cap:
   tmpOffset = tmpCurCap->length + sizeof( tmpCurCap->length ) + sizeof( tmpCurCap->cap_type );
   
   tmpCurCap = ( struct generic_capability* )( ( ( char* )( tmpCurCap ) + tmpOffset ) );
   aReadSize -= tmpOffset;
  }
 }

 return tmpResult;
}

static void send_abort( packager_t *packager, int s, int seid )
{
 struct abort_cmd       abort_stream;
 struct abort_rsp       abort_resp;

 DBG( "Sending AVDTP_ABORT, SEID %d...\n", seid );

 assert( NULL != packager );

                // Sending Abort command
                init_request( packager, &abort_stream.header, AVDTP_ABORT );
                abort_stream.acp_seid = seid;

                if( write(s, &abort_stream, sizeof(abort_stream)) != sizeof(abort_stream)) {
                  char errorText[ 200 ];
                  
                  sprintf( errorText, "Can not send ABORT to stream SEID = %d\n", seid);
                  AppendError( errorText, 1 );
                }
                else
                { 
                 DBG( "Reading abort response...\n" );

                 read( s, &abort_resp, sizeof( abort_resp ) );

                 DBG( "Received abort response, message %d\n", abort_resp.header.message_type );
                }
}

#ifndef ENABLE_PACKET_FRAGMENTATION
static int process_seid(packager_t *packager,int s, struct acp_seid_info * get_seid_resp, unsigned short *psm)
{       
  packager_t *p=(packager_t *)packager;
  //a2dp_packager_t *h=(a2dp_packager_t *)(p->udata);
        //int v;
               int             seid = get_seid_resp->acp_seid;
        struct getcap_resp     cap_resp;
        struct getcap_req      put_req;
        struct set_config      s_config;
        struct set_config_resp s_resp;
        struct open_stream_cmd open_stream;
        struct open_stream_rsp open_resp;
        struct get_config      s_getconf;
        struct get_config_resp get_resp;
               char            errorText[ 2048 ];
               
        int    tmpRetry = 10, tmpReadSize = 0;

        assert( NULL != p );

        a2dp_packager_t *h=(a2dp_packager_t *)(p->udata);        

        //seid++;
        DBG("SEID = %d, Busy %d, MT %d, TSEP %d\n", seid, get_seid_resp->inuse0, get_seid_resp->media_type, get_seid_resp->tsep );        

        assert( NULL != h );

        if( get_seid_resp->inuse0 )
        {
         DBG( "SEP is in use, aborting it\n" );
         send_abort( packager, s, seid );
         return -1; // Comment ?
        }

        memset(&put_req, 0, sizeof(put_req));
        init_request(packager,&put_req.header, AVDTP_GET_CAPABILITIES);
        put_req.acp_seid = seid;
        
        if (write(s, &put_req, sizeof(put_req)) != sizeof(put_req)) {
                sprintf( errorText, "Couldn't request capabilities for SEID = %d", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
        else DBG("Requested Capabilities for SEID = %d\n",seid);

/*read_cap_again:
        if ( ( tmpReadSize = read(s, &cap_resp, sizeof(cap_resp) ) ) < ( int )sizeof(cap_resp) ||
                        cap_resp.header.message_type == MESSAGE_TYPE_REJECT ||
                        cap_resp.media_type != AUDIO_MEDIA_TYPE ||
                        cap_resp.media_codec_type != SBC_MEDIA_CODEC_TYPE) {
#ifdef DEBUG
            DumpCapResp( tmpReadSize, &cap_resp );
#endif
                sprintf( errorText, "Didn't receive SBC codec parameters for SEID = %d", seid );
                if( ( EAGAIN == errno || EINTR == errno || EBUSY == errno ) && tmpRetry-- )
                {
                 DBG( "Retrying because of interrupted system call, retry %d\n", tmpRetry );
                 goto read_cap_again;
                }
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
*/
        
read_cap_again:
// Enable this to use packets fragmenting
//        if( ( tmpReadSize = ReassembleAVDTPSignalPacket( h, &( cap_resp.header ), sizeof( cap_resp ) ) ) < ( int )sizeof(cap_resp) ||
        errno = 0;
        if( ( tmpReadSize = read(s, &cap_resp, sizeof(cap_resp) ) ) < ( int )sizeof(cap_resp) ||
            cap_resp.header.message_type == MESSAGE_TYPE_REJECT ) {            
#ifdef DEBUG
            DumpCapResp( tmpReadSize, &cap_resp );
#endif
                sprintf( errorText, "Didn't receive SBC codec parameters for SEID = %d", seid );
                if( ( EAGAIN == errno || EINTR == errno || EBUSY == errno ) && tmpRetry-- )
                {
                 DBG( "Retrying because of interrupted system call, retry %d\n", tmpRetry );
                 goto read_cap_again;
                }
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }


        DBG("Got capabilities response: ");

#ifdef DEBUG
            DumpCapResp( tmpReadSize, &cap_resp );
#endif

        if( !process_cap_resp( tmpReadSize - sizeof( cap_resp.header ), &cap_resp ) )
        {
                sprintf( errorText, "Can not find Audio capabilities for SEID = %d", seid );
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }

        memset(&s_config, 0, sizeof(s_config));
        init_request(packager,&s_config.header, AVDTP_SET_CONFIGURATION);

        s_config.serv_cap         = MEDIA_TRANSPORT_CATEGORY;
        s_config.serv_cap_len     = 0 ; 
        s_config.acp_seid         = seid;
        s_config.int_seid         = INT_SEID;  // how should I choose the int_seid??
        s_config.cap_type         = MEDIA_CODEC_CATEGORY;
        s_config.length           = 6;
        s_config.media_type       = AUDIO_MEDIA_TYPE;
        s_config.media_codec_type = SBC_MEDIA_CODEC_TYPE;

        fill_sbc_capabilities( &( s_config.sbc_elements ), p, cap_resp.sbc_elements.min_bitpool, cap_resp.sbc_elements.max_bitpool );
/*
        // Channel mode
   	    v = 0x02; // default to stereo
        switch(p->encoder->channel_mode) {
                case MONO:
                           v = 0x08;
                           break;

                case DUAL_CHANNEL:
                           v = 0x04;
                           break;

                case STEREO:
                           v = 0x02;
                           break;

                case JOINT:
                           v = 0x01;
                           break;

                default:
                        assert( 0 );
                        break;
        }
        s_config.sbc_elements.channel_mode = v;

        // Rate
        switch(p->encoder->rate) 
        {
                case 16000:
                        v = 0x08;
                        break;

                case 32000:
                        v = 0x04;
                        break;

                case 44100:
                        v = 0x02;
                        break;

                case 48000:
                        v = 0x01;
                        break;
                
                default:
                        fprintf( stderr, "Unknown sampling rate %d\n", p->encoder->rate );
                        fflush( stderr );
                        assert( 0 );
                        break;
        }
        s_config.sbc_elements.frequency = v;

        // Allocation method
        switch( p->encoder->allocation_method )
        {
          case SNR:
                    v = 0x2;
                    break;

          case LOUDNESS:
                    v = 0x1;
                    break;


          default:
              assert( 0 );
              break;
        }
        s_config.sbc_elements.allocation_method = v;

        // Subbands
        switch(p->encoder->subbands) {
                case 4:
                        v = 0x02;
                        break;

                case 8:
                        v = 0x01;
                        break;

                default:
                        assert( 0 );
                        break;
        }
        s_config.sbc_elements.subbands = v;

        // Blocks
        switch(p->encoder->blocks) {
                case 4:
                        v = 0x8;
                        break;

                case 8:
                        v = 0x4;
                        break;

                case 12:
                        v = 0x2;
                        break;

                case 16:
                        v = 0x1;
                        break;

                default:
                    assert( 0 );
                    break;
        }
        s_config.sbc_elements.block_length = v;
        */
        /*
           int_seid was 1 before
           s_config was set from sbc_info before:
           s_config.sbc_elements.channel_mode = 8 >> sbc_info.channel_mode;
           s_config.sbc_elements.frequency = 8 >> sbc_info.sampling_frequency;
           s_config.sbc_elements.allocation_method = 1 << sbc_info.allocation_method;
           s_config.sbc_elements.subbands = 2 >> sbc_info.subbands;
           s_config.sbc_elements.block_length = 8 >> sbc_info.blocks;
         */
        s_config.sbc_elements.min_bitpool = cap_resp.sbc_elements.min_bitpool;
        s_config.sbc_elements.max_bitpool = cap_resp.sbc_elements.max_bitpool;

        //dump_packet(&s_config.sbc_elements, sizeof(s_config.sbc_elements)); //exit(0);

        if (!(cap_resp.sbc_elements.channel_mode & s_config.sbc_elements.channel_mode))
        {
                AppendError( "headset does not support this channel mode\n", 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (!(cap_resp.sbc_elements.frequency & s_config.sbc_elements.frequency))
        {
                sprintf( errorText, "headset does not support this frequency\n");
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (!(cap_resp.sbc_elements.allocation_method & s_config.sbc_elements.allocation_method))
        {
                sprintf( errorText, "headset does not support this allocation_method\n");
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (!(cap_resp.sbc_elements.subbands & s_config.sbc_elements.subbands))
        {
               sprintf( errorText, "headset does not support this subbands setting\n");
               AppendError( errorText, 0 );
               iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (write(s, &s_config, sizeof(s_config)) != sizeof(s_config)) {
                sprintf( errorText, "couldn't set config seid = %d\n", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
        DBG("Sent set configurations command\n");

        int size = read(s, &s_resp, sizeof(s_resp));
        if (size == sizeof(s_resp) - 2) {
                //if(size>0){
                DBG("Set configurations command accepted\n");
                h->configured = 1;
        } else {
                DBG("Set configurations command rejected, size %d, code 0x%x, serv cat 0x%x\n", size, s_resp.error_code, s_resp.serv_cat );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                h->configured = 0;
                return (-1); // ?? comment ?
        }

        // Sending Get Configuration
       
        init_request(packager,&s_getconf.header, AVDTP_GET_CONFIGURATION);
        s_getconf.acp_seid = seid;

        if (write(s, &s_getconf, sizeof(s_getconf)) != sizeof(s_getconf)) {
                sprintf( errorText, "couldn't get config seid = %d\n", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
        DBG("Sent GET configurations command\n");

        size = read(s, &get_resp, sizeof(get_resp));
        if (size > ( int )( sizeof( get_resp.header ) + sizeof( uint8_t ) ) )  {                
                DBG("Get configurations command accepted\n");
        } else {
                DBG("Get configurations command rejected, size %d, code 0x%x\n", size, get_resp.error_code );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                h->configured = 0;
                return (-1); // ?? comment ?
        }

//        DBG("Sent ABORT stream command\n");
//        send_abort( packager, s, seid );
//        return (-1);

        memset(&open_stream, 0, sizeof(open_stream));
        init_request(packager,&open_stream.header, AVDTP_OPEN);
        open_stream.acp_seid = seid;

        if (write(s, &open_stream, sizeof(open_stream)) != sizeof(open_stream)) {
                sprintf( errorText, "Couldn't open stream SEID = %d\n", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
                return (-1);
        }

        DBG("Sent open stream command\n");

        if (read(s, &open_resp, sizeof(open_resp)) < ( int )sizeof(open_resp) - 1 ||
                        open_resp.header.message_type == MESSAGE_TYPE_REJECT) {
                sprintf( errorText, "Didn't receive open response confirm for SEID = %d\n", seid);
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;

                send_abort( packager, s, seid );

                return (-1);
        }

        DBG("Got open stream confirm\n");

        *psm = CMD_PSM;

        return 0;
}
#else

static int process_seid(packager_t *packager,int s, struct acp_seid_info * get_seid_resp, unsigned short *psm)
{       
  packager_t *p=(packager_t *)packager;
  //a2dp_packager_t *h=(a2dp_packager_t *)(p->udata);
        //int v;
               int             seid = get_seid_resp->acp_seid;
        a2dp_generic_responce_t cap_resp;
        struct getcap_req      put_req;
        struct set_config      s_config;
        struct set_config_resp s_resp;
        struct open_stream_cmd open_stream;
        struct open_stream_rsp open_resp;
        struct get_config      s_getconf;
        struct get_config_resp get_resp;
               char            errorText[ 2048 ];
               
        int    tmpRetry = 10, tmpReadSize = 0;

        assert( NULL != p );

        a2dp_packager_t *h=(a2dp_packager_t *)(p->udata);        

        //seid++;
        DBG("SEID = %d, Busy %d, MT %d, TSEP %d\n", seid, get_seid_resp->inuse0, get_seid_resp->media_type, get_seid_resp->tsep );        

        assert( NULL != h );

        if( get_seid_resp->inuse0 )
        {
         DBG( "SEP is in use, aborting it\n" );
         send_abort( packager, s, seid );
         return -1; // Comment ?
        }

        memset(&put_req, 0, sizeof(put_req));
        init_request(packager,&put_req.header, AVDTP_GET_CAPABILITIES);
        put_req.acp_seid = seid;
        
        if (write(s, &put_req, sizeof(put_req)) != sizeof(put_req)) {
                sprintf( errorText, "Couldn't request capabilities for SEID = %d", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
        else DBG("Requested Capabilities for SEID = %d\n",seid);

/*read_cap_again:
        if ( ( tmpReadSize = read(s, &cap_resp, sizeof(cap_resp) ) ) < ( int )sizeof(cap_resp) ||
                        cap_resp.header.message_type == MESSAGE_TYPE_REJECT ||
                        cap_resp.media_type != AUDIO_MEDIA_TYPE ||
                        cap_resp.media_codec_type != SBC_MEDIA_CODEC_TYPE) {
#ifdef DEBUG
            DumpCapResp( tmpReadSize, &cap_resp );
#endif
                sprintf( errorText, "Didn't receive SBC codec parameters for SEID = %d", seid );
                if( ( EAGAIN == errno || EINTR == errno || EBUSY == errno ) && tmpRetry-- )
                {
                 DBG( "Retrying because of interrupted system call, retry %d\n", tmpRetry );
                 goto read_cap_again;
                }
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
*/
        
read_cap_again:
        errno    = 0;
        h->cmdfd = s;
        if( ( tmpReadSize = ReassembleAVDTPSignalPacket( h, &( cap_resp.responce_header ), sizeof( cap_resp ) ) ) < ( int )( sizeof(cap_resp.responce_header) + sizeof( uint8_t ) ) ||
            cap_resp.responce_header.message_type == MESSAGE_TYPE_REJECT ) {            
#ifdef DEBUG
            DumpCapResp( tmpReadSize, &cap_resp.resp_getcap_cmd );
#endif
                sprintf( errorText, "Didn't receive SBC codec parameters for SEID = %d", seid );
                if( ( EAGAIN == errno || EINTR == errno || EBUSY == errno ) && tmpRetry-- )
                {
                 DBG( "Retrying because of interrupted system call, retry %d\n", tmpRetry );
                 goto read_cap_again;
                }
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                h->cmdfd = -1;
                return (-1);
        }


        h->cmdfd = -1;
        
        DBG("Got capabilities response: ");

#ifdef DEBUG
            DumpCapResp( tmpReadSize, &cap_resp.resp_getcap_cmd );
#endif

        if( !process_cap_resp( tmpReadSize - sizeof( cap_resp.responce_header ), &cap_resp.resp_getcap_cmd ) )
        {
                sprintf( errorText, "Can not find Audio capabilities for SEID = %d", seid );
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }

        memset(&s_config, 0, sizeof(s_config));
        init_request(packager,&s_config.header, AVDTP_SET_CONFIGURATION);

        s_config.serv_cap         = MEDIA_TRANSPORT_CATEGORY;
        s_config.serv_cap_len     = 0 ; 
        s_config.acp_seid         = seid;
        s_config.int_seid         = INT_SEID;  // how should I choose the int_seid??
        s_config.cap_type         = MEDIA_CODEC_CATEGORY;
        s_config.length           = 6;
        s_config.media_type       = AUDIO_MEDIA_TYPE;
        s_config.media_codec_type = SBC_MEDIA_CODEC_TYPE;

        fill_sbc_capabilities( &( s_config.sbc_elements ), 
                               p, 
                               cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.min_bitpool, 
                               cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.max_bitpool );
/*
        // Channel mode
   	    v = 0x02; // default to stereo
        switch(p->encoder->channel_mode) {
                case MONO:
                           v = 0x08;
                           break;

                case DUAL_CHANNEL:
                           v = 0x04;
                           break;

                case STEREO:
                           v = 0x02;
                           break;

                case JOINT:
                           v = 0x01;
                           break;

                default:
                        assert( 0 );
                        break;
        }
        s_config.sbc_elements.channel_mode = v;

        // Rate
        switch(p->encoder->rate) 
        {
                case 16000:
                        v = 0x08;
                        break;

                case 32000:
                        v = 0x04;
                        break;

                case 44100:
                        v = 0x02;
                        break;

                case 48000:
                        v = 0x01;
                        break;
                
                default:
                        fprintf( stderr, "Unknown sampling rate %d\n", p->encoder->rate );
                        fflush( stderr );
                        assert( 0 );
                        break;
        }
        s_config.sbc_elements.frequency = v;

        // Allocation method
        switch( p->encoder->allocation_method )
        {
          case SNR:
                    v = 0x2;
                    break;

          case LOUDNESS:
                    v = 0x1;
                    break;


          default:
              assert( 0 );
              break;
        }
        s_config.sbc_elements.allocation_method = v;

        // Subbands
        switch(p->encoder->subbands) {
                case 4:
                        v = 0x02;
                        break;

                case 8:
                        v = 0x01;
                        break;

                default:
                        assert( 0 );
                        break;
        }
        s_config.sbc_elements.subbands = v;

        // Blocks
        switch(p->encoder->blocks) {
                case 4:
                        v = 0x8;
                        break;

                case 8:
                        v = 0x4;
                        break;

                case 12:
                        v = 0x2;
                        break;

                case 16:
                        v = 0x1;
                        break;

                default:
                    assert( 0 );
                    break;
        }
        s_config.sbc_elements.block_length = v;
        */
        /*
           int_seid was 1 before
           s_config was set from sbc_info before:
           s_config.sbc_elements.channel_mode = 8 >> sbc_info.channel_mode;
           s_config.sbc_elements.frequency = 8 >> sbc_info.sampling_frequency;
           s_config.sbc_elements.allocation_method = 1 << sbc_info.allocation_method;
           s_config.sbc_elements.subbands = 2 >> sbc_info.subbands;
           s_config.sbc_elements.block_length = 8 >> sbc_info.blocks;
         */
        s_config.sbc_elements.min_bitpool = cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.min_bitpool;
        s_config.sbc_elements.max_bitpool = cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.max_bitpool;

        //dump_packet(&s_config.sbc_elements, sizeof(s_config.sbc_elements)); //exit(0);

        if (!(cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.channel_mode & s_config.sbc_elements.channel_mode))
        {
                AppendError( "headset does not support this channel mode\n", 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (!(cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.frequency & s_config.sbc_elements.frequency))
        {
                sprintf( errorText, "headset does not support this frequency\n");
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (!(cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.allocation_method & s_config.sbc_elements.allocation_method))
        {
                sprintf( errorText, "headset does not support this allocation_method\n");
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (!(cap_resp.resp_get_config_cmd.serv_cap.sbc_elements.subbands & s_config.sbc_elements.subbands))
        {
               sprintf( errorText, "headset does not support this subbands setting\n");
               AppendError( errorText, 0 );
               iLastA2DPErrorCode = A2DP_ERROR_CODEC_NOT_SUPPORTED;
        }

        if (write(s, &s_config, sizeof(s_config)) != sizeof(s_config)) {
                sprintf( errorText, "couldn't set config seid = %d\n", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
        DBG("Sent set configurations command\n");

        int size = read(s, &s_resp, sizeof(s_resp));
        if (size == sizeof(s_resp) - 2) {
                //if(size>0){
                DBG("Set configurations command accepted\n");
                h->configured = 1;
        } else {
                DBG("Set configurations command rejected, size %d, code 0x%x, serv cat 0x%x\n", size, s_resp.error_code, s_resp.serv_cat );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                h->configured = 0;
                return (-1); // ?? comment ?
        }

        // Sending Get Configuration
       
        init_request(packager,&s_getconf.header, AVDTP_GET_CONFIGURATION);
        s_getconf.acp_seid = seid;

        if (write(s, &s_getconf, sizeof(s_getconf)) != sizeof(s_getconf)) {
                sprintf( errorText, "couldn't get config seid = %d\n", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                return (-1);
        }
        DBG("Sent GET configurations command\n");

        size = read(s, &get_resp, sizeof(get_resp));
        if (size > ( int )( sizeof( get_resp.header ) + sizeof( uint8_t ) ) )  {                
                DBG("Get configurations command accepted\n");
        } else {
                DBG("Get configurations command rejected, size %d, code 0x%x\n", size, get_resp.error_code );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CAPABILITIES;
                h->configured = 0;
                return (-1); // ?? comment ?
        }

//        DBG("Sent ABORT stream command\n");
//        send_abort( packager, s, seid );
//        return (-1);

        memset(&open_stream, 0, sizeof(open_stream));
        init_request(packager,&open_stream.header, AVDTP_OPEN);
        open_stream.acp_seid = seid;

        if (write(s, &open_stream, sizeof(open_stream)) != sizeof(open_stream)) {
                sprintf( errorText, "Couldn't open stream SEID = %d\n", seid);
                AppendError( errorText, 1 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
                return (-1);
        }

        DBG("Sent open stream command\n");

        if (read(s, &open_resp, sizeof(open_resp)) < ( int )sizeof(open_resp) - 1 ||
                        open_resp.header.message_type == MESSAGE_TYPE_REJECT) {
                sprintf( errorText, "Didn't receive open response confirm for SEID = %d\n", seid);
                AppendError( errorText, 0 );
                iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;

                send_abort( packager, s, seid );

                return (-1);
        }

        DBG("Got open stream confirm\n");

        *psm = CMD_PSM;

        return 0;
}

#endif









int DetectA2dp(packager_t *packager) {
        packager_t      *p=(packager_t *)packager;
        a2dp_packager_t *h=(a2dp_packager_t *)(p->udata);
        sdp_session_t   *sess;
        sdp_list_t      *attrid, *search, *seq, *next;
        sdp_data_t      *pdlist;
        uuid_t           group;
        uint32_t         range = 0x0000ffff;
        int              err;
        //char             errorText[ 2048 ];

        h->psm_cmd = CMD_PSM;
        h->flags   = 0;
        
        sess = sdp_connect(&(h->src), &(h->dst), SDP_RETRY_IF_BUSY);
        if (!sess) {
                /*DBG( "SDP connect error, retrying err %d, %s...\n", errno, strerror(errno) );
                sess = sdp_connect(&(h->src), &(h->dst), SDP_RETRY_IF_BUSY);
                if( !sess )*/
                {                
                 AppendError( "Failed to connect to SDP server", 1 );
                 iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_CONNECT_SDP;
                 return 0;
                }
        }

        /* 0x1108->all? 0x1101->rf sink 0x111e->handsfree 0x1108->headset */
        sdp_uuid16_create(&group, 0x110d);
        search = sdp_list_append(0, &group);
        attrid = sdp_list_append(0, &range);
        err = sdp_service_search_attr_req(sess, search,
                        SDP_ATTR_REQ_RANGE, attrid, &seq);
        sdp_list_free(search, 0);
        sdp_list_free(attrid, 0);

        if (err) {
                AppendError( "Service Search failed", 1 );
                sdp_close(sess);
                iLastA2DPErrorCode = A2DP_ERROR_A2DP_NOT_SUPPORTED;
                return 0;
        }

        for (; seq; seq = next) {
                sdp_record_t *rec = (sdp_record_t *) seq->data;

                DBG("Found A2DP Sink\n");

                next = seq->next;
                free(seq);
                sdp_record_free(rec);
        }
        sdp_uuid16_create(&group, PNP_INFO_SVCLASS_ID);
        search = sdp_list_append(0, &group);
        attrid = sdp_list_append(0, &range);
        err = sdp_service_search_attr_req(sess, search,
                        SDP_ATTR_REQ_RANGE, attrid, &seq);
        sdp_list_free(search, 0);
        sdp_list_free(attrid, 0);

        if (err)
        {
                //iLastA2DPErrorCode = A2DP_ERROR_A2DP_NOT_SUPPORTED;
                goto done;
        }


        for (; seq; seq = next) {
                sdp_record_t *rec = (sdp_record_t *) seq->data;

                pdlist = sdp_data_get(rec, 0x0201);
                h->vendor = pdlist ? pdlist->val.uint16 : 0x0000;

                pdlist = sdp_data_get(rec, 0x0202);
                h->product = pdlist ? pdlist->val.uint16 : 0x0000;

                pdlist = sdp_data_get(rec, 0x0203);
                h->version = pdlist ? pdlist->val.uint16 : 0x0000;

                DBG("Product ID %04x:%04x:%04x\n", h->vendor, h->product, h->version);

                if (h->vendor == 0x1310 && h->product == 0x0100 && h->version == 0x0104) {
                        DBG("Enabling GCT media payload workaround\n");
                        h->flags |= NONSPECAUDIO;
                }

                next = seq->next;
                free(seq);
                sdp_record_free(rec);
        }

done:
        sdp_close(sess);
        return 1;
}

int StartA2dpCommand(packager_t *packager) {
  packager_t *p=(packager_t *)packager;
  a2dp_packager_t *h=(a2dp_packager_t *)(p->udata);
  unsigned short   psm_cmd, psm_stream;
  int              size,i;
  struct sepd_req  put_req;
  struct sepd_resp get_resp;
  int              cmdfd;
  int              last_seid_index;
  char             errorText[ 2048 ];

  SetRoleSwitch( h, 1 );  

  psm_cmd           = CMD_PSM;
  h->signalling_mtu = DEFAULT_SIGNALLING_MTU;
  cmdfd             = do_connect(&(h->src), &(h->dst), psm_cmd, &( h->signalling_mtu ), &( h->signalling_mtu ) );

  if( cmdfd < 0 ) {
          sprintf( errorText, "cannot open psm_cmd = %d\n", psm_cmd);
          AppendError( errorText, 1 );
          iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
          return 0;
  }

  //A2DPSwitchRole( packager, 1 );
  
  // avdt_discover_req
  memset(&put_req, 0, sizeof(put_req));
  init_request(packager,&put_req.header, AVDTP_DISCOVER);

  if (write(cmdfd, &put_req, sizeof(put_req)) != sizeof(put_req)) {
          AppendError( "Couldn't send avdtp_discover", 1 );
          close(cmdfd);
          iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
          return 0;
  }
  else DBG("Sent the Stream End Point Discovery Command\n");


  size = read(cmdfd, &get_resp, sizeof(get_resp));
  if (size < (int)(sizeof(get_resp) - MAX_ADDITIONAL_CODEC_OCTETS) || 
      MESSAGE_TYPE_REJECT == get_resp.header.message_type ) {
          AppendError( "Couldn't get avdtp_discover", 0 );
          close(cmdfd);
          iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
#ifdef DEBUG
            int i;
        
            fprintf( stderr, "ReadSize %d, CP size %d, data:", size, sizeof( get_resp ) );
            for( i = 0; i < size; ++i )
            {
             fprintf( stderr, "0x%x ", ( ( char * )( &get_resp ) )[ i ] );
            }

            fprintf( stderr, "\n" );
#endif
          return 0;
  }
  else 
  {
      DBG("Got a Stream End Point Discovery Response\n");
#ifdef DEBUG
            int i;
        
            fprintf( stderr, "ReadSize %d, CP size %d, data:", size, sizeof( get_resp ) );
            for( i = 0; i < size; ++i )
            {
             fprintf( stderr, "0x%x ", ( ( char * )( &get_resp ) )[ i ] );
            }

            fprintf( stderr, "\n" );
#endif
  }

  h->seid = -1;
  last_seid_index = MAX_ADDITIONAL_CODEC - ((sizeof(get_resp)-size)/sizeof(struct acp_seid_info));

  DBG("received %d capabilities\n", last_seid_index + 1);

  for(i=0; i <= last_seid_index; i++) {
          if (process_seid(packager, cmdfd, &get_resp.infos[i], &psm_stream) == 0) {
                  h->seid = get_resp.infos[i].acp_seid;
                  break;
          }
  }

  if(h->seid == -1) {
          //We have not found the seid that we want
          AppendError( "Couldn't locate the correct seid", 0 );
          return 0;
  }

  h->psm_cmd    = psm_cmd;
  h->psm_stream = psm_stream;
  h->cmdfd      = cmdfd;

  //SetRoleSwitch( h, 0 );

  return 1;
}




int StartA2dpStream(packager_t *p, int aSendStartCommand ) {
  a2dp_packager_t            *h=(a2dp_packager_t *)(p->udata);
  struct start_stream_cmd     start_stream;
  struct start_stream_rsp     start_resp;
  char                        errorText[ 2048 ];

 assert( NULL != p );
 assert( NULL != h );

  if( !h->paused )
  {
   if( -1 == h->streamfd )
   {
    DBG( "Opening MEDIA CHANNEL...\n" );
    
    h->streamfd = do_connect(&(h->src), &(h->dst), h->psm_stream, &(h->mtu), NULL );
    if (h->streamfd < 0) {
          sprintf( errorText, "cannot open psm_stream = %d\n", h->psm_stream );
          AppendError( errorText, 1 );
          iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
          return 0;
    }
   }
   else
   {
    DBG( "Stream channel already opened\n" );
   }
  }

  assert( h->streamfd >= 0 );

  DBG( "Send start %d\n", aSendStartCommand );

  if( aSendStartCommand && !h->started )
  {
   memset(&start_stream, 0, sizeof(start_stream));
   init_request(p,&start_stream.header, AVDTP_START);
   start_stream.acp_seid = h->seid;

   if (write(h->cmdfd, &start_stream, sizeof(start_stream)) != sizeof(start_stream)) {
          sprintf( errorText, "couldn't send start_stream\n");
          AppendError( errorText, 1 );
          close(h->streamfd);
          close(h->cmdfd);
          h->streamfd = -1;
          h->cmdfd    = -1;
          iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
	  return 0;
   }
                                
   DBG("Sent stream START\n");

   if (read(h->cmdfd, &start_resp, sizeof(start_resp)) < 
	( int )sizeof(start_resp) - 2 ||start_resp.header.message_type == MESSAGE_TYPE_REJECT) {
          sprintf( errorText, "didn't receive start_resp confirm for seid = %d\n", h->seid);

          // Aborting
          send_abort( p, h->cmdfd, h->seid );
          
          AppendError( errorText, 0 );
          close(h->streamfd);
          close(h->cmdfd);
          h->streamfd = -1;
          h->cmdfd    = -1;
          iLastA2DPErrorCode = A2DP_ERROR_CAN_NOT_OPEN_CHANNEL;
	  return 0;
   }   

   DBG( "Got START stream confirm\n" );
  }
  else
  {
   DBG( "Not sending START command\n" );
  }

  h->paused  = 0;
  h->started = 1;
  
  DBG("mtu is %i\n",h->mtu);

  return 1;
}

int getA2dpNrFrames(packager_t *p) {
  a2dp_packager_t *a=(a2dp_packager_t *)p->udata;
  int fsize,nrframes;

	// frame size determined by encoder
  fsize=p->encoder->framelength(p->encoder);
  DBG("fsize=%i\n",fsize);
        // number of frames is detemined by what fits in 
	// the MTU after the headers
  nrframes=(a->mtu - 
              (sizeof(struct media_packet_header) +
               sizeof(struct media_payload_header))) / fsize;
  DBG("nrframes=%i\n",nrframes);

  return nrframes;
}

int getA2dpPacketLen(packager_t *p) {
  a2dp_packager_t *a=(a2dp_packager_t *)p->udata;
  int fsize,nrframes,size;

	// frame size determined by encoder
  fsize=p->encoder->framelength(p->encoder);
  DBG("fsize=%i\n",fsize);

        // number of frames is detemined by what fits in 
	// the MTU after the headers
  nrframes=(a->mtu - 
              (sizeof(struct media_packet_header) +
               sizeof(struct media_payload_header))) / fsize;
  DBG("nrframes=%i\n",nrframes);

	// so a packet is the nr of frames times the frame size,
	// and the size of the headers added to that product
  size=(nrframes*fsize) + 
       (sizeof(struct media_packet_header) +
        sizeof(struct media_payload_header));
  DBG("size=%i\n",size);

  return size; 
}


static int find_conn(int s, int dev_id, long arg)
{
	struct hci_conn_list_req *cl;
	struct hci_conn_info *ci;
	int i;

	if (!(cl = malloc(10 * sizeof(*ci) + sizeof(*cl)))) {
		DBG( "Can't allocate memory, code %d, %s\n", errno, strerror( errno ) );
		return 0;
	}
	cl->dev_id = dev_id;
	cl->conn_num = 10;
	ci = cl->conn_info;

	if (ioctl(s, HCIGETCONNLIST, (void *) cl)) {
		DBG( "Can't get connection list, code %d, %s", errno, strerror( errno ) );
		return 0;
	}

	for (i = 0; i < cl->conn_num; i++, ci++)
		if (!bacmp((bdaddr_t *) arg, &ci->bdaddr))
			return 1;

	return 0;
}

/*static int A2DPSwitchRole( packager_t* p, int aRole )
{
 int     dev_id    = -1;
 int     dd;

 a2dp_packager_t* h = ( a2dp_packager_t* )p->udata; 

#ifdef DEBUG
  char ba[ 100 ];

 ba2str( &( h->dst ), ba );

 DBG( "Switching to role to %d on %s\n", aRole, ba );
#endif

 assert( NULL != p );
 assert( NULL != h );
 
 dev_id = hci_for_each_dev( HCI_UP, find_conn, (long) &( h->dst ));
 if( dev_id < 0 ) 
 {
  DBG( "Not connected to %s\n", ba );
  
  return -1;			
 }

 dd = hci_open_dev( dev_id );
 if (dd < 0) 
 {
  DBG( "HCI device open failed , code %d, %s", errno, strerror( errno ) );
  return -2;
 }

 if( hci_switch_role( dd, &( h->dst ), aRole, 10000 ) < 0 ) 
 {
  DBG( "Can not switch role: %d %s\n", errno, strerror( errno ) );
 }

 close( dd );

 return 0;
}
*/
static int GetA2DPLinkQuality( packager_t* p )
{
 int     dev_id    = -1;
 struct  hci_conn_info_req *cr;
 uint8_t lq;
 int     dd;
 a2dp_packager_t* h = ( a2dp_packager_t* )p->udata; 

#ifdef DEBUG
  char ba[ 100 ];

 ba2str( &( h->dst ), ba );
 //DBG( "Addr: %s ", ba );
#endif

 assert( NULL != p );
 assert( NULL != h );
 
 dev_id = hci_for_each_dev( HCI_UP, find_conn, (long) &( h->dst ));
 if( dev_id < 0 ) 
 {
  DBG( "Not connected to %s\n", ba );
  
  return -1;			
 }

 dd = hci_open_dev( dev_id );
 if (dd < 0) 
 {
  DBG( "HCI device open failed , code %d, %s", errno, strerror( errno ) );
  return -2;
 }

 cr = malloc( sizeof( *cr ) + sizeof( struct hci_conn_info ) );
 if( !cr )
 {
  DBG( "Can not allocate memory for hci_conn_info_req: code %d, %s\n", errno, strerror( errno ) );

  close( dd );
  return -3;
 }

 bacpy( &cr->bdaddr, &( h->dst ) );
 cr->type = ACL_LINK;
 if( ioctl( dd, HCIGETCONNINFO, (unsigned long) cr) < 0 ) 
 {
  DBG( "Get connection info failed, code %d, %s\n", errno, strerror( errno ) );

  close( dd );
  return -4;
 }

 if( hci_read_link_quality( dd, htobs( cr->conn_info->handle ), &lq, 1000 ) < 0 ) 
 {
  DBG( "HCI read_link_quality request failed, code %d, %s\n", errno, strerror( errno ) );
  if( ETIMEDOUT == errno || EBUSY == errno || EAGAIN == errno || EINTR == errno )
  { // In case of timeout no error
   lq = MIN_LINK_QUALITY;
  }
  else
  {
   lq = -5;
  }
 }
 else
 {
  DBG( "Link quality: %d\n", lq);
 }

 close( dd );
 free( cr );

 return lq;
}

static int CheckA2DPLinkQuality( packager_t* p, int aDirectCheck )
{
       int tmpResult = 1;
 const int tmpLQ     = GetA2DPLinkQuality( p );

 assert( NULL != p );

 a2dp_packager_t* h = ( a2dp_packager_t* )p->udata; 
 assert( NULL != h );

 if( tmpLQ < MIN_LINK_QUALITY /*&& tmpLQ > 0 */ ) 
 {
  DBG( "Link quality too low (min allowed %d)\n", MIN_LINK_QUALITY );

  if( h->link_bad++ > LINK_BAD_TRESHOLD || aDirectCheck )
  {
   DBG( "Link qality bad for a long time, exiting\n" );
   tmpResult          = 0;
   h->link_bad        = 0;
   iLastA2DPErrorCode = A2DP_ERROR_BAD_LINK_QUALITY;
  }  
 }
 else
 {
  h->link_bad = 0;
 }
 
 /*if( tmpLQ < 0 )
 { // Not connected !!!
   a2dp_packager_t* a2dp = ( a2dp_packager_t* )( p->udata );
   char tmpErrorText[ 2048 ];

   DBG( "Re-initialising in listening mode\n" );

   assert( NULL != a2dp );

   InitError( tmpErrorText );

   close( a2dp->streamfd );
   a2dp->streamfd = -1;
   CloseA2dpCommand( p );
   ListenA2dp( a2dp, a2dp->listen_data.iBTAddr );
 }*/

 return tmpResult;
}

static void CloseA2dp( packager_t* packager )
{
 struct close_stream_cmd    close_stream;
 struct close_stream_rsp    close_resp;
 a2dp_packager_t*           h = ( a2dp_packager_t* )( packager->udata );

 assert( NULL != packager );
 assert( NULL != h );

 DBG( "Closing A2DP, cmdfd %d, strfd %d...\n", h->cmdfd, h->streamfd );

 if( h->cmdfd >= 0 )
 {
  // Closing command channel
  if( h->streamfd >= 0 )
  {
   if( h->started )
   {
    memset( &close_stream, 0, sizeof( close_stream ) );
    init_request( packager, &close_stream.header, AVDTP_CLOSE );
    close_stream.acp_seid = h->seid;

    if( write( h->cmdfd, &close_stream, sizeof( close_stream ) ) != sizeof( close_stream ) ) 
    {
     DBG( "Couldn't close stream SEID = %d\n", h->seid );                
    } else {
     DBG( "Sent close stream command\n" );
    }

    memset( &close_resp, 0, sizeof( close_resp ) );
    if( read( h->cmdfd, &close_resp, sizeof( close_resp)) <  ( int )sizeof( close_resp ) - 1 || 
        close_resp.header.message_type == MESSAGE_TYPE_REJECT) 
    {
      DBG( "Didn't received good close confirm for SEID %d, msg type %d, error 0x%x\n", h->seid, close_resp.header.message_type, close_resp.error );

      send_abort( packager, h->cmdfd, h->seid );
    }
    else
    {
     DBG( "Got close stream confirm\n" );
    }
   }
   else
   {
    DBG( "Not sending CLOSE command\n" );
   }

   close( h->streamfd );
   h->streamfd = -1;
  }
  
  // Closing ACL sockets
  close( h->cmdfd );
  SetRoleSwitch( h, 0 );
 }
 
 h->cmdfd      = -1;
 h->started    = 0;
 h->paused     = 0;
 h->configured = 0;
}

// Returns 1 in case of success
int ConnectA2dp( void *packager, const char *dstaddr, char* aErrorText ) 
{
 DBG( "Connect on %p to %s, err %p\n", packager, dstaddr, aErrorText );

 packager_t*      p    = ( packager_t* )packager;
 a2dp_packager_t* a2dp = ( a2dp_packager_t* )( p->udata );

 assert( NULL != p );
 assert( NULL != a2dp );

 InitError( aErrorText );

 iLastA2DPErrorCode = A2DP_ERROR_UNKNOWN; 

 // Stopping listener
 DBG( "Aborting listening...\n" );
 StopListen( &( a2dp->listen_data ) );
 
 if( -1 == a2dp->cmdfd )
 {
  DBG( "Starting a2dp command channel...\n" );
  if( !StartA2dpCommand( p ) ) goto error;
 }
 else
 {
  DBG( "Command channel already started\n" );
 }

 DBG( "Checking link quality..." );
 if( !CheckA2DPLinkQuality( p, 1 ) ) 
  {
   AppendError( "Bad link quality\n", 0 );
   goto error;
  }
  
 if( -1 == a2dp->streamfd || !a2dp->started )
 {
  DBG( "Starting a2dp stream channel...\n" );
  if( !StartA2dpStream( p, 1 ) ) goto error;
 }
 else
 {
  DBG( "Stream channel already opened\n" );
 }

 // now initialize the size fields
 p->packetlen   = getA2dpPacketLen( p );
 a2dp->nrframes = getA2dpNrFrames( p );

 DBG( "frame duration: %lums\n",p->encoder->sbc.duration * a2dp->nrframes );
 DBG( "timer: %lums\n",( p->encoder->sbc.duration * a2dp->nrframes ) / 2 );

 DBG( "A2dp connected\n" );

 SetRoleSwitch( a2dp, 0 ); // Disabling role switch

 return 1;
 
error:
  DBG( "Can not connect\n" );

  SetRoleSwitch( a2dp, 0 ); // Disabling role switch
  
  return 0;
}

static void ListenA2dp( a2dp_packager_t* a2dp, const char* dstaddr );

static void RestartListenA2dp( a2dp_packager_t* a2dp )
{
 assert( NULL != a2dp );

 ba2str( &( a2dp->dst ), a2dp->listen_data.iBTAddr );

 DBG( "Restarting A2DP listen at %s...\n", a2dp->listen_data.iBTAddr );
 
 ListenA2dp( a2dp, a2dp->listen_data.iBTAddr );
}

static void RemoteSinkConnectedCallback( a2dp_listen_t* aListenData )
{
 assert( NULL != aListenData );

 a2dp_packager_t* a2dp = ( a2dp_packager_t* )( aListenData->iUserData );
 struct l2cap_options opts;
 int                  opt;

 DBG( "Addr %s, FD %d, res %d\n", aListenData->iBTAddr, aListenData->iFile, aListenData->iResult );

 assert( NULL != a2dp );
 
 switch( aListenData->iResult )
 {
  case EErrorNone:
   assert( -1 == a2dp->cmdfd );
   a2dp->cmdfd = aListenData->iFile;

   opt = sizeof(opts);
   if( getsockopt( a2dp->cmdfd, SOL_L2CAP, L2CAP_OPTIONS, &opts, &opt) < 0) {

                DBG( "Can't get L2CAP options. %s(%d)\n", strerror(errno), errno);
                close(a2dp->cmdfd);
                a2dp->cmdfd = -1;
                RestartListenA2dp( a2dp );
                return;
        }

    DBG( "Connected [imtu %d, omtu %d, flush_to %d]\n", opts.imtu, opts.omtu, opts.flush_to );

    a2dp->mtu = opts.omtu;    
   break;

  case EErrorFailed:
   DBG( "BT listening failed, restart\n" );
   RestartListenA2dp( a2dp );
   break;

  case EErrorInterrupted:
   // Do nothing
   DBG( "A2DP Listening interrupted\n" );

   aListenData->iStopListen = 1;
   break;

  default:
   DBG( "Incorrect listen result code\n" );
 }
}

static void ListenA2dp( a2dp_packager_t* a2dp, const char* dstaddr )
{
 assert( NULL != a2dp ); 

 memset( &( a2dp->listen_data ), 0, sizeof( a2dp->listen_data ) );

 if( NULL != dstaddr && dstaddr != a2dp->listen_data.iBTAddr )
 {
  strcpy( a2dp->listen_data.iBTAddr, dstaddr );
 }

 a2dp->listen_data.iChannel        = CMD_PSM; 
 a2dp->listen_data.RemoteConnected = RemoteSinkConnectedCallback;
 a2dp->listen_data.iUserData       = a2dp;

 // Starting listener 
 DBG( "Staring listening for connections on %s...\n", a2dp->listen_data.iBTAddr );
 
 if( StartListen( &( a2dp->listen_data ) ) == EErrorNone )
 {
  DBG( "Started listening\n" );
 }
 else
 {
  DBG( "Can not start listening\n" );
 }
}

void SetRoleSwitch( a2dp_packager_t *a2dp, int aEnable )
{
 assert( NULL != a2dp );
 
 int                 tmpNewPolicy = -1, tmpNewLinkMode = -1;
 struct hci_dev_info tmp_device_info;

 tmp_device_info.dev_id = DEFAULT_HCI_DEVICE;

 DBG( "Ctl Fd %d, Dev %d, Enable %d\n", a2dp->device_ctl, tmp_device_info.dev_id, aEnable );

 if( !ioctl( a2dp->device_ctl, HCIGETDEVINFO, (void *)&( tmp_device_info ) ) ) 
  {
   DBG( "Current link policy: %s\n", hci_lptostr( tmp_device_info.link_policy ) );
   DBG( "Current link mode  : %s\n", hci_lmtostr( tmp_device_info.link_mode ) );

   if( aEnable )
   {
    if( !( tmp_device_info.link_policy & HCI_LP_RSWITCH ) )
    { // Not set
     tmpNewPolicy = tmp_device_info.link_policy | HCI_LP_RSWITCH;
    }

    if( tmp_device_info.link_mode & HCI_LM_MASTER )
    {
     tmpNewLinkMode = ( tmp_device_info.link_mode & ~HCI_LM_MASTER );// | HCI_LM_ACCEPT;
    }
   }
   else
    if( tmp_device_info.link_policy & HCI_LP_RSWITCH )
    {
     tmpNewPolicy = tmp_device_info.link_policy & ~HCI_LP_RSWITCH;
    }

    if( !( tmp_device_info.link_mode & HCI_LM_MASTER ) )
    {
     tmpNewLinkMode = tmp_device_info.link_mode | HCI_LM_MASTER;
    }

#ifdef ENABLE_ROLE_SWITCH
    if( -1 != tmpNewPolicy )
    {
	 struct hci_dev_req dr;

	 dr.dev_id  = DEFAULT_HCI_DEVICE;
	 dr.dev_opt = tmpNewPolicy;

 	 if( ioctl( a2dp->device_ctl, HCISETLINKPOL, (unsigned long) &dr ) < 0 ) 
     {
	  DBG( "Can't set link policy on hci%d: %s (%d)\n", dr.dev_id, strerror(errno), errno );
     }
    }
#endif

  /*  if( -1 != tmpNewLinkMode )
    {
	 struct hci_dev_req dr;

	 dr.dev_id  = DEFAULT_HCI_DEVICE;
	 dr.dev_opt = tmpNewLinkMode;

 	 if( ioctl( a2dp->device_ctl, HCISETLINKMODE, (unsigned long) &dr ) < 0 ) 
     {
	  DBG( "Can't set link mode on hci%d: %s (%d)\n", dr.dev_id, strerror(errno), errno );
     }
    }*/

    if( !ioctl( a2dp->device_ctl, HCIGETDEVINFO, (void *)&( tmp_device_info ) ) ) 
    {
     DBG( "New link policy: %s\n", hci_lptostr( tmp_device_info.link_policy ) );
     DBG( "New link mode  : %s\n", hci_lmtostr( tmp_device_info.link_mode ) );
    }
    else
    {
     DBG( "Can not get dev info: %d %s\n", errno, strerror( errno ) );
    }
 }
 else
 {
  DBG( "Can not get dev info: %d %s\n", errno, strerror( errno ) );
 }
}

a2dp_packager_t *InitA2dp( packager_t *p, const char *dstaddr, char* aErrorText ) 
{
  assert( NULL != p );
  assert( NULL != dstaddr );
  assert( NULL != aErrorText );

  a2dp_packager_t *a2dp = NULL; 
  
  InitError( aErrorText );

  a2dp = malloc( sizeof( a2dp_packager_t ) );
  if( a2dp == NULL ) 
  {
      AppendError( "Can not allocate memory for A2DP protocol stack", 1 );
      goto error;
  }

  memset( a2dp, 0, sizeof( a2dp_packager_t ) ); 
  memset(&(a2dp->packet_header),0,sizeof(struct media_packet_header));
  memset(&(a2dp->payload_header),0,sizeof(struct media_payload_header));

  a2dp->mtu       = 0;
  a2dp->timestamp = 0;
  a2dp->seq_num   = 1;

  a2dp->cmdfd     = -1;
  a2dp->streamfd  = -1;

  bacpy(&(a2dp->src), BDADDR_ANY);  

  str2ba(dstaddr, &(a2dp->dst));

  p->udata=(void*)a2dp;

  iLastA2DPErrorCode = A2DP_ERROR_UNKNOWN;

  // Getting default device info
  if( ( a2dp->device_ctl = socket( AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI ) ) < 0 ) 
  {
   AppendError( "Can't open HCI socket", 1 );
   goto error;
  }

  a2dp->device_info.dev_id = DEFAULT_HCI_DEVICE;

  if( ioctl( a2dp->device_ctl, HCIGETDEVINFO, (void *)&( a2dp->device_info ) ) ) 
  {
   AppendError( "Can't get device info", 1 );
   goto error;
  }

  DBG( "Link policy: %s\n", hci_lptostr( a2dp->device_info.link_policy ) );
  DBG( "Link mode  : %s\n", hci_lmtostr( a2dp->device_info.link_mode ) );

  // Start listening immediately
  ListenA2dp( a2dp, dstaddr );

  DBG( "Detecting a2dp on headset %s...\n", dstaddr );
  if( !DetectA2dp( p ) ) 
  {
    if( a2dp->cmdfd >= 0 )   
    { // Already have opened the connection
      goto ok;
    }
    goto error;
  }

  DBG( "Starting a2dp command channel...\n" );
  if( !StartA2dpCommand( p ) ) goto error;
  if( !StartA2dpStream( p, 0 ) ) goto error;

  DBG( "Checking link quality..." );
  if( !CheckA2DPLinkQuality( p, 1 ) ) 
  {
   AppendError( "Bad link quality\n", 0 );
   goto error;
  }
  
  // Closing command channel
  CloseA2dp( p );
  assert( -1 == a2dp->cmdfd );
  assert( -1 == a2dp->streamfd );

/*
  DBG( "Starting a2dp stream channel...\n" );
  if( !StartA2dpStream( p ) ) goto error;
 
  // now initialize the size fields
  p->packetlen   = getA2dpPacketLen( p );
  a2dp->nrframes = getA2dpNrFrames( p );

  DBG("frame duration: %lums\n",p->encoder->sbc.duration * a2dp->nrframes);
  DBG("timer: %lums\n",(p->encoder->sbc.duration * a2dp->nrframes)/2);
*/

ok:
  DBG( "A2dp initialized\n" );

  return a2dp;
 
error:
  DBG("error, freeing memory in A2Dp init...\n");  

  if( NULL != a2dp ) 
  {
    StopListen( &( a2dp->listen_data ) );
    
    close( a2dp->device_ctl );

    if( a2dp->cmdfd != -1 ) {
      CloseA2dp( p );
      assert( -1 == a2dp->cmdfd );
      assert( -1 == a2dp->streamfd );
    } else {
      AppendError( "Unable to open command or stream channel\n", 0 );
    }
    free( a2dp );
  }
  else
  {
   aErrorText = "Can not allocate memory";
  }

  p->udata = NULL;

  return NULL;
}

int CreateA2dpPackage(void *packager) {
  packager_t *p=(packager_t *)packager;
  a2dp_packager_t *a=(a2dp_packager_t *)(p->udata);
  int ec;
  int fsize,free,nrframes;
  int size,i;
  unsigned char *packetptr;
  unsigned char *packetstart;

  assert( NULL != p->in );
  assert( NULL != p->out );
  assert( NULL != p->encoder );
  
  //DBG( "Packager %p\n", packager );

  free=GetStreamBufferLength(p->out) - p->bytesWritten;

  if( p->packetlen > free ) return PACKAGER_FLUSH;

  fsize=p->encoder->framelength(p->encoder);

	// number of frames is detemined by what fits in the MTU after the headers
  nrframes=a->nrframes;
  size=p->packetlen;

  memset(&(a->packet_header), 0, sizeof(a->packet_header));
  a->packet_header.v               = 2;
  a->packet_header.pt              = 0x60; //1;
  a->packet_header.sequence_number = htons( a->seq_num );
  a->packet_header.timestamp       = htonl( a->timestamp );
  a->packet_header.ssrc            = htonl( 1 );

  //a->timestamp += (p->encoder->blocks + 1)*4 * (p->encoder->subbands + 1)*4;
  //a->timestamp += 128;
  a->timestamp += 224 * nrframes;
  a->seq_num++;

  a->payload_header.frame_count=nrframes;

  packetstart=packetptr=GetStreamBufferPtr(p->out) + p->bytesWritten; 

  memcpy(packetptr,&(a->packet_header),sizeof(struct media_packet_header));
  packetptr += sizeof(struct media_packet_header);
  p->bytesWritten += sizeof(struct media_packet_header);
  
  memcpy(packetptr,&(a->payload_header),sizeof(struct media_payload_header));
  packetptr += sizeof(struct media_payload_header);
  p->bytesWritten += sizeof(struct media_payload_header);
  
  for(i=0;i<nrframes;i++) {
    ec=p->encoder->encode(p->encoder, p->in, packetptr);

    switch(ec) {
      case ENCODER_CONTINUE:
        assert(p->encoder->sbc.len == fsize);
        if(__builtin_expect((p->packetStream!=NULL),1)) {
          if(fwrite(packetptr,1,fsize,p->packetStream)!=((size_t)(fsize))) {
            fprintf(stderr,"Error writing to packet stream\n");
          }
        }
        packetptr += fsize;
        p->bytesWritten += fsize;
        break;
      case ENCODER_EOS:
        return PACKAGER_EOS;
      case ENCODER_UNDERRUN:
	return PACKAGER_UNRECOVERABLE;
      case ENCODER_TERM:
        return PACKAGER_UNRECOVERABLE;
      default:
        return PACKAGER_UNRECOVERABLE;
    }
  }


  return PACKAGER_CONTINUE;
}

void DoneA2dp(void *packager) 
{
 packager_t                *p=(packager_t *)packager;
 a2dp_packager_t           *h=(a2dp_packager_t *)(p->udata);
 char                       tmpError[ 2048 ];

 assert( NULL != p );
 assert( NULL != h );

 InitError( tmpError );

 // Closing all channels
 CloseA2dp( p );
 
 close( h->device_ctl );
 
 // reset all fields
 // memset( h, 0, sizeof( a2dp_packager_t ) );
 free( h );
 free( packager );
}

int PauseA2dp( void *packager, char* aErrorText )
{
 packager_t                *p=(packager_t *)packager;
 a2dp_packager_t           *h=(a2dp_packager_t *)(p->udata);
 struct suspend_stream_cmd  suspend_stream;
 struct suspend_stream_rsp  suspend_resp;
 int                        tmpResult = 1;
 char                       errorText[ 2048 ];

 assert( NULL != p );
 assert( NULL != h );

 InitError( aErrorText );

 h->paused  = 1;
 h->started = 0;

 // Suspending stream
 memset( &suspend_stream, 0, sizeof( suspend_stream ) );
 init_request( packager, &suspend_stream.header, AVDTP_SUSPEND );
 suspend_stream.acp_seid = h->seid;

 if( write( h->cmdfd, &suspend_stream, sizeof( suspend_stream ) ) != sizeof( suspend_stream ) ) 
  {
   sprintf( errorText, "Couldn't suspend stream SEID = %d", h->seid );                
   AppendError( errorText, 1 );
  } 
 else 
 {
  DBG( "Sent suspend stream command\n" );
 }

 if( read( h->cmdfd, &suspend_resp, sizeof( suspend_resp ) ) < ( int )sizeof( suspend_resp ) - 2 ||
     suspend_resp.header.message_type == MESSAGE_TYPE_REJECT ) 
  {
    sprintf( errorText, "Didn't receive supsend response confirm for SEID = %d\n", h->seid );
    AppendError( errorText, 0 );
    tmpResult = 0;
  }
  else
  {
   DBG( "Got suspend stream confirm\n" );
   //close( h->streamfd );
   //h->streamfd = 0;
  }

 return tmpResult;
}


int ResumeA2dp( void *packager, char* aErrorText )
{
 InitError( aErrorText );

 return StartA2dpStream( packager, 1 );
}


static void init_responce( 
              const struct avdtp_header* aCommandHeader, 
                    struct avdtp_header* aResponceHeader,
                    int                  aResponceStatus )
{
 assert( NULL != aCommandHeader );
 assert( NULL != aResponceHeader );
 //assert( MESSAGE_TYPE_ACCEPT == aResponceStatus || MESSAGE_TYPE_REJECT == aResponceStatus );

 memcpy( aResponceHeader, aCommandHeader, sizeof( *aResponceHeader ) );
 aResponceHeader->packet_type  = PACKET_TYPE_SINGLE;
 aResponceHeader->message_type = aResponceStatus;
}


int ProcessA2DPSinkRequests( void *packager )
{
 struct pollfd           poll_data;
 a2dp_generic_command_t  tmpCommand;
 //uint8_t                 padding[ MAX_PACKET_SIZE ]; // Do not delete this !!!!!
 a2dp_generic_responce_t tmpResponce;
 int                     tmpResult = 0, tmpBytesRead = 0;
 char                    tmpErrorText[ 2048 ];
 int                     tmpPollResult;

 //DBG( "Begin pack %p\n", packager );

 assert( NULL != packager );

 packager_t*             p = ( packager_t * )packager;
 assert( NULL != p );

 a2dp_packager_t*        h = ( a2dp_packager_t * )( p->udata );
 assert( NULL != h );

 if( h->cmdfd < 0 )
 { // No connection yet
  return 0;
 }

 InitError( tmpErrorText ); 

 //DBG( "Begin 2 pack %p\n", packager );

 // Checking link quality
 if( h->link_check++ > LINK_CHECK_PERIOD )
 {
  h->link_check = 0;
  if( !CheckA2DPLinkQuality( p, 0 ) )
  {
   return AVDTP_ABORT;
  }
 }
  
 //DBG( "Begin 3 pack %p\n", packager );
 
 // Polling command connection
 poll_data.fd     = h->cmdfd;
 poll_data.events = POLLIN;

 if( ( tmpPollResult = poll( &poll_data, 1, 0 ) ) > 0 )
 { // There are some data
  if( POLLHUP == poll_data.revents )
  {
   DBG( "BT Socket disconnetced, exiting...\n" );
   return AVDTP_ABORT;
  }

#ifdef ENABLE_PACKET_FRAGMENTATION
  // Reading header
  if( ( tmpBytesRead = ReassembleAVDTPSignalPacket( 
                        h, 
                        &( tmpCommand.command_header ), 
                        sizeof( tmpCommand ) ) ) 
                        >= ( int )sizeof( tmpCommand.command_header ) )
#else
  if( ( tmpBytesRead = read( h->cmdfd, &tmpCommand, sizeof( tmpCommand ) ) ) >= ( int )sizeof( tmpCommand.command_header ) )
#endif
  { // Command read
   int tmpResponceSize = 0;
   int tmpResponceCode = 0;

   memset( &tmpResponce, 0, sizeof( tmpResponce ) );

   DBG( "Received SINK A2DP command code 0x%x, msg type %d, read %d bytes\n", tmpCommand.command_header.signal_id, tmpCommand.command_header.message_type, tmpBytesRead );

   if( MESSAGE_TYPE_COMMAND != tmpCommand.command_header.message_type )
   {
    DBG( "Incorrect message type, ignoring\n" );
   }
   else
   switch( tmpCommand.command_header.signal_id )
   {
    case AVDTP_DISCOVER:         
     DBG( "SINK: AVDTP_DISCOVER, Stream FD %d\n", h->streamfd );   

     if( tmpBytesRead == sizeof( tmpCommand.cmd_sepd_req ) )
     {
      if( h->configured || -1 != h->streamfd || h->started )
      {
       DBG( "Wrong state, rejecting\n" );

       tmpResponceSize                         = sizeof( tmpResponce.resp_close_stream_cmd );
       tmpResponceCode                         = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_close_stream_cmd.error = AVDTP_ERROR_BAD_STATE;//0xC0;
      }
      else
      {
       tmpResponceSize = sizeof( tmpResponce.responce_header ) + 
                       sizeof( struct acp_seid_info );
       tmpResponceCode = MESSAGE_TYPE_ACCEPT;
       tmpResponce.resp_sepd_cmd.infos[ 0 ].inuse0     = ( h->streamfd == -1 ) ? 0: 1;
       tmpResponce.resp_sepd_cmd.infos[ 0 ].acp_seid   = INT_SEID;
       tmpResponce.resp_sepd_cmd.infos[ 0 ].media_type = AUDIO_MEDIA_TYPE;
       tmpResponce.resp_sepd_cmd.infos[ 0 ].tsep       = END_POINT_SOURCE; 
      }
     }
     else
     {
      DBG( "Wrong command length %d, need %d, rejecting\n", tmpBytesRead, sizeof( tmpCommand.cmd_sepd_req ) );

      tmpResponceSize                         = sizeof( tmpResponce.resp_close_stream_cmd );
      tmpResponceCode                         = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_close_stream_cmd.error = AVDTP_ERROR_BAD_LENGTH;
     }
     break;

    case AVDTP_GET_CAPABILITIES:   
    case AVDTP_GET_CONFIGURATION:    
     DBG( "SINK: AVDTP_GET_CAPABILITIES or CONFIG, SEID %d\n", tmpCommand.cmd_getcap_req.acp_seid );   

     if( tmpBytesRead != sizeof( tmpCommand.cmd_getcap_req ) )
     {
      DBG( "Wrong command length, rejecting\n" );
      
      tmpResponceSize                         = sizeof( tmpResponce.resp_close_stream_cmd );
      tmpResponceCode                         = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_close_stream_cmd.error = AVDTP_ERROR_BAD_LENGTH;
      break;
     }

     if( AVDTP_GET_CAPABILITIES == tmpCommand.command_header.signal_id )
     {
      if( ( h->configured || -1 != h->streamfd || h->started )  )
      {
cap_err:
       DBG( "Wrong state, rejecting\n" );
      
       tmpResponceSize                         = sizeof( tmpResponce.resp_close_stream_cmd );
       tmpResponceCode                         = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_close_stream_cmd.error = AVDTP_ERROR_BAD_STATE; // 0xC0; //
       break;
      }
     }
     else
     {
      if( !h->configured )
      {
       goto cap_err;
      }
     }

     if( INT_SEID == tmpCommand.cmd_getcap_req.acp_seid )
     {
      tmpResponceSize                              = sizeof( tmpResponce.resp_getcap_cmd );
      tmpResponceCode                              = MESSAGE_TYPE_ACCEPT;
      tmpResponce.resp_getcap_cmd.serv_cap         = MEDIA_TRANSPORT_CATEGORY;
      tmpResponce.resp_getcap_cmd.serv_cap_len     = 0;                                                       
      tmpResponce.resp_getcap_cmd.cap_type         = MEDIA_CODEC_CATEGORY;
      tmpResponce.resp_getcap_cmd.length           = 6;
      tmpResponce.resp_getcap_cmd.media_type       = AUDIO_MEDIA_TYPE;
      tmpResponce.resp_getcap_cmd.media_codec_type = SBC_MEDIA_CODEC_TYPE;
      fill_sbc_capabilities( &( tmpResponce.resp_getcap_cmd.sbc_elements ), p, MIN_BITPOOL, MAX_BITPOOL );
     }
     else
     {
      tmpResponceSize                         = sizeof( tmpResponce.resp_close_stream_cmd );
      tmpResponceCode                         = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_close_stream_cmd.error = AVDTP_ERROR_BAD_SEID;
     }
     break;

    case AVDTP_SET_CONFIGURATION: 
     DBG( "SINK: AVDTP_SET_CONFIGURATION, SEID %d, INT SEID %d\n", tmpCommand.cmd_set_config.acp_seid, tmpCommand.cmd_set_config.int_seid );   

     if( tmpBytesRead < ( int )sizeof( tmpCommand.cmd_set_config ) )
     {
      DBG( "Wrong command length, rejecting\n" );

      tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_BAD_LENGTH;
      tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_CODEC_CATEGORY;
      break;
     }

     if( tmpCommand.cmd_set_config.serv_cap_len )
     {
      DBG( "Wrong media_transport cap length %d, should be %d, rejecting\n", tmpCommand.cmd_set_config.serv_cap_len, 0 );

      tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_BAD_MEDIA_TRANSPORT;
      tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_TRANSPORT_CATEGORY;
      break;
     }

     /*{
      tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_set_config_cmd.error_code = 0xC0;
      tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_TRANSPORT_CATEGORY;
      break;
     }*/

     if( INT_SEID == tmpCommand.cmd_set_config.acp_seid )
     {
      if( h->configured )
      {
       DBG( "SEP already configured, rejecting\n" );

       tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
       tmpResponceCode                            = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_SEP_IN_USE;
       tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_TRANSPORT_CATEGORY;
       break;
      }

      if( !h->started && -1 == h->streamfd  )
      {
       if( process_cap_resp( 
            tmpBytesRead - sizeof( tmpCommand.cmd_set_config.header ) - 2 * sizeof( uint8_t ),
             ( struct getcap_resp *)( ( char * )&( tmpCommand.cmd_set_config ) + 2 * sizeof( uint8_t ) ) ) )
       {       
        tmpCommand.cmd_set_config.sbc_elements.min_bitpool = MIN_BITPOOL;
        tmpCommand.cmd_set_config.sbc_elements.max_bitpool = MAX_BITPOOL;
        fill_sbc_capabilities( &( tmpResponce.resp_getcap_cmd.sbc_elements ), p, MIN_BITPOOL, MAX_BITPOOL );

        if( memcmp( &( tmpCommand.cmd_set_config.sbc_elements ),
                    &( tmpResponce.resp_getcap_cmd.sbc_elements ),
                    sizeof( tmpResponce.resp_getcap_cmd.sbc_elements ) ) )  
        { // Error     
         tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
         tmpResponceCode                            = MESSAGE_TYPE_REJECT;
         tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_UNSUPPORTED_CONFIGURATION;
         tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_CODEC_CATEGORY;
        }
        else
        {
         tmpResponceSize = sizeof( tmpResponce.resp_set_config_cmd ) - 
                          sizeof( tmpResponce.resp_set_config_cmd.error_code ) - 
                          sizeof( tmpResponce.resp_set_config_cmd.serv_cat );
         tmpResponceCode = MESSAGE_TYPE_ACCEPT;
         h->psm_cmd      = CMD_PSM;
         h->psm_stream   = CMD_PSM;
         h->seid         = tmpCommand.cmd_set_config.int_seid;
         h->configured   = 1;
        }
       }
       else
       {
         tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
         tmpResponceCode                            = MESSAGE_TYPE_REJECT;
         tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_UNSUPPORTED_CONFIGURATION;
         tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_CODEC_CATEGORY;
       }
/*       tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
       tmpResponceCode                            = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_SEP_IN_USE;
       tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_TRANSPORT_CATEGORY;*/
      }
      else
      { // error
       tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
       tmpResponceCode                            = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_SEP_IN_USE;
       tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_TRANSPORT_CATEGORY;
      }
     }
     else
     { // error
       tmpResponceSize                            = sizeof( tmpResponce.resp_set_config_cmd );
       tmpResponceCode                            = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_BAD_SEID;
       tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_TRANSPORT_CATEGORY;
     }
     break;

/*    case AVDTP_GET_CONFIGURATION:
     DBG( "SINK: AVDTP_GET_CONFIGURATION, SEID %d\n", tmpCommand.cmd_get_config.acp_seid );           

     if( tmpBytesRead != sizeof( tmpCommand.cmd_get_config ) )
     {
      DBG( "Wrong command length, rejecting\n" );
      
      tmpResponceSize                            = sizeof( struct avdtp_header ) + sizeof( uint8_t );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_get_config_cmd.error_code = AVDTP_ERROR_BAD_LENGTH;
      break;
     }

     if( !h->configured )
     {
      DBG( "Wrong state, rejecting\n" );

      tmpResponceSize                            = sizeof( struct avdtp_header ) + sizeof( uint8_t );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_get_config_cmd.error_code = AVDTP_ERROR_BAD_STATE; //0xC0; //
      break;
     }

     if( INT_SEID == tmpCommand.cmd_get_config.acp_seid )
     {
      fill_sbc_capabilities( &( tmpResponce.resp_get_config_cmd.serv_cap.sbc_elements ), p, MIN_BITPOOL, MAX_BITPOOL );
      tmpResponceSize = sizeof( tmpResponce.resp_get_config_cmd );
      tmpResponceCode = MESSAGE_TYPE_ACCEPT;
     }
     else
     { // error
      tmpResponceSize                            = sizeof( struct avdtp_header ) + sizeof( uint8_t );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_get_config_cmd.error_code = AVDTP_ERROR_BAD_SEID;
     }
     break;*/

    case AVDTP_RECONFIGURE:
     DBG( "SINK: AVDTP_RECONFIGURE\n" );                   
     
     tmpResponceSize                            = sizeof( tmpResponce.resp_reconfigure_cmd );
     tmpResponceCode                            = MESSAGE_TYPE_REJECT;
     tmpResponce.resp_set_config_cmd.error_code = AVDTP_ERROR_COMMAND_NOT_SUPPORTED;
     tmpResponce.resp_set_config_cmd.serv_cat   = MEDIA_TRANSPORT_CATEGORY;
     break;

    case AVDTP_OPEN:
     DBG( "SINK: AVDTP_OPEN, SEID %d\n", tmpCommand.cmd_open_stream.acp_seid );           

     if( tmpBytesRead != sizeof( tmpCommand.cmd_open_stream ) )
     {
      DBG( "Wrong command length, rejecting\n" );

      tmpResponceSize                          = sizeof( tmpResponce.resp_open_stream_cmd );
      tmpResponceCode                          = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_open_stream_cmd.error   = AVDTP_ERROR_BAD_LENGTH;
      break;
     }

     //if( tmpBytesRead != sizeof( tmpCommand.cmd_open_stream ) )
     /*{
      //DBG( "Wrong command length, rejecting\n" );

      tmpResponceSize                          = sizeof( tmpResponce.resp_open_stream_cmd );
      tmpResponceCode                          = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_open_stream_cmd.error   = 0xC0;
      break;
     }*/
     
     if( INT_SEID == tmpCommand.cmd_open_stream.acp_seid )
     {
      if( -1 == h->streamfd && h->configured )
      {
       if( StartA2dpStream( p, 0 ) )
       {
        tmpResponceSize                        = sizeof( tmpResponce.resp_open_stream_cmd ) - 
                                                 sizeof( uint8_t );
        tmpResponceCode                        = MESSAGE_TYPE_ACCEPT;       
        assert( h->streamfd >= 0 );
        h->started                             = 0;
       }
       else
       {
        tmpResponceSize                              = sizeof( tmpResponce.resp_open_stream_cmd );
        tmpResponceCode                              = MESSAGE_TYPE_REJECT;
        tmpResponce.resp_open_stream_cmd.error       = AVDTP_ERROR_BAD_HEADER_FORMAT;
       }
      }
      else
      {
       tmpResponceSize                              = sizeof( tmpResponce.resp_open_stream_cmd );
       tmpResponceCode                              = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_open_stream_cmd.error       = AVDTP_ERROR_BAD_STATE;
      }
     }
     else
     {
      tmpResponceSize                              = sizeof( tmpResponce.resp_open_stream_cmd );
      tmpResponceCode                              = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_open_stream_cmd.error       = AVDTP_ERROR_BAD_SEID;
     }
     break;

    case AVDTP_START:
     DBG( "SINK: AVDTP_START, SEID %d, CMD %d STM %d\n", tmpCommand.cmd_start_stream.acp_seid, h->cmdfd, h->streamfd );           

     if( tmpBytesRead != sizeof( tmpCommand.cmd_start_stream ) )
     {
      DBG( "Wrong command length, rejecting\n" );

      tmpResponceSize                            = sizeof( tmpResponce.resp_start_stream_cmd );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_start_stream_cmd.error    = AVDTP_ERROR_BAD_LENGTH;
      tmpResponce.resp_start_stream_cmd.acp_seid = tmpCommand.cmd_start_stream.acp_seid;
      break;
     }

     /*{
      //DBG( "Wrong command length, rejecting\n" );

      tmpResponceSize                          = sizeof( tmpResponce.resp_start_stream_cmd );
      tmpResponceCode                          = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_start_stream_cmd.error   = 0xC0;
      tmpResponce.resp_start_stream_cmd.acp_seid = tmpCommand.cmd_start_stream.acp_seid;
      break;
     }*/

     if( INT_SEID == tmpCommand.cmd_start_stream.acp_seid )
     {
      if( -1 != h->streamfd && !h->started && h->configured )   
      {
       tmpResponceSize                            = sizeof( tmpResponce.resp_start_stream_cmd ) - 
                                                     2 * sizeof( uint8_t );
       tmpResponceCode                            = MESSAGE_TYPE_ACCEPT;
       h->started                                 = 1;
      }
      else
      {
       tmpResponceSize                            = sizeof( tmpResponce.resp_start_stream_cmd );
       tmpResponceCode                            = MESSAGE_TYPE_REJECT;
       tmpResponce.resp_start_stream_cmd.acp_seid = tmpCommand.cmd_start_stream.acp_seid;
       tmpResponce.resp_start_stream_cmd.error    = AVDTP_ERROR_BAD_STATE;
      }
     }
     else
     {
      tmpResponceSize                              = sizeof( tmpResponce.resp_start_stream_cmd );
      tmpResponceCode                              = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_start_stream_cmd.error      = AVDTP_ERROR_BAD_SEID;
      tmpResponce.resp_start_stream_cmd.acp_seid   = tmpCommand.cmd_start_stream.acp_seid;
     }
     break;

    case AVDTP_CLOSE:
     DBG( "SINK: AVDTP_CLOSE, SEID %d\n", tmpCommand.cmd_close_stream.acp_seid );           

     if( tmpBytesRead != sizeof( tmpCommand.cmd_close_stream ) )
     {
      DBG( "Wrong command length, rejecting\n" );

      tmpResponceSize                            = sizeof( tmpResponce.resp_close_stream_cmd );
      tmpResponceCode                            = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_close_stream_cmd.error    = AVDTP_ERROR_BAD_LENGTH;      
      break;
     }

      /*{
      //DBG( "Wrong command length, rejecting\n" );

      tmpResponceSize                          = sizeof( tmpResponce.resp_close_stream_cmd );
      tmpResponceCode                          = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_close_stream_cmd.error  = 0xC0;
      break;
     }*/

     if( INT_SEID == tmpCommand.cmd_close_stream.acp_seid )
     {
      if( -1 != h->streamfd )   
      {
       tmpResponceSize = sizeof( tmpResponce.resp_close_stream_cmd ) - 
                                                sizeof( uint8_t );
       tmpResponceCode = MESSAGE_TYPE_ACCEPT;
       tmpResult       = AVDTP_CLOSE;
       h->started      = 0;
       h->configured   = 0;
      }
      else
      {
       tmpResponceSize                            = sizeof( tmpResponce.resp_close_stream_cmd );
       tmpResponceCode                            = MESSAGE_TYPE_REJECT;       
       tmpResponce.resp_close_stream_cmd.error    = AVDTP_ERROR_BAD_STATE;
      }
     }
     else
     {
      tmpResponceSize                         = sizeof( tmpResponce.resp_close_stream_cmd );
      tmpResponceCode                         = MESSAGE_TYPE_REJECT;
      tmpResponce.resp_close_stream_cmd.error = AVDTP_ERROR_BAD_SEID;
     }
     break;

    case AVDTP_SUSPEND:
     DBG( "SINK: AVDTP_SUSPEND\n" );           

     tmpResponceSize                              = sizeof( tmpResponce.resp_suspend_stream_cmd );
     tmpResponceCode                              = MESSAGE_TYPE_REJECT;
     tmpResponce.resp_suspend_stream_cmd.acp_seid = tmpCommand.cmd_suspend_stream.acp_seid;
     tmpResponce.resp_suspend_stream_cmd.error    = AVDTP_ERROR_COMMAND_NOT_SUPPORTED;
     break;

    case AVDTP_ABORT:
     DBG( "SINK: AVDTP_ABORT\n" );           

     tmpResponceSize = sizeof( tmpResponce.resp_abort_cmd );
     tmpResponceCode = MESSAGE_TYPE_ACCEPT;
     tmpResult       = AVDTP_ABORT;
     h->started      = 0;
     h->configured   = 0;
     break;

    default:
     DBG( "Wrong A2DP command code %d, issuing GENERAL REJECT\n", tmpCommand.command_header.signal_id );

     // General reject
     tmpResponceSize                         = sizeof( tmpResponce.resp_close_stream_cmd );
     tmpResponceCode                         = 0;
     tmpResponce.resp_close_stream_cmd.error = 0;
     tmpCommand.command_header.signal_id     = 0;
   }

   DBG( "Responce size %d, state %d (accept %d), result %d\n", tmpResponceSize, tmpResponceCode, MESSAGE_TYPE_ACCEPT, tmpResult );
   
   if( tmpResponceSize )
   { // Writing response
    init_responce( &( tmpCommand.command_header ), 
                   &( tmpResponce.responce_header ),
                   tmpResponceCode );
    
#ifdef DEBUG
    fprintf( stderr, "Responce: " );
    {
     int i = 0;
     for( i = 0; i < tmpResponceSize; ++i )
     {
      fprintf( stderr, "0x%x ", ( ( char* )&tmpResponce )[ i ] );
     }
    }
    fprintf( stderr, "\n" );
#endif

    errno = 0;
    
#ifdef ENABLE_PACKET_FRAGMENTATION
    tmpBytesRead = SendAVDTPSignalPacket( h->signalling_mtu, h->cmdfd, ( char *)( &tmpResponce ), tmpResponceSize );
#else
    tmpBytesRead = write( h->cmdfd, &tmpResponce, tmpResponceSize );
#endif
    DBG( "Written %d bytes, errno %d, %s\n", tmpBytesRead, errno, strerror( errno ) );
   }
  }
  else
  {
#ifdef DEBUG   
   if( tmpBytesRead >= 0 )
   {
    DBG( "Wrong bytes read count %d\n", tmpBytesRead );
   }
#endif
  }
 }
 else
 {
  if( tmpPollResult < 0 )
  {
   DBG( "Poll error %d, errno %d, %s, closing command channed\n", tmpPollResult, errno, strerror( errno ) );

   char tmpErrorText[ 2048 ];

   InitError( tmpErrorText );

   CloseA2dp( packager );

   RestartListenA2dp( h );
  }
 }

 //DBG( "End, res %d\n", tmpResult );

 return tmpResult;
}

/* EOF */
