/* 
 * File   : Encoder.c
 * Author : Koen Martens <koen.martens@tomtom.com>
 * Desc.  : Wrapper for the SBC encoder plus some helper functions
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <dlfcn.h>
#include <assert.h>
#include <unistd.h>

#include "Debug.h"
#include "Encoder.h"
#include "Packager.h"
#include "A2dp.h"
#include "DynamicSBC.h"

#define RESUME_EMPTY_PACKETS  1400
#define DISCARD_EMPTY_PACKETS 200

static int DoEncode( void *encoder,StreamBufferHandle *in,void *out );
static void DoneEncoder(void *);

// calculates the byterate
static int SBCBitRate( void *encoder )
{
 encoder_t *h=(encoder_t *)encoder;

 // P 12.9 of A2DP
 return ( ( h->framelength( encoder ) * h->rate ) /*<< 3*/ ) / ( h->subbands * h->blocks )
        + 1; // bacuse of bytes, + 1
}

// returns length including frame header
static int SBCFrameLength(void *encoder) {
        encoder_t *h=(encoder_t *)encoder;
        
        int tmp, nrof_subbands, nrof_blocks;

        nrof_subbands = h->subbands; //(hdr->subbands + 1) * 4;
        nrof_blocks = h->blocks;     //(hdr->blocks + 1) * 4;

        switch (h->channel_mode) {
                case MONO:
                        nrof_subbands /= 2;
                        tmp = nrof_blocks * h->bitpool;
                        break;

                case DUAL_CHANNEL:
                        tmp = nrof_blocks * h->bitpool * 2;
                        break;

                case STEREO:
                        tmp = nrof_blocks * h->bitpool;
                        break;

                case JOINT:
                        tmp = nrof_blocks * h->bitpool + nrof_subbands;
                        break;

                default:
                        return 0;
        }

        return (nrof_subbands + ((tmp + 7) / 8)) + sizeof(struct sbc_frame_header);
}


static void* GetDLLSymbol( void* aDLL, const char* aSymbolName, char* aErrorText )
{
 assert( NULL != aDLL );
 assert( NULL != aSymbolName );
 assert( NULL != aErrorText );

 void* tmpAddr = dlsym( aDLL, aSymbolName );
 char* sbc_dl_err;
 
 if( ( sbc_dl_err = dlerror() ) != NULL )  {
    sprintf( aErrorText, "Unable to load symbol %s: %s\n", aSymbolName, sbc_dl_err );
    fprintf( stderr, "%s", aErrorText );
    tmpAddr = NULL;
  }

 return tmpAddr;
}

encoder_t *InitEncoder( int aBitPool, char* aErrorText ) {
  encoder_t   *h;
  TSBCEncoder *t;

  assert( NULL != aErrorText );
  *aErrorText = 0;

  h = malloc(sizeof(encoder_t));
  if( NULL == h ) 
  {
   strcpy( aErrorText, "Can not allocate memory for SBC encoder\n" );
   goto free_and_exit;
  }

  memset( h, 0, sizeof( encoder_t ) );
  
  t = h->functions = malloc( sizeof( TSBCEncoder ) );
  if( NULL == h->functions ) 
  {
   strcpy( aErrorText, "Can not allocate memory for SBC encoder\n" );
   goto free_and_exit;
  }

  // load sbc encoder
  DBG("Loading dynamic object \"%s\"\n", SBCLIB );
  h->sbc_dl = dlopen(SBCLIB,RTLD_NOW);
  if(!h->sbc_dl) {
    sprintf( aErrorText, "SBC dynamic library %s not loaded: %s\n", SBCLIB, dlerror() ); 
    fprintf( stderr,"%s", aErrorText );
    goto free_and_exit;
  }

  t->iSBCInputType = GetDLLSymbol( h->sbc_dl, "SBCEncoderInputType", aErrorText );
  t->iSBCInit      = GetDLLSymbol( h->sbc_dl, "SBCInit",             aErrorText );
  t->iSBCEncode    = GetDLLSymbol( h->sbc_dl, "SBCEncode",           aErrorText );
  t->iSBCFinish    = GetDLLSymbol( h->sbc_dl, "SBCFinish",           aErrorText );

  if( NULL == t->iSBCInputType || 
      NULL == t->iSBCInit      ||
      NULL == t->iSBCEncode    ||
      NULL == t->iSBCFinish )
  {
   dlclose( h->sbc_dl );
   goto free_and_exit;
  }

  if( h->functions->iSBCInit( &(h->sbc), SBC_NULL, aBitPool ) == -EIO ) 
  {
      strcpy( aErrorText, "Can not initialise SBC encoder\n" );
      fprintf( stderr, "%s", aErrorText );

      dlclose( h->sbc_dl );

      goto free_and_exit;
  }

  h->framelength = SBCFrameLength;
  h->byterate    = SBCBitRate;
  h->encode      = DoEncode;
  h->destroy     = DoneEncoder;

  h->channels          = h->sbc.channels;
  h->channel_mode      = h->sbc.channel_mode;
  h->rate              = h->sbc.rate;
  h->subbands          = h->sbc.subbands;
  h->blocks            = h->sbc.blocks;
  h->bitpool           = h->sbc.bitpool;
  h->allocation_method = h->sbc.allocation_method;

  h->sbc.len = h->framelength( h );

  h->empty_data = NULL;
  h->resumed    = 1;
  
  return h;

free_and_exit:
  DBG("error, free_and_exit\n");
  if(h->functions) free(h->functions);  
  if(h->sbc_dl) dlclose(h->sbc_dl);
  if( NULL != h->empty_data ) free( h->empty_data );

  free( h );

  return NULL;
}


static int DoEncode(void *encoder,StreamBufferHandle *in,void *out) {
  encoder_t *e=(encoder_t *)encoder;
  int ec;
  void *ptr;

  //DBG( "ReadStreamBuffer()\n" );
  
  // Read, assuming that input buffer is multiple of input_len
read_again:
  
  while( e->paused )
  {
   usleep( 1000 );

#ifdef DEBUG
      static int tmpCount = 0;

      if( tmpCount++ > 400 )
      {
       DBG( "Encoder paused, input buffer fill %d\n", GetStreamBufferAvailable( in ) );
       tmpCount = 0;
      }
#endif
  }

  if( e->resumed || e->discarded )
  {
   if( NULL == e->empty_data )
   {
    e->empty_data = malloc( e->sbc.input_len );

    if( NULL != e->empty_data )
    {
     memset( e->empty_data, 0, e->sbc.input_len );
    }
    else
    {
     DBG( "Fatal: can not allocate memory for empty packet\n" );
    }
   }

   ec  = e->sbc.input_len;
   ptr = e->empty_data;

   if( e->empty_count++ > ( ( e->discarded ) ? DISCARD_EMPTY_PACKETS: RESUME_EMPTY_PACKETS ) )
   {
    e->resumed     = 0;
    e->empty_count = 0;

    DBG( "Finished sending empty packets, input buffer fill %d\n", GetStreamBufferAvailable( in ) );

    if( e->discarded )
    {
     DBG( "Pausing...\n" );
     e->discarded = 0;
     e->paused    = 1;
    }
   }
  }
  else
  {
    ec = ReadStreamBuffer( in, &ptr, e->sbc.input_len );
    if( ec == 0 ) {
     if(StreamBufferWaitFlush(in)) return ENCODER_TERM;
    
     //DBG( "SBC ENCODER BUFFER UNDERFLOW!!!\n" );

     goto read_again;
   }
  }

    // handle errors
  if( ec == -1 ) 
  {
   DBG( "=========!!! EOF for SBC encoder signalled !!!=======\n" );
   return ENCODER_EOS;
  }

  if( ec != e->sbc.input_len ) {
    DBG("======= !!!!!! GOT INPUT UNDERRUN (%i, want %i) discarding data and retrying !!!!!! ======= \n",ec,e->sbc.input_len); 
    //return ENCODER_UNDERRUN;
    goto read_again;
  }

    // prep encoder
  e->sbc.data = out;
  e->sbc.size = e->sbc.len;

    // do the work
  ec = (e->functions->iSBCEncode( &(e->sbc), ptr, e->sbc.input_len ));

  return ENCODER_CONTINUE;
}

static void DoneEncoder(void *encoder) {
  encoder_t *e=(encoder_t *)encoder;

  e->functions->iSBCFinish( &(e->sbc) );
  
  if( NULL != e->functions ) free( e->functions );  

  //DestroySBCEncoder();
  dlclose(e->sbc_dl);  

  if( NULL != e->empty_data )
  {
   free( e->empty_data );
  }

  free(encoder);
}


/* EOF */
