/*      
 * File   : Packager.c
 * Author : Koen Martens <koen.martens@tomtom.com>
 * Desc.  : Packager abstraction
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */  

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <assert.h>
#include "Debug.h"
#include "StreamBuffer.h"
#include "Packager.h"
#include "A2dp.h"

packager_t *InitPackager(
             encoder_t          *encoder,
             const char         *bdaddr,
             FILE               *pstream,
             char               *errorText ) {
  packager_t *p;

  assert(encoder!=NULL);
  assert( NULL != errorText );

  p = malloc(sizeof(packager_t));
  if( NULL == p ) 
  {
      strcpy( errorText, "Can not allocate memory for A2DP protocol stack" );
      return NULL;
  }

  memset( p, 0, sizeof( packager_t ) );
  
  p->createpackage     = CreateA2dpPackage;
  p->destroy           = DoneA2dp;
  p->pause             = PauseA2dp;
  p->resume            = ResumeA2dp;
  p->processpeerevents = ProcessA2DPSinkRequests;
  p->connect           = ConnectA2dp;
  
  p->in      = NULL;
  p->out     = NULL;
  p->encoder = encoder;

  p->bytesWritten = 0;
  p->outFlushed   = 0;

  p->totalFrames  = 0;
  p->missedFrames = 0;
  p->lastFrame    = 0;

  p->terminated = 0;

  p->packetStream = pstream;

  p->udata = InitA2dp( p, bdaddr, errorText ); 

  if(p->udata == NULL) goto free_error;

  return p;

free_error:
  if(p->udata) free(p->udata);
  free(p);
  return NULL;
}


void *packagerProducer(void *args) {
  StreamBufferHandle *h=(StreamBufferHandle *)args;
  packager_t *packager=(packager_t *)(h->udata);
  int ec;

  packager->bytesWritten=0;

  	// wait until output streambuffer is initialized 
  DBG("packager waiting for out buffer\n");
  while(packager->out==NULL) usleep(10000);
  DBG("packager got out buffer\n");

  while(!StreamBufferKillThread(packager->out)) {

    ec = packager->createpackage((void*)packager);

    switch(ec) {
      case PACKAGER_FLUSH:
        if(FlushStreamBuffer(packager->out,packager->bytesWritten)) {
	  packager->bytesWritten=0;
	  UnflushStreamBuffer(packager->out);
          goto flush_and_exit;
        }
        packager->bytesWritten=0;
        break;
      case PACKAGER_CONTINUE:
        break;
      case PACKAGER_EOS:
        goto flush_and_exit;
        break;
      default:
        fprintf(stderr,"undefined packager error\n");
		// fall-through
      case PACKAGER_UNRECOVERABLE:
        fprintf(stderr,"unrecoverable packager error\n");
	goto flush_and_exit;
        break;
    }
  }

flush_and_exit:

  DBG("Packager thread exiting..\n");

  StreamBufferEos(packager->out);

  if( !StreamBufferKillThread( packager->out ) )
  {
   FlushStreamBuffer( packager->out, packager->bytesWritten );
  }

  packager->terminated = 1;
  
  pthread_exit(0);
}

void DonePackager(packager_t *packager) {
  DoneA2dp(packager);
  free(packager);
}

/* EOF */
