#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h> 
#include <netinet/in.h>
#include <sys/time.h>
#include <limits.h>
#include <dlfcn.h>
#include <stdio.h>
#include <string.h>

#include "Debug.h"
#include "AudioAccept.h"
#include "A2DPPlugIn.h"
#include "A2DPErrors.h"
#include "A2dpTypes.h"

extern int ProcessA2DPSinkRequests( void *packager );

#define DEFAULT_BITPOOL           56
#define DEFAULT_SEMAPHORE_NAME    "/var/run/a2dp.sem"
#define CLIENT_PIN_FILE_NAME      "/etc/bluetooth/client_pin"
#define DEFAULT_PIN               "0000"
#define DEFAULT_BT_ADDRESS        "00:0D:44:2A:5E:24" /* HP */

//"00:0b:0d:50:8d:79" /* bluewalker w/o buttons */
//"00:0b:0d:50:97:a4"  // bluewalker w/ buttons

// Audio parameters
#define PCMBUF_SIZE	    ( 64 * 1024 )	// total size of all decoded PCM sample buffers
#define BUF_COUNT       4
#define BUF_SIZE        ( PCMBUF_SIZE / BUF_COUNT )
#define SAMPLING_RATE   44100
#define NUM_CHANNELS    2
#define BITS_PER_SAMPLE 16

typedef struct {
  // General
  int                 niceValue;  
  int                 isDaemon;
  char                lastErrorText[ 2048 ];
  int                 lastError;
  int                 done;
  EPauseMode          pause_mode;
  int                 paused;
  int                 bufnumber; // audio accept buffer number
  char                audiosockname[ 2048 ];
  char                pincode[ 100 ];
  a2dp_mode_t         a2dpMode;
  void*               emptyData;
  int                 emptyDataLen;
  int                 emptyPacketSent;
} main_config_t;

void usage(main_config_t *config) {
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"-s name    audio socket name\n");
  fprintf(stderr,"-a btaddr  Bluetooth device address\n");
  fprintf(stderr,"-d         Daemonize\n");
  fprintf(stderr,"-n ##      Nice value (default %i)\n",config->niceValue);
  exit(0);
}


static main_config_t config;

// Generic setup
static void InitInternalUIController( ui_output_contoller_t* aController );


int main(int argc,char *argv[]) 
{
 ui_output_contoller_t tmpUIController;
 
 int   option;
 char  tmpStr[ 100 ];

 DBG( "Starting %s...\n", argv[ 0 ] );

 strcpy( tmpStr, DEFAULT_BT_ADDRESS );

 InitInternalUIController( &tmpUIController );

 // set defaults on config
 memset( &config, 0, sizeof( config ) );

 strcpy( config.audiosockname, DEFAULT_AUDIO_SOCKET_NAME ); 
 config.niceValue = -10;
 
  while((option=getopt(argc,argv,"ds:n:a:"))!=-1) {
    switch(option) {
     case 'd':
          config.isDaemon=1;
          break;

	 case 's':
	  if(optarg==NULL)
            usage(&config);
          else 
          {
            strncpy(config.audiosockname,optarg,2048);
	        config.audiosockname[2047]=0;
          }
          break;

	 case 'a':
	  if(optarg==NULL)
            usage(&config);
          else 
          {
            strncpy( tmpStr, optarg, 100 );
	        tmpStr[ 99 ] = 0;
          }
          break;

	 case 'n':
	  if(optarg==NULL || atoi(optarg)<-20 || atoi(optarg)>19 )
	    usage(&config);
	  else
	    config.niceValue=atoi(optarg);
	  break;

	default:
	  usage(&config);
	  break;
    }
  } 
	// renice ourselves to not be nice (we need accurate timing),
	// do this _before_ forking off any threads
  errno=0;
  if( (nice(config.niceValue)==-1) && (errno!=0) ) {
    perror(__func__);
    exit(1);
  }


restart: 
 if( AudioAcceptInit( BUF_SIZE, 
                      BUF_COUNT,  
                      DEFAULT_SEMAPHORE_NAME, 
                      config.audiosockname, 
                      1,
                      SAMPLING_RATE,
                      NUM_CHANNELS,
                      BITS_PER_SAMPLE,
		      0 ) < 0)
 {
  fprintf( stderr, "FATAL: Can not initialise AudioAccept\n" );
  exit( 1 );
 }

 InitUIOutputController( &tmpUIController ); 

 /*tmpUIController.iSetBlueToothOutput( 
                    tmpStr,   
                    DEFAULT_BITPOOL,
                    DEFAULT_PIN,
                    ENormalA2DPMode );*/

 config.bufnumber = config.lastError = config.paused = config.emptyPacketSent = 0;
 config.pause_mode = KNoPause;

 // Resetting error flag
 *config.lastErrorText = 0;

 while( !config.done ) 
 {  
    DispatchNextUIOutputMessage( &tmpUIController );      

    // Processing events from sink
    switch( PlugInProcessPeerEvents() )
    {
     case AVDTP_ABORT:
     case AVDTP_CLOSE:
      DBG( "Stopping playback on SINK request\n" );
      tmpUIController.iStopPlayback();
      break;
    }

    usleep( 100000 );

    if( PlugInPlaybackFinished( NULL, NULL ) )
    { // At the end exiting
     DBG( "Playback finished, exiting...\n" );
     config.done = 1;
    }
  }

  DBG( "Deleting output controller\n" );
  // de-init everything (in reverse order)
  DoneUIOutputController( &tmpUIController );

  DBG( "Done A2DP...\n" );
  PlugInDone();  

  DBG( "Done Audio Accept...\n" );
  AudioAcceptDestroy();

  // Resetting all fields
  if( config.isDaemon ) 
  {
    DBG("daemon mode -> restart\n");
    goto restart;
  }  

  DBG( "Exiting...\n" );

  return 0;
}

static int PausePlayback( int aDiscardBuffers  );
static int StopPlayback( void );

static int PCMDataReadFunction( void** ptr, int len )
{
 int tmpRetryCount = 50;
 int tmpResult     = -1;

 if( KDiscard == config.pause_mode )
 {
  DBG( "Discarding data...\n" );
  return 0;
 }

retry:
 tmpResult = AudioAcceptReadData( ptr, len ); 

 if( -1 == tmpResult )
 {
  if( KSync == config.pause_mode || KDiscard == config.pause_mode )
  {
   //config.sync = 0;
   tmpResult = 0;
  }
  else
  {
   if( EAlwaysSendEmptyPacketsA2DPMode == config.a2dpMode )
   { // Sending empty packet
    if( NULL == config.emptyData || config.emptyDataLen < len )
    {
     if( NULL != config.emptyData )
     {
      free( config.emptyData );
      config.emptyData    = NULL;
      config.emptyDataLen = 0;
     }
     config.emptyData = malloc( len );

     if( NULL != config.emptyData )
     {
      memset( config.emptyData, 0, len );
      config.emptyDataLen = len;

      DBG( "Allocated %d bytes for empty SBC packet\n", len );
     }
     else
     {
      DBG( "Fatal: can not allocate memory for empty packet\n" );
     }       
    }

    if( NULL != config.emptyData )
    { 
     *ptr                   = config.emptyData;
     tmpResult              = len;
     config.emptyPacketSent = 1;

     //usleep( 500 ); // Small pause to reduce CPU usage
    }
    else
    { // Error, falling back to normal case
     goto old_mode;
    }
   }
   else
   {
old_mode:
    if( --tmpRetryCount )
    {
     DBG( "Buffer not ready, sleeping\n" );

     usleep( 1000 );
  
     config.emptyPacketSent = 0;
     
     goto retry;
    }
    DBG( "Buffer underflow or stream end !!!\n" );
   }
  }
 }
 else
 {
  config.emptyPacketSent = 0;
 }
 
 //DBG( "ReadData %d, res %d\n", len, tmpResult );

 return tmpResult;
}

// UI controller management
static int SetBlueToothOutput( const char* aAddress, int aQuality, const char* aPin, int aMode )
{
 a2dp_audio_output_plugin_data tmpData; 

 assert( NULL != aAddress );
 assert( NULL != aPin );

#ifdef DEBUG
 if( strcmp( aAddress, "Test" ) )
#endif
 strcpy( tmpData.iBTAddress, aAddress );
 tmpData.iBitPool = aQuality;
  
 *config.lastErrorText = 0;

 // Dynamic change of mode and PIN is allowed
 strncpy( config.pincode, aPin, 100 );
 config.a2dpMode = aMode;

 config.lastError = PlugInInit( EPullModel, &tmpData );

 DBG( "Result %d\n", config.lastError );

 return config.lastError;
}

static int ResumePlayback( void )
{
 if( config.paused )
 {
  if( ( config.lastError = PlugInResumePlayback( config.lastErrorText ) ) )
  {
   config.paused     = 0;
   config.pause_mode = KNoPause;
  }
 }
 else
 {
  strcpy( config.lastErrorText, "A2DP player not in paused state" );
  config.lastError = 0;
 }

 return config.lastError;
}

static int StartPlayback( void )
{ 
 if( config.paused )
 { // Resuming
  return ResumePlayback();    
 }
 else
 {
  FILE* tmpClientPin = NULL;

  // Creating client_pin file
  if( ( tmpClientPin = fopen( CLIENT_PIN_FILE_NAME, "w+b" ) ) != NULL )
  {
   fprintf( tmpClientPin, "%s", config.pincode );
   fclose( tmpClientPin );
  }
  else
  {
   DBG( "Can not rewrite %s file!\n", CLIENT_PIN_FILE_NAME );
  }

  // Start the player
  config.lastError = PlugInStartPlayback( PCMDataReadFunction, config.lastErrorText );

/*  if( !config.lastError )
  { // Error
   DBG( "Start playback error, stopping everything...\n" );
   PlugInStopPlayback();
  }*/
 }

 return config.lastError;
}


static int StopPlayback( void )
{
 *config.lastErrorText = 0;

 DBG( "Stop playback\n" );
 
 /*if( ( */config.lastError = PlugInStopPlayback();// ) )
  {
   config.done = 1;
  }

 return config.lastError;
}

static int InternalPausePlayback( EPauseMode aPauseMode )
{
 DBG( "Pause mode %d\n", aPauseMode );
 
 if( config.paused )
 {
  strcpy( config.lastErrorText, "A2DP player already paused state" );
  config.lastError = 0;
 }
 else
 {
  config.pause_mode = aPauseMode;
  
  if( KDiscard == aPauseMode )
  {
   AudioAcceptDiscardBuffers();
  }

  if( ( config.lastError = PlugInPausePlayback( aPauseMode, config.lastErrorText ) ) )
  {
   config.paused = 1;
  }
 }
 
 return config.lastError;
}

static int SyncPlayback( void )
{
 return InternalPausePlayback( KSync );
}

static int PausePlayback( int aDiscardBuffers )
{
 return InternalPausePlayback( aDiscardBuffers ? KDiscard: KPauseImmediately );
}

int GetLastResult( char** aErrorText, int* aStatus )
{
 assert( NULL != aErrorText );
 assert( NULL != aStatus );

 int tmpPlaybackStatus  = 0;
 int tmpConnectedStatus = 0;

 *aErrorText = config.lastErrorText;
 *aStatus    = A2DP_STATUS_UNKNOWN;

 PlugInPlaybackFinished( &tmpPlaybackStatus, &tmpConnectedStatus );

 if( tmpPlaybackStatus )
 {
  if( config.paused )
  {
   *aStatus = A2DP_STATUS_PAUSED;
  }
  else
   if( config.emptyPacketSent )
   {
    *aStatus = A2DP_STATUS_PLAYING_EMPTY;
   }
   else
   {
    *aStatus = A2DP_STATUS_PLAYING;
   }
 }
 else
  if( tmpConnectedStatus )
  {
   *aStatus = A2DP_STATUS_CONNECTED;
  }

 DBG( "Playing %d, Connected %d, Status %d, Error %d, Str %s\n", 
      tmpPlaybackStatus, 
      tmpConnectedStatus, 
      *aStatus, 
      config.lastError, 
      config.lastErrorText );

 return config.lastError;
}


static void InitInternalUIController( ui_output_contoller_t* aController )
{
 assert( NULL != aController );

 aController->iSetBlueToothOutput         = SetBlueToothOutput; 
 aController->iStartPlayback              = StartPlayback;
 aController->iPausePlayback              = PausePlayback;
 aController->iResumePlayback             = ResumePlayback;
 aController->iStopPlayback               = StopPlayback;
 aController->iSyncPlayback               = SyncPlayback;
 aController->iLastResult                 = GetLastResult;
}





















