#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <limits.h>
#include <dlfcn.h>
#include <time.h>
#include <signal.h>

#include <sysvsem.h>

#ifdef NMP3DEBUG
#undef DEBUG
#endif

#include "Debug.h"
#include "StreamBuffer.h"
#include "ReadFile.h"
#include "MadProducer.h"
#include "UIController.h"
#include "AudioAccept.h"


// As much possible debug output
//#define FULL_DEBUG

#define PCMBUF_SIZE	    ( 96  * 1024 )  // total size of all decoded PCM sample buffers

#define BUF_COUNT       2
#define BUF_SIZE        ( PCMBUF_SIZE / BUF_COUNT )

#define DISCARD_TIMEOUT	5	// timeout (in seconds) for discardAndWait notification

static void DiscardAndWait(void);
static void ResumeDiscardAndWait(void);


typedef struct {
  // General
  int                 niceValue;  
  int                 isDaemon;
  int                 playType;
  char                lastErrorText[ 2048 ];
  char                currentFileName[ 2048 ];
  char                nextFileName[ 2048 ];
  int                 done;
  int                 stop_mode;
  char                audiosockname[2048];
  int                 paused;
  int                 customCodec;
  unsigned int        oldBytePos;
  unsigned long       currentTime;
  unsigned long       currentChapter;

  // debugging option to write out stream to file
  char                pkgFile[ 2048 ];
  FILE               *pkgFileFp;

  // file reader related
  filereader_feature_t currentFeature;
  fileReaderData      *fileReader;
  StreamBufferHandle  *fileReaderBuffer;

  // mp3 decoder related
  MadOpt              madOpt;
  StreamBufferHandle *madProducerBuffer;
  fileReaderData     *mp3_file_reader;

} main_config_t;

void usage(main_config_t *config) {
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"-s foo     Set TTMP socket name (default %s)\n",config->audiosockname);
  fprintf(stderr,"-d         Daemonize\n");
  fprintf(stderr,"-n ##      Nice value (default %i)\n",config->niceValue);
  exit(0);
}

static main_config_t config;

// Generic setup
static void InitInternalUIController( ui_contoller_t* aController );

int main(int argc,char *argv[]) 
{
 ui_contoller_t tmpUIController; 

  // 
  int option;
  //
  int ec;

  DBG( "Starting %s...\n", argv[ 0 ] );
  
  InitInternalUIController( &tmpUIController );

  // set defaults on config
  memset( &config, 0, sizeof( config ) );

  strcpy( config.audiosockname, DEFAULT_AUDIO_SOCKET_NAME ); 
  //config.niceValue = -10;
 
  while((option=getopt(argc,argv,"ds:n:"))!=-1) {
    switch(option) {
        case 'd':
          config.isDaemon=1;
          break;
	case 's':
	  if(optarg==NULL)
            usage(&config);
          else {
            strncpy(config.audiosockname,optarg,2048);
	    config.audiosockname[2047]=0;
	  }
          break;
	case 'n':
	  if(optarg==NULL || atoi(optarg)<-20 || atoi(optarg)>19 )
	    usage(&config);
	  else
	    config.niceValue=atoi(optarg);
	  break;
	default:
	  usage(&config);
	  break;
    }
  } 
	// renice ourselves to not be nice (we need accurate timing),
	// do this _before_ forking off any threads
  errno=0;
  if( (nice(config.niceValue)==-1) && (errno!=0) ) {
    perror(__func__);
    exit(1);
  }

 InitUIController( &tmpUIController ); 

restart: 
 config.done      = 0;
 config.stop_mode = 0;

 // Waiting for first configuration commands
 DBG( "Waiting for configuration commands...\n" );

 while( NULL == config.fileReader && !config.done )
 {
  DispatchNextUIMessage( &tmpUIController, ENoExitIfError );

  if( config.done )
  { // Normal shutdown
   DBG( "Exiting before playback start\n" );
   
   DoneUIController( &tmpUIController );

   DBG( "Exiting...\n" );   
   exit( 0 );
  }

  if( NULL == config.fileReader )
  { // Sleep
   usleep( 10000 );
  }
 }

  if( NULL == config.fileReader || config.done )
  { // Sleep
   DBG( "File Reader not initialised (0x%p) or exit signalled (%d), exiting...\n", config.fileReader, config.done );
   exit( 1 );
  }

 DBG( "File Reader set up success, ptr 0x%p\n", config.fileReader );

 // Resetting error flag
 *config.lastErrorText = 0;

 // !!!!!!!!!!!!!!!! Buffers should be ready after UI init !!!!!!!!!!!!!!!!!!!!!
 assert(config.fileReader != NULL);

 const stream_parameters_t* tmpStreamParameters = config.fileReader->iGetStreamParameters( config.fileReader );
 assert( NULL != tmpStreamParameters );
 
  // StreamBuffer associated with fileReader 
  DBG( "Setting up fileReaderBuffer, size %d, readsize %d, sr %d, ch %d, bps %d, type %d, fb %d...\n", 
       tmpStreamParameters->iBufferSize,
       tmpStreamParameters->iMinReadSize,
       tmpStreamParameters->iSamplingRate,
       tmpStreamParameters->iNumChannels,
       tmpStreamParameters->iBitsPerSample,
       tmpStreamParameters->iBufferType,
       tmpStreamParameters->iFeatureBufferSize );
      
  config.fileReader->udata = &config.madOpt;
  config.fileReaderBuffer=InitStreamBuffer( MAX( DEFAULT_FILEBUF_SIZE, tmpStreamParameters->iBufferSize ), 
                                            config.fileReader->iFileProducer,
					                        (void*)(config.fileReader), BO_NOIBLK );
  assert(config.fileReaderBuffer != NULL);

  // sound_accept interface
  ec=AudioAcceptInit(
      BUF_SIZE,
      BUF_COUNT,
      "/var/run/mp3.sem",
      config.audiosockname,
      0,
      tmpStreamParameters->iSamplingRate,
      tmpStreamParameters->iNumChannels,
      tmpStreamParameters->iBitsPerSample,
      SAH_NOTIFY_STOP | SAH_NOTIFY_SYNC );
  
  if( ec < 0 )
  {
   DBG( "FATAL: Can not connect to audio server, exiting !!!\n" );
   exit( 1 );
  }

    // mp3 decoder
  DBG("Setting up mp3 decoder..\n");
  memset(&(config.madOpt),0,sizeof(MadOpt));
  pthread_mutex_init(&(config.madOpt.skipMutex),0);
  pthread_mutex_init(&(config.madOpt.resumeMutex),0);
  pthread_cond_init(&(config.madOpt.resumeCond),0);

  config.madOpt.options         = MO_PCM | MO_SWP;
  config.madOpt.h               = config.fileReaderBuffer;
  config.madOpt.fileReader      = config.fileReader;
  config.madOpt.haveAudioAccept = 1;
  config.madOpt.skipped         = 1;
  config.madOpt.timeStamp       = 0;
  config.madOpt.bytesRead       = config.fileReader->openFilePos;
  
  config.madProducerBuffer=InitStreamBufferSet(
                            BUF_SIZE,
                            config.fileReader->iDecoderProducer,
                            (void*)&(config.madOpt),
							BO_NOIBLK,
                            (void*)AudioAcceptBuffer() );
  if( config.madProducerBuffer == NULL )
  {
   DBG( "FATAL: Can not creare MAD stream buffer, exiting\n" );
   AudioAcceptDestroy();
   exit( 1 );
  }

  // set up outgoing connection to audio_accept
  DBG("Init done, entering main loop..\n");

	// main loop
  while( !config.done ) {    

    usleep(100000);

    DispatchNextUIMessage( &tmpUIController, ENoExitIfError );          

	// this is usually normal, but might be an error:
    if(config.madOpt.terminated) 
    {
        DBG( "MAD thread terminated, exiting...\n" );
        config.done=1;
    }

	// this only happens when there is an error:
    if( NULL != config.fileReader )
    {
     if(config.fileReader->terminated)
     {
        DBG("fileReader thread terminated, exiting...\n");
        config.done=1;
     }

     if( config.fileReader->switchedToNext )
     {
      config.fileReader->switchedToNext = 0;
      strcpy( config.currentFileName, config.nextFileName );

      *config.nextFileName = 0;

      config.oldBytePos = 0;
     
      DBG( "Song switched to %s\n", config.currentFileName );
     }
    }
  }

  DBG( "Finishing main loop, exit mode %d\n", config.stop_mode );

  
  if( config.stop_mode )
  {
   DBG( "EXITING IN SUDDEN DEATH MODE NOW\n" );
   DoneUIController( &tmpUIController );
   
   DBG( "Killing MAD thread...\n" );
   pthread_kill( config.madProducerBuffer->thread, SIGKILL );

   DBG( "Killing File Reader thread...\n" );
   pthread_kill( config.fileReaderBuffer->thread, SIGKILL );

   DBG( "Done AudioAccept...\n" );
   AudioAcceptDestroy();

   DBG( "SUDDEN DEATH...\n" );
   return 0;
  }

  // Give UI chance to read 100%
  for( ec = 0; ec < 10; ++ec )
  {
   if( DispatchNextUIMessage( &tmpUIController, EExitIfError ) )
   {
    break;
   }

   usleep(100000);
  }


	// de-init everything (in reverse order)
//destoy:
  DBG( "Freeing resources...\n" );

#ifdef FULL_DEBUG
   DBG( "BR %d MAD %d FR %d R file %s\n", 
        config.madOpt.bytesRead, 
        config.madOpt.fileSize, 
        config.fileReader->fileSize,
        config.currentFileName );
#endif

  // mp3 decoder
  //  = madProducerBuffer 
  if( NULL != config.madProducerBuffer )
  {
   DBG("free> madProducerBuffer\n");
   DoneStreamBufferWakeUp( config.madProducerBuffer,1,1,
				&(config.madOpt.resumeMutex),
				&(config.madOpt.resumeCond) );
  }

#ifdef FULL_DEBUG
   DBG( "BR %d MAD %d FR %d file %s\n", 
        config.madOpt.bytesRead, 
        config.madOpt.fileSize, 
        config.fileReader->fileSize,
        config.currentFileName );
#endif

  // 
  // file reader
  if( NULL != config.fileReaderBuffer )
  {
   DBG("free> fileReaderBuffer\n");
   DoneStreamBufferWakeUp(config.fileReaderBuffer,0,3,
				&(config.fileReader->resumeMutex),
				&(config.fileReader->resumeCond),
				&(config.fileReader->nextMutex),
				&(config.fileReader->nextCond),
				&(config.fileReader->wakeupMutex),
				&(config.fileReader->wakeupCond) );
  }

#ifdef FULL_DEBUG
   DBG( "BR %d MAD %d FR %d file %s\n", 
        config.madOpt.bytesRead, 
        config.madOpt.fileSize, 
        config.fileReader->fileSize,
        config.currentFileName );
#endif

  if( NULL != config.fileReader )
  {
   DBG("free> fileReader\n");
   config.fileReader->iDestroyFileProducer( config.fileReader );
  }

  pthread_mutex_destroy(&(config.madOpt.skipMutex));
  pthread_mutex_destroy(&(config.madOpt.resumeMutex));
  pthread_cond_destroy(&(config.madOpt.resumeCond));

  // Resetting all fields
  config.madProducerBuffer = NULL;
  config.fileReaderBuffer  = NULL;
  config.fileReader        = NULL;

  DBG( "Done AudioAccept...\n" );
  AudioAcceptDestroy();

  if( config.isDaemon ) 
  {
    DBG("daemon mode -> restart\n");
    goto restart;
  }

  DoneUIController( &tmpUIController );

  DBG( "Exiting...\n" );

  return 0;
}


// UI controller management
static int StartPlayback( const char* aFileName, const char* aCodecName, unsigned long aByteOffset )
{ 
 DBG("Setting up file reader...\n");
 
 assert( NULL != aFileName );

 int tmpResult = 1;

 *config.lastErrorText = 0;

 // temporary hack to test audiobooks
 //aFileName = "/mnt/sdcard/Audiobooks/test1.aa";
 //aCodecName = "/lib/audible.so";

 DBG("Filename %s CodecName %s, offset %ld\n", aFileName, aCodecName, aByteOffset );

 if( NULL == config.fileReader )
 {
  strcpy( config.currentFileName, aFileName );

  config.madOpt.fileSize  = 0;
  config.madOpt.bytesRead = 0;

  if( NULL == aCodecName )
  {
   config.fileReader  = InitMP3FileProducer( config.currentFileName, aByteOffset ); 
   config.customCodec = 0;
  }
  else
  { // Custom codec
   config.fileReader  = InitPlugInFileProducer( config.currentFileName, aCodecName, aByteOffset, config.lastErrorText );

   if( NULL != config.fileReader )
   {
    config.customCodec = 1;
   }
  }

  if( NULL == config.fileReader )
  {
   tmpResult = 0;
   if( !*config.lastErrorText ) 
   {
    strcpy( config.lastErrorText, strerror( errno ) );
   }
  }
  else
  {
   DBG( "Current song: %s\n", config.currentFileName );
  }
 }
 else
  if( config.customCodec )
  {
   strcpy( config.lastErrorText, "Next song not supported in for custom codecs" );
   tmpResult = 0;

   DBG( "%s\n", config.lastErrorText );
  }
  else
  {
   strcpy( config.nextFileName, aFileName );
  
   // enqueue next mp3 (test)
   if( !config.fileReader->iSetFileProducerNextFile( config.fileReader, config.nextFileName ) )
   {
    tmpResult = 0;
    strcpy( config.lastErrorText, strerror( errno ) );
   }
   else
   {
    DBG( "Next song: %s\n", config.nextFileName );
   }
  }

 DBG( "File reader set up, result %d\n", tmpResult );
 
 return tmpResult;
}


static int PausePlayback( int aMode )
{
 if( aMode >= 0 )
 {
  config.paused = ( aMode + 1 ) % 2;
 }

 DBG( "Paused state %d, mode %d\n", config.paused, aMode );
 
 if(config.paused) {   
   AudioAcceptSendResume();
   config.paused = 0;
 } else {
   AudioAcceptSendPause();
   config.paused = 1;
 } 
 return 1;
}

static int StopPlayback( int aMode )
{
 int tmpResult = 1; 

 DBG( "Stop playback, mode %d\n", aMode );
 
 *config.lastErrorText = 0;

 config.stop_mode = aMode;   
 config.done      = 1;
 
 //config.madOpt.fileSize  = 0;
 //config.madOpt.bytesRead = 0;

 if( NULL == config.fileReader )
 {
  tmpResult = 0;
  strcpy( config.lastErrorText, "Playback not started" );
 }

 return tmpResult;
}

/*static void ClearBuffers( void )
{
  DBG( "Stopping File Reader buffer flush...\n" );
  InterruptFlushStreamBuffer( config.fileReaderBuffer );

  DBG( "Stopping MAD buffer flush...\n" );
  InterruptFlushStreamBuffer( config.madProducerBuffer );

  // Clearing buffers
  DBG( "Discarding MAD buffer...\n" );
  DiscardStreamBufferFull( config.madProducerBuffer );

  DBG( "Discarding File Reader buffer...\n" );
  DiscardStreamBufferFull( config.fileReaderBuffer );
}*/

static int SkipPlayback( const char *aFileName )
{
  int tmpResult=0;
  int skipRetries=1000; // 10 seconds

  DBG( "Skip playback to %s...\n", aFileName );

  assert( NULL != aFileName );
  
	// if another skip is still in progress, wait
  while(config.madOpt.skipped==0 && --skipRetries) usleep(10000);

  if(skipRetries==0) {
    fprintf(stderr,"skip timeout!!\n");
    return 0;
  }

  config.madOpt.skipped=0;

    // stop playback immediately, better feedback effect UI-wise
  DiscardAndWait();
  //AudioAcceptSendStop();
  //AudioAcceptWaitForStop();

  DBG("received STOP notify\n");

    // reset paused flag
  config.paused=0;

    // if required, set next file name
  if(strlen(aFileName)) {
    strcpy( config.nextFileName, aFileName );
  
    if( !config.fileReader->iSetFileProducerNextFile( config.fileReader, config.nextFileName ) )
    {
     tmpResult = 1;
     strcpy( config.lastErrorText, strerror( errno ) );
    }
    else
    {
     DBG( "Next song: %s\n", config.nextFileName );
    }
  }

  pthread_mutex_lock(&(config.fileReader->skipMutex));
  config.fileReader->skip=1;
  pthread_mutex_unlock(&(config.fileReader->skipMutex));

  DBG("fileReader->skip set\n");

  pthread_mutex_lock(&(config.madOpt.skipMutex));
  config.madOpt.skip=1;
  pthread_mutex_unlock(&(config.madOpt.skipMutex));

  DBG("madOpt->skip set\n");

  //ClearBuffers();  

  config.madOpt.fileSize  = 0;
  config.madOpt.bytesRead = 0;

  ResumeDiscardAndWait();
  
  //AudioAcceptNotifyConditional();
 
  return tmpResult;
}

static int GetCurrentPlayPercentage( void )
{
 unsigned long long tmpResult = 0;

 if( NULL == config.fileReader )
 {
  return 0; // Playback not started
 }

#ifdef FULL_DEBUG
#ifdef DEBUG
 static int tmpD = 0;

 if(!config.madOpt.bytesRead) {
   DBG("!!!!!!! bytesRead is zero...\n");
 }
#endif
#endif

 if( config.madOpt.fileSize )
 { 
  tmpResult = config.madOpt.bytesRead;
  tmpResult = tmpResult * 100LL / config.madOpt.fileSize;
 }
 
 if( tmpResult >= 100 )
 {
  if( config.fileReader->switchedToNext || 
      config.madOpt.terminated || 
      config.fileReader->terminated || 
      config.fileReader->isIdle )
  {
   tmpResult = 100; // Last read
  }
  else
  {
   tmpResult = 99;
  }
 }

#ifdef FULL_DEBUG
#ifdef DEBUG
  if( !( ++tmpD % 10 ) )//|| tmpResult > 97 )
  {
   DBG( "BR %d MAD %d FR %d R %lld, file %s\n", 
        config.madOpt.bytesRead, 
        config.madOpt.fileSize, 
        config.fileReader->fileSize,
        tmpResult,
        config.currentFileName );
  }
#endif
#endif

  return tmpResult;
}

static void CalculatePlaybackChapterAndTime( void )
{
 if( NULL != config.fileReader )
 {
  if( config.oldBytePos != config.madOpt.bytesRead )
  {
   config.fileReader->iBytePosToChapter( config.fileReader, config.madOpt.bytesRead, &config.currentChapter, &config.currentTime );
  
   config.oldBytePos = config.madOpt.bytesRead;

   //DBG( "Byte pos %d, Chapter %ld, Time %ld\n", config.madOpt.bytesRead, config.currentChapter, config.currentTime );
  }

#ifdef FULL_DEBUG
#ifdef DEBUG
  static int dd = 0;

  if( !( dd++ %10 ) )
  {
   DBG( "File %s, Byte pos %d, Chapter %ld, Time %ld\n", config.currentFileName, config.madOpt.bytesRead, config.currentChapter, config.currentTime );
  }
#endif
#endif
 }
}

static unsigned long GetCurrentChapter( void )
{
 CalculatePlaybackChapterAndTime();
  
 return config.currentChapter;
}

static unsigned long GetCurrentPlaybackPosition( void )
{
 CalculatePlaybackChapterAndTime();
 
 return ( config.currentTime ) ? config.currentTime: config.madOpt.timeStamp; 
}

static const char* GetCurrentSongName( void )
{
 return config.currentFileName;
}

static const char* GetNextSongName( void )
{
 return config.nextFileName;
}

static const char* GetLastError( void )
{ 
 return config.lastErrorText;
}

static int StartFeatureMode( int aMode, int aSpeed )
{
 int tmpResult = 0;
 
 *config.lastErrorText = 0;

 if( NULL != config.fileReader )
 { 
  if( config.fileReader->isIdle && aMode && config.madOpt.bytesRead >= config.fileReader->fileSize )
  {
   DBG( "Playback in idle mode, can not start feature mode !\n" );
  }
  else
  {
   if( ( int )config.currentFeature != aMode && aMode ) // Stopping only for beginning of feature
   {
    DiscardAndWait();   
   }
  
   config.currentFeature = aMode;

   DBG( "Starting feature...\n" );
   tmpResult = config.fileReader->iStartFeatureMode( config.fileReader, aMode, aSpeed, config.lastErrorText );

   // Resuming
   ResumeDiscardAndWait();
  }
 }
 else
 {
  strcpy( config.lastErrorText, "Playback not started" );
 }

 return tmpResult;
}

static int GoToChapter( int aChapterNumber, int aAbsolute )
{
 int tmpResult = 0;

 *config.lastErrorText = 0;

 if( NULL != config.fileReader )
 {
  // Stopping any feature
  StartFeatureMode( ENoFeature, 0 );

  // Discarding buffers
  DiscardAndWait();
  
  // Jumping
  tmpResult = config.fileReader->iGoToChapter( config.fileReader, aChapterNumber, aAbsolute, config.lastErrorText );

  // Resuming
  ResumeDiscardAndWait();
 }
 else
 {
  strcpy( config.lastErrorText, "Playback not started" );
 }

 return tmpResult;
}

static void DiscardAndWait(void) {
  time_t timeStart;
  int    tmpRetry = 10;

  // Killing all instantly first
  if( !config.fileReader->discardAndWait )
  {
   config.fileReader->discardAndWait = 2;
  }

  if( !config.madOpt.discardAndWait )
  {
   config.madOpt.discardAndWait = 2;  
  }

  // TODO: check errors on AudioAccept calls  
  AudioAcceptSendStop();
  AudioAcceptWaitForStop();

  assert( config.fileReader->discardAndWait );
  assert( config.madOpt.discardAndWait );

    // filereader
  while(config.fileReader->discardAndWait==2) 
  {
    DBG("kicking filereader into discard mode\n");  

    pthread_mutex_lock(&(config.fileReader->nextMutex));
    pthread_cond_signal(&(config.fileReader->nextCond));
    pthread_mutex_unlock(&(config.fileReader->nextMutex));
  
    pthread_mutex_lock(&(config.fileReader->wakeupMutex));
    pthread_cond_signal(&(config.fileReader->wakeupCond));
    pthread_mutex_unlock(&(config.fileReader->wakeupMutex));
  
    InterruptFlushStreamBuffer(config.fileReaderBuffer);

    DBG("waiting for filereader to go into discard mode\n");
    timeStart=time(0);
    while( ((time(0)-timeStart)<DISCARD_TIMEOUT) && (config.fileReader->discardAndWait==2) ) {
      usleep(1000); // TODO: some timeout
    }
  }
  DBG("filereader went into discard mode\n");

  // madProducer
  while(config.madOpt.discardAndWait==2 && tmpRetry-- ) {
    DBG("kicking madProducer into discard mode\n");  

    InterruptFlushStreamBuffer(config.madProducerBuffer);

    AudioAcceptNotifyConditional();

    DBG("waiting for madproducer to go into discard mode\n");
    timeStart=time(0);
    while( ((time(0)-timeStart)<DISCARD_TIMEOUT) && (config.madOpt.discardAndWait==2) ) 
    { 
      usleep(1000); // TODO: some timeout
    }
  }

  if( config.madOpt.discardAndWait != 1 )
  {
   // Forcing the discard anyway
   config.madOpt.discardAndWait = 1;

   DBG( "ERROR MAD PRODUCER NOT IN DISCARD MODE !!!" );
  }
  else
  {
   DBG("madproducer went into discard mode\n");
  }

  DBG( "Paused %d\n", config.paused );
}

static void ResumeDiscardAndWait(void) {

   DBG( "RESUMING from Discard\n" );

   config.fileReader->discardAndWait=0;
   pthread_mutex_lock(&(config.fileReader->resumeMutex));
   pthread_cond_signal(&(config.fileReader->resumeCond));
   pthread_mutex_unlock(&(config.fileReader->resumeMutex));

   config.madOpt.discardAndWait=0;
   pthread_mutex_lock(&(config.madOpt.resumeMutex));
   pthread_cond_signal(&(config.madOpt.resumeCond));
   pthread_mutex_unlock(&(config.madOpt.resumeMutex));

   config.oldBytePos = 0; // Reset the position to automatically calculate chapters 
}

static unsigned long GetCurrentByteOffset( void )
{
 return config.madOpt.bytesRead;
}

static void InitInternalUIController( ui_contoller_t* aController )
{
 assert( NULL != aController );

 aController->iStartPlayback              = StartPlayback;
 aController->iPausePlayback              = PausePlayback;
 aController->iStopPlayback               = StopPlayback;
 aController->iSkip                       = SkipPlayback;
 aController->iGetCurrentPlayPercentage   = GetCurrentPlayPercentage;
 aController->iGetCurrentPlaybackPosition = GetCurrentPlaybackPosition;
 aController->iGetCurrentChapter          = GetCurrentChapter;
 aController->iGetCurrentSongName         = GetCurrentSongName;
 aController->iGetNextSongName            = GetNextSongName;
 aController->iGetLastError               = GetLastError;
 aController->iFeatureMode                = StartFeatureMode;
 aController->iGoToChapter                = GoToChapter;
 aController->iGetCurrentByteOffset       = GetCurrentByteOffset;
}
